/**
 * cfo/fo_methods/p2/P2impl.hpp
 * @author The RECORD Development Team (INRA) and the VLE Development Team
 * See the AUTHORS or Authors.txt file
 */

/*
 * Copyright (C) 2009-2010 INRA
 *
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef CFO_FO_METHODS_P2_IMPL_HPP_
#define CFO_FO_METHODS_P2_IMPL_HPP_

#include <vle/value/Map.hpp>
#include <vle/utils/Rand.hpp>

#include <cfo/tools/VleSimulator.hpp>
#include <cfo/tools/accumulators/accu_multi.hpp>

#include <cfo/fo_methods/p2/leaf_cutting.hpp>
#include <cfo/fo_methods/p2/accu_leaf.hpp>
#include <cfo/fo_methods/p2/accu_strat.hpp>
#include <cfo/fo_methods/p2/factory_accu.hpp>
#include <cfo/fo_methods/p2/leaf_simulation.hpp>
#include <cfo/fo_methods/p2/leaf_selection.hpp>
#include <cfo/fo_methods/p2/utils.hpp>


namespace cfo {
namespace fo_methods {
namespace p2 {

namespace ct = cfo::tools;


/**
 * @brief P2 alternative implementation.
 * template class DIM_TYPE is
 */
template<class DIM_TYPE, st::AccuType STRAT_ACCU_TYPE, CutType CUT_TYPE,
LeafPerformanceType PERF_TYPE, LeafDispersionType DISP_TYPE, LeafSelectionType LEAF_SEL_TYPE>

class P2impl
{

    const static double MAX_DOUBLE = 2000000;
    const static double MIN_DOUBLE = -2000000;

    typedef StratP2<DIM_TYPE, STRAT_ACCU_TYPE> strat_type;
    typedef boost::shared_ptr<strat_type> strat_type_ptr;
    typedef LeafP2<strat_type, PERF_TYPE, DISP_TYPE> leaf_type;
    typedef boost::shared_ptr<leaf_type> leaf_type_ptr;
    typedef AccuFactory<leaf_type> factory_type;
    typedef LeafSimulation<leaf_type> simulation_type;
    typedef Cutting<leaf_type, CUT_TYPE> cutting_type;
    typedef LeafPool<leaf_type, LEAF_SEL_TYPE> leaf_pool_type;

private:

    factory_type maccuFactory;
    /**
     *@brief abstract tool for the simulation of a leaf
     */
    simulation_type msim;
    /**
     *@brief abstract tool to cut a leaf in two parts
     */
    cutting_type mcutTool;
    /**
     *@brief object storing the set of leaves that can be cut in two sub leaves.
     */
    leaf_pool_type mleafPool;

    /**
     * @brief Maximum number of simulations
     */
    int mmaxSimulationNbr;
    /**
     * @brief minimal target value to reach. Once it is reached, the
     * optimization process ends.
     */
    double mtarget;

public:

    P2impl(ct::VleSimulator& sim, const P2args& p2args, vu::Rand& rand) :
        maccuFactory(p2args), msim(sim, p2args, rand), mcutTool(p2args),
        mleafPool(p2args), mmaxSimulationNbr(p2args.max_sim_number),
        mtarget(p2args.eval_target)
    {
    }

    ~P2impl()
    {
    }
    void optimises(vv::Map& results)
    {
        int nbLoops = 0;
        int nbSimulsGlob = 0;
        double bestPerf = MAX_DOUBLE;
        leaf_type_ptr father  = maccuFactory.buildLeaf(0);
        leaf_type_ptr child1;
        leaf_type_ptr child2;
        leaf_type_ptr bestLeaf = father;

        if (father->getNbPotentialSubLeaves() > 0) {
            mleafPool.pushLeaf(father);
            nbSimulsGlob += msim.simulates(father);
            bestPerf = father->performance();
        }

        while ((mleafPool.size() > 0) &&
                (bestPerf > mtarget) &&
                (nbSimulsGlob < mmaxSimulationNbr)) {
            nbLoops++;

            /** leaf selection **/
            father = mleafPool.popLeaf();

            /** cutting process **/
            std::pair<leaf_type_ptr, leaf_type_ptr>
            childs = mcutTool.cuts(father);
            child1 = childs.first;
            child2 = childs.second;
            /** for each child: simulate **/
            int nbSimulsLoc1 = msim.simulates(child1);
            int nbSimulsLoc2 = msim.simulates(child2);
            nbSimulsGlob += nbSimulsLoc1 + nbSimulsLoc2;
            /** update best region known **/
            if (child1->performance() < bestPerf) {
                bestLeaf = child1;
                bestPerf = bestLeaf->performance();
            }
            if (child2->performance() < bestPerf) {
                bestLeaf = child2;
                bestPerf = bestLeaf->performance();
            }
            /*** update lists (pendingList)***/
            if (child1->getNbPotentialSubLeaves() > 1) {
                mleafPool.pushLeaf(child1);
            }
            if (child2->getNbPotentialSubLeaves() > 1) {
                mleafPool.pushLeaf(child2);
            }

        }//main loop ends
        results.addInt("nb-loops", nbLoops);
        results.addInt("nb-simulations", nbSimulsGlob);
        results.addInt("leaf-pool-size",mleafPool.size());
        results.addDouble("min-performance", bestPerf);
        results.addDouble("target",mtarget);
        results.addBoolean("stop-because-target-reached",
            (bestPerf <= mtarget));
        results.addBoolean("stop-because-empty-leaf-pool",
            (mleafPool.size() <= 0));
        results.addBoolean("stop-because-max-sim-reached",
            (nbSimulsGlob >= mmaxSimulationNbr));
        vv::Set& best_leaf = results.addSet("best-leaf");
        for (unsigned int i = 0; i < bestLeaf->getInputSize(); i++) {
            vv::Map& oneInput = best_leaf.addMap();
            oneInput.addDouble("min",bestLeaf->getDomain(i).min());
            oneInput.addDouble("max",bestLeaf->getDomain(i).max());
        }
    }
};

}
}
}//namespaces

#endif
