/**
 * cfo/fo_methods/p2/accu_leaf.hpp
 * @author The RECORD Development Team (INRA) and the VLE Development Team
 * See the AUTHORS or Authors.txt file
 */

/*
 * Copyright (C) 2009-2010 INRA
 *
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef CFO_FO_METHODS_P2_ACCU_LEAF_HPP_
#define CFO_FO_METHODS_P2_ACCU_LEAF_HPP_

#include <vector>
#include <iostream>
#include <limits>
#include <math.h>
#include <boost/shared_ptr.hpp>

#include <cfo/fo_methods/p2/accu_leaf_alternatives.hpp>
#include <cfo/fo_methods/p2/utils.hpp>

namespace cfo {
namespace fo_methods {
namespace p2 {

namespace st = cfo::tools;

/**
 * @brief LeafP2 instantiation class. A Leaf is seen as an accumulator of strats.
 * It provides performance and dispersion heuristics. It also provides acces
 * to a strats. Alternatives depend on choice of the computation and dispersion
 * heuristics.
 */
template<class STRAT, LeafPerformanceType PERF_TYPE = LEAF_PERF_MEAN,
        LeafDispersionType DISP_TYPE = LEAF_DISP_GLOBAL_VAR>
class LeafP2: public ct::Leaf
{
public:
    typedef typename STRAT::dim_type dim_type;
    typedef boost::shared_ptr<STRAT> strat_type_ptr;
    typedef STRAT strat_type;
    typedef typename std::vector<strat_type_ptr> const_iterator;

    /**
     * @brief LeafP2 constructor
     * @param  p2args, the set of P2 initializations
     * @param depth, the depth of the leaf
     */
    LeafP2(const P2args& p2args, unsigned int depth) :
        Leaf(p2args.leaf_params, depth), mperfAccu(p2args),
        mdispAccu(p2args), mstrats(), mperformance(), mdispersion(),
        mupdated(false)
    {
    }
    /**
     * @brief Inserts a Strat into the accumulator
     * @param strat the strategy (point) to insert
     */
    void insert(const strat_type_ptr& strat)
    {
        mstrats.push_back(strat);
        mperfAccu.insert(strat);
        mdispAccu.insert(strat);
        mupdated = false;
    }
    /**
     * @brief Inserts a Strat into the accumulator
     * @param strat the strategy (point) to insert
     */
    unsigned int count() const
    {
        return mstrats.size();
    }
    /**
     * @brief Performance heuristic extractor
     * @return the performance value (either a MultiDim or a MonoDim)
     */
    dim_type performance()
    {
        if(!mupdated){
            mperformance = mperfAccu.performance(mstrats);
            mdispersion = mdispAccu.dispersion(mstrats);
            mupdated = true;
        }
        return mperformance;
    }
    /**
     * @brief Dispersion heuristic extractor
     * @return the dispersion value (a MonoDim)
     */
    dim_type dispersion()
    {
        if(!mupdated){
            mperformance = mperfAccu.performance(mstrats);
            mdispersion = mdispAccu.dispersion(mstrats);
            mupdated = true;
        }
        return mdispersion;
    }
    /**
     * @brief Gets iterator on first strat
     * @return a iterator on first strat
     */
    const_iterator begin()
    {
        return mstrats.begin();
    }
    /**
     * @brief Gets iterator on the end of the strat vector
     * @return a iterator on the end of the strat vector
     */
    const_iterator end()
    {
        return mstrats.end();
    }
    /**
     * @brief Access to the index-th strat stored.
     * @param index, the index of strat
     * @return a pointer on the index-th strat.
     */
    strat_type_ptr access(unsigned int index)
    {
        return mstrats[index];
    }
    /**
     * @brief Gets a reference on the strat vector
     * @return the set of strats
     */
    const std::vector<strat_type_ptr>& getStrats()
    {
        return mstrats;
    }
private:
    /**
     * @brief The performance evaluation alternative
     */
    LeafPerfAccu<STRAT, PERF_TYPE> mperfAccu;
    /**
     * @brief The dispersion evaluation alternative
     */
    LeafDispAccu<STRAT, DISP_TYPE> mdispAccu;
    /**
     * @brief the set of strat accumulated.
     */
    std::vector<strat_type_ptr> mstrats;
    /**
     * @brief Last performance computed
     */
    dim_type mperformance;
    /**
     * @brief Last dispersion computed
     */
    dim_type mdispersion;
    /**
     * @brief Bool to true if last computations of dispersion and performance
     * are still reliable.
     */
    bool mupdated;

};

}
}
}//namespace cfo::fo_methods::p2

#endif //CFO_FO_METHODS_P2_ACCU_LEAF_HPP_
