/**
 * cfo/fo_methods/p2/accu_leaf_alternatives.hpp
 * @author The RECORD Development Team (INRA) and the VLE Development Team
 * See the AUTHORS or Authors.txt file
 */

/*
 * Copyright (C) 2009-2010 INRA
 *
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#ifndef CFO_FO_METHODS_P2_LEAF_ALTERNATIVES_HPP_
#define CFO_FO_METHODS_P2_LEAF_ALTERNATIVES_HPP_

#include <vector>
#include <iostream>
#include <limits>
#include <math.h>
#include <boost/shared_ptr.hpp>

#include <cfo/tools/accumulators/accu_multi.hpp>
#include <cfo/fo_methods/p2/utils.hpp>

namespace cfo {
namespace fo_methods {
namespace p2 {


namespace st = cfo::tools;

/**
 * @brief LeafPerfAccu declaration.
 * LeafPerfAccu classes instantiations are thought as Strat accumulators
 * and should provide the heuristic performance. Performance computation
 * alternatives are listed in the type LeafPerformanceType
 */
template<class STRAT, LeafPerformanceType PERF_TYPE> class LeafPerfAccu;
/**
 * @brief LeafDispAccu declaration.
 * LeafDispAccu classes instantiations are thought as Strat accumulators
 * and should provide the heuristic dispersion. Dispersion computation
 * alternatives are listed in the type LeafDispersionType
 */
template<class STRAT, LeafDispersionType DISP_TYPE> class LeafDispAccu;

/**
 * @brief Partial instantiation of a LeafPerfAccu based on heuristic
 * LEAF_PERF_MEAN.
 */
template<class STRAT>
class LeafPerfAccu<STRAT, LEAF_PERF_MEAN>
{
public:
    typedef boost::shared_ptr<STRAT> strat_type_ptr;
    typedef typename STRAT::dim_type dim_type;
    /**
     * @brief LeafPerfAccu constructor
     * @param p2args, the set of P2 initializations
     */
    LeafPerfAccu(const P2args& /*p2args*/)
    {
    }
    /**
     * @brief Inserts a Strat into the accumulator
     * @param strat the strategy (point) to insert
     */
    void insert(const strat_type_ptr& strat)
    {
        mstats.insert(strat->mean());
    }
    /**
     * @brief Performance heuristic extractor
     * @param the set of strored strats.
     * @return the performance value (either a MultiDim or a MonoDim)
     */
    dim_type performance(const std::vector<strat_type_ptr>&)
    {
        return mstats.mean();
    }
private:
    /**
     * @brief Accumulator used for performance computation
     */
    ct::Accu<dim_type, ct::MEAN_VAR> mstats;
};

/**
 * @brief Partial instantiation of a LeafPerfAccu based on heuristic
 * LEAF_PERF_MIN.
 */
template<class STRAT>
class LeafPerfAccu<STRAT, LEAF_PERF_MIN>
{
public:

    typedef boost::shared_ptr<STRAT> strat_type_ptr;
    typedef typename STRAT::dim_type dim_type;
    /**
     * @brief LeafPerfAccu constructor
     * @param p2args, the set of P2 initializations
     */
    LeafPerfAccu(const P2args& /*p2args*/)
    {
    }
    /**
     * @brief Inserts a Strat into the accumulator
     * @param strat the strategy (point) to insert
     */
    void insert(const strat_type_ptr& strat)
    {
        mstats.insert(strat->mean());
    }
    /**
     * @brief Performance heuristic extractor
     * @param the set of strored strats.
     * @return the performance value (either a MultiDim or a MonoDim)
     */
    dim_type performance(const std::vector<strat_type_ptr>&)
    {
        return mstats.min();
    }

private:
    /**
     * @brief Accumulator used for performance computation
     */
    ct::Accu<dim_type, ct::MEAN_VAR> mstats;
};

/**
 * @brief Partial instantiation of a LeafPerfAccu based on heuristic
 * LEAF_PERF_QUANTILE.
 */
template<class STRAT>
class LeafPerfAccu<STRAT, LEAF_PERF_QUANTILE>
{

public:

    typedef boost::shared_ptr<STRAT> strat_type_ptr;
    typedef typename STRAT::dim_type dim_type;
    /**
     * @brief LeafPerfAccu constructor
     * @param p2args, the set of P2 initializations
     */
    LeafPerfAccu(const P2args& p2args) :
        mleafQuantileOrder(p2args.eval_leaf_quantile_order),
        mstratQuantileOrder(p2args.eval_strat_quantile_order)
    {
    }
    /**
     * @brief Inserts a Strat into the accumulator
     * @param strat the strategy (point) to insert
     */
    inline void insert(const strat_type_ptr& strat)
    {
        mstats.insert(strat->quantile(mstratQuantileOrder));
    }
    /**
     * @brief Performance heuristic extractor
     * @param the set of strored strats.
     * @return the performance value (either a MultiDim or a MonoDim)
     */
    inline dim_type performance(const std::vector<strat_type_ptr>&)
    {
        return mstats.quantile(mleafQuantileOrder);
    }

private:
    /**
     * @brief Accumulator used for performance computation
     */
    ct::Accu<dim_type, ct::QUANTILE> mstats;
    /**
     * @brief Quantile order to compute quantile value at leaf scale
     */
    double mleafQuantileOrder;
    /**
     * @brief Quantile order to compute quantile value at strat scale
     */
    double mstratQuantileOrder;
};

/**
 * @brief Partial instantiation of a LeafDispAccu based on heuristic
 * LEAF_DISP_GLOBAL_VAR. Note : only the MonoDim is implemented yet.
 * To use the MultiDim version, addition and multiplication should
 * be implemented for MultiDim data.
 */
template<class STRAT>
class LeafDispAccu<STRAT, LEAF_DISP_GLOBAL_VAR>
{

public:

    typedef boost::shared_ptr<STRAT> strat_type_ptr;
    typedef typename STRAT::dim_type dim_type;
    /**
     * @brief LeafDispAccu constructor
     * @param p2args, the set of P2 initializations
     */
    LeafDispAccu(const P2args& /*p2args*/)
    {
    }
    /**
     * @brief Inserts a Strat into the accumulator
     * @param strat the strategy (point) to insert
     */
    inline void insert(const strat_type_ptr& strat)
    {
        mcount.insert(strat->count());
        msquareSum.insert(strat->squareSum());
        msum.insert(strat->sum());
    }
    /**
     * @brief Dispersion heuristic extractor
     * @param the set of strored strats.
     * @return the dispersion value (a MonoDim)
     */
    inline dim_type dispersion(const std::vector<strat_type_ptr>&)
    {
        double sum = msum.sum();
        double squareSum = msquareSum.sum();
        double count = mcount.sum();
        return (squareSum / count - pow(sum / count, 2));
    }

private:
    /**
     * @brief Count the number of strats results accumulated
     */
    ct::Accu<dim_type, ct::MEAN_VAR> mcount;
    /**
     *@brief Accumulator of square sum of strat
     */
    ct::Accu<dim_type, ct::MEAN_VAR> msquareSum;
    /**
     *@brief Accumulator of square sum of strat
     */
    ct::Accu<dim_type, ct::MEAN_VAR> msum;


};


/**
 * @brief Partial instantiation of a LeafDispAccu based on heuristic
 * LEAF_DISP_SPATIAL_VAR1. Note : only the MonoDim is implemented yet.
 * To use the MultiDim version, addition and multiplication should
 * be implemented for MultiDim data.
 */
template<class STRAT>
class LeafDispAccu<STRAT, LEAF_DISP_SPATIAL_VAR1>
{

public:

    typedef boost::shared_ptr<STRAT> strat_type_ptr;
    typedef typename STRAT::dim_type dim_type;
    /**
     * @brief LeafDispAccu constructor
     * @param p2args, the set of P2 initializations
     */
    LeafDispAccu(const P2args& /*p2args*/)
    {
    }
    /**
     * @brief Inserts a Strat into the accumulator
     * @param strat the strategy (point) to insert
     */
    inline void insert(const strat_type_ptr& strat)
    {
        mstats.insert(strat->mean());
    }
    /**
     * @brief Dispersion heuristic extractor
     * @param the set of strored strats.
     * @return the dispersion value (a MonoDim)
     */
    inline dim_type dispersion(const std::vector<strat_type_ptr>&)
    {
        return mstats.deviationSquareSum() / mstats.count();
    }

private:
    /**
     * @brief Accumulator used for performance computation
     */
    ct::Accu<dim_type, ct::MEAN_VAR> mstats;
};


/**
 * @brief Partial instantiation of a LeafDispAccu based on heuristic
 * LEAF_DISP_SPATIAL_VAR2. Note : only the MonoDim is implemented yet.
 * To use the MultiDim version, addition and multiplication should
 * be implemented for MultiDim data.
 */
template<class STRAT>
class LeafDispAccu<STRAT, LEAF_DISP_SPATIAL_VAR2>
{

public:

    typedef boost::shared_ptr<STRAT> strat_type_ptr;
    typedef typename STRAT::dim_type dim_type;
    /**
     * @brief LeafDispAccu constructor
     * @param p2args, the set of P2 initializations
     */
    LeafDispAccu(const P2args& /*p2args*/) :
        mstatByReplica()
    {
    }
    /**
     * @brief Inserts a Strat into the accumulator
     * @param strat the strategy (point) to insert
     */
    inline void insert(const strat_type_ptr& strat)
    {
        unsigned int nbReplicas = strat->count();
        mstatByReplica.resize(nbReplicas);
        for (unsigned int repl = 0; repl < nbReplicas; repl++) {
            mstatByReplica[repl].insert(strat->access(repl));
        }
        mstatStrat.insert(strat->mean());
    }
    /**
     * @brief Dispersion heuristic extractor
     * @return the dispersion value (MonoDim data)
     */
    inline double dispersion(const std::vector<strat_type_ptr>& strats)
    {
        unsigned int nbStrats = mstatStrat.count();// =? strats.size()
        unsigned int nbReplicas = mstatByReplica.size(); //?= strats[0].count()
        double globalMean = mstatStrat.mean();
        double intern_sum = 0;
        double replicaMean = 0;
        double stratMean = 0;
        double v = 0;
        for (unsigned int s = 0; s < nbStrats; s++) {
            stratMean = strats[s]->mean();
            for (unsigned int r = 0; r < nbReplicas; r++) {
                replicaMean = mstatByReplica[r].mean();
                v = strats[s]->access(r);
                intern_sum += pow((v - stratMean - replicaMean + globalMean), 2);
            }
        }
        double sigma_res = intern_sum / ((nbStrats - 1) * (nbReplicas - 1));
        return (mstatStrat.deviationSquareSum() / (nbStrats - 1)) - sigma_res / (nbReplicas);

    }
private:
    /**
     * @brief Accumulator that provides statistics by replica
     */
    std::vector<ct::Accu<dim_type, ct::MEAN_VAR> > mstatByReplica;
    /**
     * @brief Accumulator that provides statistics by strat
     */
    ct::Accu<dim_type, ct::MEAN_VAR> mstatStrat;
};

}
}
}//namespace cfo::fo_methods::p2

#endif //CFO_FO_METHODS_P2_LEAF_ALTERNATIVES_HPP_
