/**
 * cfo/fo_methods/p2/accu_strat.hpp
 * @author The RECORD Development Team (INRA) and the VLE Development Team
 * See the AUTHORS or Authors.txt file
 */

/*
 * Copyright (C) 2009-2010 INRA
 *
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef CFO_FO_METHODS_P2_ACCU_STRAT_HPP_
#define CFO_FO_METHODS_P2_ACCU_STRAT_HPP_

#include <vector>
#include <iostream>
#include <limits>
#include <math.h>
#include <boost/shared_ptr.hpp>

#include <vle/value/Tuple.hpp>

#include <cfo/fo_methods/p2/utils.hpp>

namespace cfo {
namespace fo_methods {
namespace p2 {

namespace vv = vle::value;
namespace ct = cfo::tools;

/**
 *@brief StratP2 implementation. A strat is seen an accumulator of
 *simulation results.
 */
template<class DIM_TYPE, ct::AccuType ACCU_TYPE>
class StratP2: public vv::Tuple, public ct::Accu<DIM_TYPE, ACCU_TYPE>
{
public:

    typedef DIM_TYPE dim_type;
    /**
     * @brief StratP2 constructor
     * @param inputSize, the dimension of simulation input
     */
    StratP2(unsigned int inputSize) :
        vv::Tuple(inputSize), ct::Accu<DIM_TYPE, ACCU_TYPE>()
    {
    }
    /**
     * @brief StratP2 constructor
     * @param input, the point
     */
    StratP2(const vv::Tuple& input) :
        vv::Tuple(input), ct::Accu<DIM_TYPE, ACCU_TYPE>()
    {
    }
    /**
     * @brief StratP2 constructor
     * @param input, the point
     * @param p2args, the set of P2 initializations
     */
    StratP2(const vv::Tuple& input, const ct::AccuArgs& arg) :
        vv::Tuple(input), ct::Accu<DIM_TYPE, ACCU_TYPE>(arg)
    {
    }
    /**
     * @brief Get an input value
     * @param i, the index of input to get
     * @return the ith input value
     */
    inline double getInput(unsigned int i)
    {
        return (((vv::Tuple*) this)->at(i));
    }
    /**
     * @brief Sets an input value
     * @param i, the index of input to set
     * @param v, the ith value to set
     */
    inline void setInput(unsigned int i, double v)
    {
        (((vv::Tuple*) this)->at(i)) = v;
    }
};


/**
 * @brief Type of evaluation of a StratP2. This is used to
 * compare strats.
 */
enum StratEvaluatorType
{
    /**
     * @brief Evaluates a StratP2 by considering the mean statistic
     */
    STRAT_EVAL_MEAN,
    /**
     * @brief Evaluates a StratP2 by considering the quantile statistic
     */
    STRAT_EVAL_QUANTILE,
    /**
     * @brief Evaluates a StratP2 by considering the projection of the point
     * on an axis.
     */
    STRAT_EVAL_INPUT_VALUE
};

/**
 * @brief StratComparator declaration. Provides
 * a comparison operator in order to sort set of StratP2.
 */
template<class STRAT, StratEvaluatorType EVAL_TYPE = STRAT_EVAL_MEAN>
struct StratComparator;

/**
 * @brief StratComparator implementation for STRAT_EVAL_MEAN
 */
template<class STRAT>
struct StratComparator<STRAT, STRAT_EVAL_MEAN>
{   typedef boost::shared_ptr<STRAT> strat_type_ptr;
    /**
     * @brief Comparison operator
     */
    bool operator()(const strat_type_ptr& i, const strat_type_ptr& j)
    {
        return (i->mean() < j->mean());
    }
};

/**
 * @brief StratComparator implementation for STRAT_EVAL_QUANTILE
 */
template<class STRAT>
struct StratComparator<STRAT, STRAT_EVAL_QUANTILE>
{
    typedef boost::shared_ptr<STRAT> strat_type_ptr;
    /**
     * @brief StratComparator constructor
     * @param quantileOrder, quantile order used to compute the quantile value
     */
    StratComparator(double quantileOrder) :
        mquantileOrder(quantileOrder)
    {
    }
    /**
     * @brief Comparison operator
     */
    bool operator()(const strat_type_ptr& i, const strat_type_ptr& j)
    {
        return (i->quantile(mquantileOrder) < j->quantile(mquantileOrder));
    }
    /**
     * @param The quantile order
     */
    double mquantileOrder;
};

/**
 * @brief StratComparator implementation for STRAT_EVAL_INPUT_VALUE
 */
template<class STRAT>
struct StratComparator<STRAT, STRAT_EVAL_INPUT_VALUE>
{
    typedef boost::shared_ptr<STRAT> strat_type_ptr;
    /**
     * @brief StratComparator constructor
     * @param inputIndex, index of input vector considered for comparison
     */
    StratComparator(unsigned int inputIndex) :
        minputIndex(inputIndex)
    {
    }
    /**
     * @brief Comparison operator
     */
    bool operator()(const strat_type_ptr& i, const strat_type_ptr& j)
    {
        return (i->getInput(minputIndex) < j->getInput(minputIndex));
    }
    /**
     * @param The index of input vector considered for comparison
     */
    unsigned int minputIndex;
};

}
}
}//namespace cfo::fo_methods::p2

#endif //CFO_FO_METHODS_P2_ACCU_STRAT_HPP_
