/**
 * cfo/fo_methods/p2/leaf_cutting.hpp
 * @author The RECORD Development Team (INRA) and the VLE Development Team
 * See the AUTHORS or Authors.txt file
 */

/*
 * Copyright (C) 2009-2010 INRA
 *
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef CFO_FO_METHODS_P2_LEAF_CUTTING_H_
#define CFO_FO_METHODS_P2_LEAF_CUTTING_H_

#include <algorithm>
#include <cstdlib>
#include <vector>
#include <cmath>

#include <cfo/tools/accumulators/accu_multi.hpp>

#include <cfo/fo_methods/p2/accu_leaf.hpp>
#include <cfo/fo_methods/p2/accu_strat.hpp>
#include <cfo/fo_methods/p2/factory_accu.hpp>
#include <cfo/fo_methods/p2/utils.hpp>
#include <cfo/fo_methods/p2/leaf_cutting_alternatives.hpp>


namespace cfo {
namespace fo_methods {
namespace p2 {

namespace st = cfo::tools;

/**
 *@brief abstract tool that cuts a region (i.e. leaf) into two parts
 */
template<class LEAF, CutType CUT_TYPE>
class Cutting
{

public:

    typedef boost::shared_ptr<LEAF> leaf_type_ptr;
    typedef typename LEAF::strat_type_ptr strat_type_ptr;
    typedef typename LEAF::strat_type strat_type;

    /**
     *@brief Cutting constructor
     */
    Cutting(const P2args& p2args) :
        mbuilder(p2args), mmethod(p2args)
    {
    }
    /**
     *@brief Cutting destructor
     */
    virtual ~Cutting()
    {

    }
    /**
     *@brief abstract method that cuts down the region pointed into two parts
     *specify attributes of the child from its father
     *@param father points onto the father region to cut
     *@return a vector of the children leaves (partition of the father)
     */
    std::pair<leaf_type_ptr, leaf_type_ptr> cuts(const leaf_type_ptr& father) const
    {
        leaf_type_ptr child_out1 = mbuilder.buildLeaf(father->getDepth()+1);
        leaf_type_ptr child_out2 = mbuilder.buildLeaf(father->getDepth()+1);
        unsigned int varToCut = father->getGreatestExtent();
        double pivot = mmethod.getPivot(father, varToCut);
        //input values
        for (unsigned int i = 0; i < child_out1->getInputSize(); i++) {
            const ct::DCDomain& fparam = father->getDomain(i);
            ct::DCDomain& c1param = child_out1->getDomain(i);
            ct::DCDomain& c2param = child_out2->getDomain(i);
            c1param.setMin(fparam.min());
            c1param.setMax(fparam.max());
            c1param.setMinInterval(fparam.minInterval());
            c2param.setMin(fparam.min());
            c2param.setMax(fparam.max());
            c2param.setMinInterval(fparam.minInterval());
            if (i == varToCut) {
                c1param.setMax(pivot);
                c2param.setMin(pivot);
            }
        }
        //split points
        for (unsigned int s = 0; s < father->count(); s++) {
            const strat_type_ptr& strat = father->access(s);
            if (strat->getInput(varToCut) <=
                    child_out1->getDomain(varToCut).max()) {
                child_out1->insert(strat);
            } else {
                child_out2->insert(strat);
            }
        }
        return (std::pair<leaf_type_ptr, leaf_type_ptr>(child_out1, child_out2));
    }

private:
    /**
     * @brief util tool to build new LeafP2 objects resulting from the cutting.
     */
    AccuFactory<LEAF> mbuilder;
    /**
     * @brief P2 alternative tool used to select the fulctrum value for the
     * cutting.
     */
    CutImpl<LEAF, CUT_TYPE> mmethod;

};

}
}
}//namespace cfo::fo_methods::p2

#endif /* CFO_FO_METHODS_P2_LEAF_CUTTING_H_ */
