/**
 * cfo/fo_methods/p2/leaf_cutting_alternatives.hpp
 * @author The RECORD Development Team (INRA) and the VLE Development Team
 * See the AUTHORS or Authors.txt file
 */

/*
 * Copyright (C) 2009-2010 INRA
 *
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef CFO_FO_METHODS_P2_LEAF_CUTTING_ALTERNATIVES_H_
#define CFO_FO_METHODS_P2_LEAF_CUTTING_ALTERNATIVES_H_

#include <algorithm>
#include <cstdlib>
#include <vector>
#include <cmath>

#include <cfo/fo_methods/p2/accu_leaf.hpp>
#include <cfo/fo_methods/p2/accu_strat.hpp>
#include <cfo/fo_methods/p2/factory_accu.hpp>
#include <cfo/fo_methods/p2/utils.hpp>

namespace cfo {
namespace fo_methods {
namespace p2 {


namespace st = cfo::tools;

/**
 * @brief CutImpl declaration.
 * Given a leaf and the index of the parameter to cut,
 * A CutImpl instantiation should provide the function getPivot that gives the
 * real between the min value and the max value of the extent of the parameter
 * that is used as fulcrum for cutting the leaf in two sub leaves.
 */
template<class LEAF, CutType CUT_TYPE>
class CutImpl;

/**
 * @brief CutImpl implementation of alternative CUT_ON_MIDDLE
 */
template<class LEAF>
class CutImpl<LEAF, CUT_ON_MIDDLE>
{
public:
    typedef boost::shared_ptr<LEAF> leaf_type_ptr;

    /**
     * @brief CutImpl constructor
     * @param p2args,  p2args, P2 initializations
     */
    CutImpl(const P2args& /*p2args*/)
    {
    }
    /**
     * @brief Gets the fulcrum value selected for cutting the leaf
     * in two sub leaves.
     * @param leaf, the leaf to cut
     */
    double getPivot(const leaf_type_ptr& leaf, unsigned int varToCut) const
    {
        const ct::DCDomain& fparam = leaf->getDomain(varToCut);
        return (fparam.min() + (fparam.max() - fparam.min()) / 2.0);
    }
};

/**
 * @brief CutImpl implementation of alternative CUT_ON_PIVOT
 */
template<class LEAF>
class CutImpl<LEAF, CUT_ON_PIVOT>
{

public:
    typedef boost::shared_ptr<LEAF> leaf_type_ptr;
    typedef typename LEAF::strat_type_ptr strat_type_ptr;
    typedef typename LEAF::strat_type strat_type;

    /**
     * @brief CutImpl constructor
     * @param p2args,  p2args, P2 initializations
     */
    CutImpl(const P2args& p2args) :
        mleafQtlOrder(p2args.cut_leaf_quantile_order),
        mstratQtlOrder(p2args.cut_strat_quantile_order),
        mstratCompare(p2args.cut_strat_quantile_order)
    {
    }
    /**
     * @brief Gets the fulcrum value selected for cutting the leaf
     * in two sub leaves.
     * @param leaf, the leaf to cut
     */

    double getPivot(const leaf_type_ptr& leaf, unsigned int varToCut) const
    {
        std::vector<strat_type_ptr> ordered(leaf->getStrats());
        std::sort(ordered.begin(), ordered.end(), mstratCompare);
        double rang_quantile = (ordered.size() - 1) * mleafQtlOrder;
        int ent_rang = floor(rang_quantile);
        strat_type_ptr pivot = ordered[ent_rang];
        return pivot->getInput(varToCut);
    }
private:
    /**
     * @brief the quantile order for evalauting the quantile at leaf scale
     */
    double mleafQtlOrder;
    /**
     * @brief the quantile order for evalauting the quantile at strat scale
     */
    double mstratQtlOrder;
    /**
     * @brief Tool to order the strat accumulated into the leaf.
     */
    StratComparator<strat_type, STRAT_EVAL_QUANTILE> mstratCompare;

};
/**
 * @brief CutImpl implementation of alternative CUT_ON_LARGEST_MEAN_DIFF
 */
template<class LEAF>
class CutImpl<LEAF, CUT_ON_LARGEST_MEAN_DIFF>
{
public:
    typedef boost::shared_ptr<LEAF> leaf_type_ptr;
    typedef typename LEAF::dim_type dim_type;
    typedef typename LEAF::strat_type strat_type;
    typedef typename LEAF::strat_type_ptr strat_type_ptr;

    /**
     * @brief CutImpl constructor
     * @param p2args,  p2args, P2 initializations
     */
    CutImpl(const P2args& /*p2args*/)
    {
    }
    /**
     * @brief Gets the fulcrum value selected for cutting the leaf
     * in two sub leaves.
     * @param leaf, the leaf to cut
     */
    double getPivot(const leaf_type_ptr& leaf, unsigned int varToCut) const
    {
        //TODO check if more than one strat
        StratComparator<strat_type, STRAT_EVAL_INPUT_VALUE> stratCompare(varToCut);
        std::vector<strat_type_ptr> ordered(leaf->getStrats());
        std::sort(ordered.begin(), ordered.end(), stratCompare);
        double max_diff = 0;
        double pivot_max_diff = 0;
        double current_diff = 0;
        ct::Accu<dim_type, ct::MEAN_VAR> accuLeft;
        for (unsigned int pivot = 0; pivot < ordered.size() - 1; pivot++) {
            accuLeft.insert(ordered[pivot]->mean());
            ct::Accu<dim_type, ct::MEAN_VAR> accuRight;
            for (unsigned int s = pivot + 1; s < ordered.size(); s++) {
                accuRight.insert(ordered[s]->mean());
            }
            current_diff = abs(accuLeft.mean() - accuRight.mean());
            if (current_diff > max_diff) {
                max_diff = current_diff;
                pivot_max_diff = pivot;
            }
        }
        return (ordered[pivot_max_diff]->getInput(varToCut) +
                ordered[pivot_max_diff + 1]->getInput(varToCut)) / 2;
    }
};

/**
 * @brief CutImpl implementation of alternative CUT_ON_LARGEST_QTL_DIFF
 */
template<class LEAF>
class CutImpl<LEAF, CUT_ON_LARGEST_QTL_DIFF>
{
public:
    typedef boost::shared_ptr<LEAF> leaf_type_ptr;
    typedef typename LEAF::dim_type dim_type;
    typedef typename LEAF::strat_type strat_type;
    typedef typename LEAF::strat_type_ptr strat_type_ptr;

    /**
     * @brief CutImpl constructor
     * @param p2args,  p2args, P2 initializations
     */
    CutImpl(const P2args& p2args) :
        mleafQtlOrder(p2args.cut_leaf_quantile_order),
        mstratQtlOrder(p2args.cut_strat_quantile_order)
    {
    }
    /**
     * @brief Gets the fulcrum value selected for cutting the leaf
     * in two sub leaves.
     * @param leaf, the leaf to cut
     */
    double getPivot(const leaf_type_ptr& leaf, unsigned int varToCut) const
    {
        //TODO check if more than one strat
        StratComparator<strat_type, STRAT_EVAL_INPUT_VALUE> stratCompare(varToCut);
        std::vector<strat_type_ptr> ordered(leaf->getStrats());
        std::sort(ordered.begin(), ordered.end(), stratCompare);
        double max_diff = 0;
        double pivot_max_diff = 0;
        double current_diff = 0;
        ct::Accu<dim_type, ct::QUANTILE> accuLeft;
        for (unsigned int pivot = 0; pivot < ordered.size() - 1; pivot++) {
            accuLeft.insert(ordered[pivot]->quantile(mstratQtlOrder));
            ct::Accu<dim_type, ct::QUANTILE> accuRight;
            for (unsigned int s = pivot + 1; s < ordered.size(); s++) {
                accuRight.insert(ordered[s]->quantile(mstratQtlOrder));
            }
            current_diff = abs(accuLeft.quantile(mleafQtlOrder) -
                accuRight.quantile(mleafQtlOrder));
            if (current_diff > max_diff) {
                max_diff = current_diff;
                pivot_max_diff = pivot;
            }
        }
        return (ordered[pivot_max_diff]->getInput(varToCut) +
                ordered[pivot_max_diff + 1]->getInput(varToCut)) / 2;
    }
private:
    /**
     * @brief the quantile order for evalauting the quantile at leaf scale
     */
    double mleafQtlOrder;
    /**
     * @brief the quantile order for evalauting the quantile at strat scale
     */
    double mstratQtlOrder;
};

}
}
}//namespace cfo::fo_methods::p2

#endif //CFO_FO_METHODS_P2_LEAF_CUTTING_ALTERNATIVES_H_
