/**
 * cfo/fo_methods/p2/leaf_simulation.hpp
 * @author The RECORD Development Team (INRA) and the VLE Development Team
 * See the AUTHORS or Authors.txt file
 */

/*
 * Copyright (C) 2009-2010 INRA
 *
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef CFO_MEHODS_P2_LEAF_SIMULATION_H_
#define CFO_MEHODS_P2_LEAF_SIMULATION_H_

#include <vector>

#include <vle/value/Double.hpp>
#include <vle/value/Set.hpp>
#include <vle/utils/Exception.hpp>
#include <vle/utils/Rand.hpp>

#include <cfo/fo_methods/p2/accu_leaf.hpp>
#include <cfo/tools/accumulators/accu_multi.hpp>

namespace vv = vle::value;
namespace vu = vle::utils;

namespace cfo {
namespace fo_methods {
namespace p2 {

namespace ct = cfo::tools;


template <class DIM_TYPE> struct OuptutScaleFit;

template<>
struct OuptutScaleFit<ct::MonoDim>
{
    const  P2args& mp2args;
    OuptutScaleFit(const P2args& p2args):
        mp2args(p2args)
    {
    }

    ct::MonoDim translate(const vv::Value& v){
       if(v.isSet()&&(v.toSet().size() >= 1)){
           return v.toSet().getDouble(0) * mp2args.sim_output_scales[0];
       } else if(v.isDouble()){
           return v.toDouble().value() * mp2args.sim_output_scales[0];
       }
       throw vu::InternalError(
           vle::fmt("P2 simulation output type recovering : %1% ")
           % v.getType());
   }


};

template<>
struct OuptutScaleFit<ct::MultiDim>
{
    const  P2args& mp2args;
    OuptutScaleFit(const P2args& p2args):
        mp2args(p2args)
    {
    }
   ct::MultiDim translate(const vv::Value& v){
       if(v.isSet()){
           ct::MultiDim res(v.toSet().size());
           for(unsigned int i = 0 ; i < v.toSet().size() ; i++){
               res[i] = v.toSet().getDouble(i) * mp2args.sim_output_scales[i];
           }
           return res;
       }
       throw vu::InternalError(" ");
   }

};


/**
 *  @brief Object to simulate points of a Leaf
 *  in accordance with a density and a distribution
 */
template<class LEAF>
class LeafSimulation
{
    typedef boost::shared_ptr<LEAF> leaf_type_ptr;
    typedef typename LEAF::strat_type_ptr strat_type_ptr;
    typedef typename LEAF::strat_type strat_type;
    typedef typename strat_type::dim_type dim_type;

public:
    /**
     *@brief LeafSimulation constructor
     *@param sim, reference to the simulator
     *@param p2args, set of parameters of p2
     */
    LeafSimulation(ct::VleSimulator& sim, const P2args& p2args, vu::Rand& rand) :
        msim(sim), mrand(rand), maccuFactory(p2args), minputs(0),
        mdim(p2args.leaf_params.size()),
        mnbOfReplicas(p2args.sim_nb_replicas), mdistribution(p2args.sim_distribution),
        mdensity(p2args.sim_density), mcommonSeed(p2args.sim_common_seed),
        moutputScaleFit(p2args)
    {
    }

    /**
     *@brief LeafSimulation destructor
     */
    virtual ~LeafSimulation()
    {
    }
    /**
     *@brief strategies generation and simulation
     *@param leaf the leaf to simulate
     *@return the number of simulations (number of strategies * number of replicas)
     *
     */
    int simulates(leaf_type_ptr& leaf)
    {
        int nbstrat = density(leaf);
        minputs.value().resize(nbstrat);
        assignRandomStrategies(leaf);
        const vv::Matrix& simRes = msim.simulatePlan(minputs,
            mnbOfReplicas, mcommonSeed);
        for (unsigned int comb = 0; comb < simRes.columns(); comb++) {
            strat_type_ptr strat = maccuFactory.buildStrat(
                minputs.get(comb)->toTuple());
            for (unsigned int repl = 0; repl < simRes.rows(); repl++) {
                strat->insert(
                        moutputScaleFit.translate(*simRes.get(comb, repl)));
            }
            leaf->insert(strat);
        }
        return nbstrat * mnbOfReplicas;
    }

private:
    /**
     *@brief one strategy generation
     *@param leaf into which the strategy is generated
     *@param the generated strategy input values
     */
    void assignRandomStrategies(leaf_type_ptr& leaf)
    {
        int nb_params = leaf->getInputSize();
        double r;
        for (unsigned int i = 0; i < minputs.size(); i++) {
            vv::Tuple* input = new vv::Tuple(mdim);
            for (int v = 0; v < nb_params; v++) {
                r = -1;
                switch (mdistribution) {
                case NORMAL: {
                    while (r < 0 || r > 1) {
                        r = mrand.normal(0.5, 0.25);
                    }
                    break;
                }
                case EXPONENTIAL: {
                    while (r < 0 || r > 1) {
                        r = mrand.exponential(0.5);
                    }
                    break;
                } case UNIFORM:{
                    r = (double) mrand.getDouble(0, 1);
                    break;
                }

                }
                ct::DCDomain& param = leaf->getDomain(v);
                input->at(v) = param.min() + r * (param.max() - param.min());
            }
            minputs.set(i, input);
        }
    }
    /**
     *@brief gives the current density of simulated points in the leaf (coming from earlier simulations)
     *@param leaf which density must be evaluated
     *@return the current density of simulated point in the Leaf
     *
     */
    int density(const leaf_type_ptr& leaf) const
    {
        unsigned int i;
        double nb;
        int nbStrategies = 1;
        for (i = 0; i < leaf->getInputSize(); i++) {
            ct::DCDomain& param = leaf->getDomain(i);
            nb = ceil(mdensity * (param.max() - param.min())
                                 / (double) param.minInterval());
            if ((int) nb > nbStrategies) {
                nbStrategies = (int) nb;
            }
        }
        return nbStrategies;
    }
    /**
     * @brief the simulation factory that manages the vle simulator
     */
    ct::VleSimulator& msim;

    vu::Rand& mrand;

    AccuFactory<LEAF> maccuFactory;

    vv::Set minputs;
    /**
     * @brief simension of the optimization problem
     */
    unsigned int mdim;
    /**
     *@brief the number of simulations for evaluating a strategy
     */
    int mnbOfReplicas;
    /**
     *@brief the type of distribution used to generate strategies
     */
    LeafSimDistribution mdistribution;
    /**
     *@brief the required  density of strategies in the leaf
     */
    double mdensity;
    /**
     *@brief the common-seed option
     */
    bool mcommonSeed;
    /**
     *@brief util object used to fit the simulation output value to the
     * dimension type object handled by accumulators (MonoDim or MultiDim).
     */
    OuptutScaleFit<dim_type> moutputScaleFit;
    /**
     * @brief
     */

};

}
}
}//namespace cfo::fo_methods::p2

#endif //CFO_MEHODS_P2_LEAF_SIMULATION_H_
