/**
 * cfo/fo_methods/p2/p2.hpp
 * @author The RECORD Development Team (INRA) and the VLE Development Team
 * See the AUTHORS or Authors.txt file
 */

/*
 * Copyright (C) 2009-2010 INRA
 *
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef CFO_FO_METHODS_P2_HPP_
#define CFO_FO_METHODS_P2_HPP_


#include <boost/format.hpp>

#include <vle/value/Map.hpp>
#include <vle/value/Set.hpp>
#include <vle/value/Double.hpp>
#include <vle/utils/Rand.hpp>

#include <cfo/interfaces/FOmethod.hpp>
#include <cfo/fo_methods/p2/P2impl.hpp>
#include <cfo/fo_methods/p2/utils.hpp>


namespace vv = vle::value;
namespace vu = vle::utils;

namespace cfo {
namespace fo_methods {
namespace p2 {

namespace ct = cfo::tools;
namespace ci = cfo::interfaces;

/**
 * @brief P2 implementation of FOmethod
 */
class p2: public ci::FOmethod
{

public:
    /**
     * @brief p2 constructor
     */
    p2(const vv::Map& init, ct::VleSimulator& sim) :
        ci::FOmethod(init,sim), malternatives(),
        margs(), mresults(), mrand()
    {

        //initialize inputs
        const vv::Map& model = init.getMap("model");
        const vv::Map& parameters = init.getMap("parameters");
        const vv::Set& inputs = model.getSet("inputs");
        const vv::Set& outputs = model.getSet("outputs");
        const vv::Map& domains = parameters.getMap("domains");


        for (unsigned int p = 0; p < inputs.size(); p++) {
            const vv::Map& dom = domains.getMap(inputs.getString(p));
            margs.leaf_params.push_back(ct::DCDomain(dom));
        }
        //initialize outputs
        margs.sim_output_size = outputs.size();
        margs.sim_output_scales.resize(margs.sim_output_size);
        if (parameters.exist("scales")) {
			const vv::Map& scales = parameters.getMap("scales");
			for (unsigned int p = 0; p < margs.sim_output_size; p++) {
				if (scales.exist(outputs.getString(p))) {
					double scale = scales.getDouble(outputs.getString(p));
					margs.sim_output_scales[p] = scale;
				} else {
					margs.sim_output_scales[p] = 1;
				}
			}
		}
        if (parameters.exist("cut-type")) {
            const std::string& v = parameters.getString("cut-type");
            if (v == "cut-on-middle") {
                malternatives.cut_type = CUT_ON_MIDDLE;
            } else if (v == "cut-on-pivot") {
                malternatives.cut_type = CUT_ON_PIVOT;
            } else if (v == "cut-on-largest-mean-diff") {
                malternatives.cut_type = CUT_ON_LARGEST_MEAN_DIFF;
            } else if (v == "cut-on-largest-qtl-diff") {
                malternatives.cut_type = CUT_ON_LARGEST_QTL_DIFF;
            } else {
                throw vu::InternalError(vle::fmt(
                    "P2 initialisation, cut-type not recognised : %1%)")
                    % v);
            }
        }
        if (parameters.exist("eval-perf-type")) {
            const std::string& v = parameters.getString("eval-perf-type");
            if (v == "leaf-perf-min") {
                malternatives.eval_perf_type = LEAF_PERF_MIN;
            } else if (v == "leaf-perf-mean") {
                malternatives.eval_perf_type = LEAF_PERF_MEAN;
            } else if (v == "leaf-perf-quantile") {
                malternatives.eval_perf_type = LEAF_PERF_QUANTILE;
            } else {
                throw vu::InternalError(vle::fmt(
                    "P2 initialisation, eval-perf-type not recognised : %1%")
                    %v);
            }
        }
        if (parameters.exist("eval-disp-type")) {
            const std::string& v = parameters.getString("eval-disp-type");
            if (v == "leaf-disp-global-var") {
                malternatives.eval_disp_type = LEAF_DISP_GLOBAL_VAR;
            } else if (v == "leaf-disp-spatial-var1") {
                malternatives.eval_disp_type = LEAF_DISP_SPATIAL_VAR1;
            } else if (v == "leaf-disp-spatial-var2") {
                malternatives.eval_disp_type = LEAF_DISP_SPATIAL_VAR2;
            } else {
                throw vu::InternalError(vle::fmt(
                    "P2 initialisation, eval-disp-type not recognised : %1%")
                    %v);
            }
        }
        if (parameters.exist("selection-type")) {
            const std::string& v = parameters.getString("selection-type");
            if (v == "leaf-sel-beta") {
                malternatives.selection_type = LEAF_SEL_BETA;
            } else if (v == "leaf-sel-conc-beta") {
                malternatives.selection_type = LEAF_SEL_CONC_BETA;
            } else if (v == "leaf-sel-performance") {
                malternatives.selection_type = LEAF_SEL_PERFORMANCE;
            } else {
                throw vu::InternalError(vle::fmt(
                    "P2 initialisation : selection-type not recognised : %1%")
                    %v);
            }
        }
        if (parameters.exist("leaf-quantile-order")) {
            double v = parameters.getDouble("leaf-quantile-order");
            margs.cut_leaf_quantile_order = v;
            margs.eval_leaf_quantile_order = v;
        }
        if (parameters.exist("strat-quantile-order")) {
            double v = parameters.getDouble("strat-quantile-order");
            margs.cut_strat_quantile_order = v;
            margs.eval_strat_quantile_order = v;
        }
        if (parameters.exist("eval-target")) {
            double v = parameters.getDouble("eval-target");
            margs.eval_target = v;
        }
        if (parameters.exist("sim-output-size")) {
            int v = parameters.getInt("sim-output-size");
            margs.eval_target = v;
        }
        if (parameters.exist("sim-nb-replicas")) {
            int v = parameters.getInt("sim-nb-replicas");
            margs.sim_nb_replicas = v;
        }
        if (parameters.exist("sim-distribution")) {
            const std::string& v = parameters.getString("sim-distribution");
            if (v == "uniform") {
                margs.sim_distribution = UNIFORM;
            } else if (v == "exponential") {
                margs.sim_distribution = EXPONENTIAL;
            } else if (v == "normal") {
                margs.sim_distribution = NORMAL;
            } else {
                throw vu::InternalError(vle::fmt(
                    "P2 initialisation, distribution-type not recognised : %1")
                    %v);
            }
        }
        if (parameters.exist("sim-density")) {
            double v = parameters.getDouble("sim-density");
            margs.sim_density = v;
        }
        if (parameters.exist("sim-common-seed")) {
            bool v = parameters.getBoolean ("sim-common-seed");
            margs.sim_common_seed = v;
        }
        if (parameters.exist("pool-max-capacity")) {
            int v = parameters.getInt("pool-max-capacity");
            margs.pool_max_capacity = v;
        }
        if (parameters.exist("max-sim-number")) {
            int v = parameters.getInt("max-sim-number");
            margs.max_sim_number = v;
        }
        if (parameters.exist("seed")) {
            mrand.seed(parameters.getInt("seed"));
        }
    }

    /**
     * @brief p2 constructor
     * @param alternatives, specification of p2 alternatives
     * @param args, specification of p2 args
     * @param sim, reference to the simulator
     */
    p2(const P2alternatives& alternatives, const P2args& args,
        ct::VleSimulator& sim) :
            ci::FOmethod(vv::Map(),sim), malternatives(alternatives),
            margs(args), mresults(), mrand()
    {
    }
    /**
     * @brief p2 destructor
     */
    ~p2()
    {
    }

    /**
     * @brief Function to implement by an optimization method.
     * Not that the overall optimization process can be implemented into
     * one step (e.g. if the concept of step has no meaning for the method).
     * @param current_step the current step
     * @return true TODO split into steps
     */
    virtual bool optimizeStep(unsigned int /*current_step*/)
    {
        if (margs.sim_output_size != 1) {
            throw vu::InternalError(" P2 initialisation : "
                    "multi objective not handled yet ");
        }
        if((malternatives.eval_disp_type == LEAF_DISP_SPATIAL_VAR2)&&
                (!margs.sim_common_seed)){
            throw vu::InternalError(" P2 initialisation : using dispersion "
                    " evaluation by spatial variance 2 requires common seed to "
                    " be true ");
        }
        buildP2run<ct::MonoDim> ();
        return true;
    }
    /**
     * @brief Observation method
     * @param port, the name of the port to observe
     * @return the observed value
     */
    virtual vv::Value* observe(const std::string& port) const
    {
    	if(mresults.exist(port)){
    		return mresults.get(port)->clone();
    	}
    	return 0;
	}

private :
    /**
     * @brief incremental building of P2 alternatives implementation
     */
    template<class DIM_TYPE> void buildP2run()
    {
        switch (malternatives.cut_type) {
        case CUT_ON_MIDDLE: {
            buildP2run1<DIM_TYPE, ct::MEAN_VAR, CUT_ON_MIDDLE> ();
            break;
        }
        case CUT_ON_PIVOT: {
            buildP2run1<DIM_TYPE, ct::QUANTILE, CUT_ON_PIVOT> ();
            break;
        }
        case CUT_ON_LARGEST_MEAN_DIFF: {
            buildP2run1<DIM_TYPE, ct::MEAN_VAR, CUT_ON_LARGEST_MEAN_DIFF> ();
            break;
        }
        case CUT_ON_LARGEST_QTL_DIFF: {
            buildP2run1<DIM_TYPE, ct::QUANTILE, CUT_ON_LARGEST_QTL_DIFF> ();
            break;
        }
        }
    }

    template<class DIM_TYPE, ct::AccuType ACCU_TYPE, CutType CUT_TYPE>
    void buildP2run1()
    {
        switch (malternatives.eval_perf_type) {
        case LEAF_PERF_MEAN: {
            buildP2run2<DIM_TYPE, ACCU_TYPE, CUT_TYPE, LEAF_PERF_MEAN> ();
            break;
        }
        case LEAF_PERF_QUANTILE: {
            buildP2run2<DIM_TYPE, ct::QUANTILE, CUT_TYPE, LEAF_PERF_QUANTILE> ();
            break;
        }
        case LEAF_PERF_MIN: {
            buildP2run2<DIM_TYPE, ACCU_TYPE, CUT_TYPE, LEAF_PERF_MIN> ();
            break;
        }
        }
    }

    template<class DIM_TYPE, ct::AccuType ACCU_TYPE, CutType CUT_TYPE,
        LeafPerformanceType LEAF_PERF>
    void buildP2run2()
    {
        switch (malternatives.eval_disp_type) {
        case LEAF_DISP_GLOBAL_VAR: {
            buildP2run3<DIM_TYPE, ACCU_TYPE, CUT_TYPE, LEAF_PERF, LEAF_DISP_GLOBAL_VAR> ();
            break;
        }
        case LEAF_DISP_SPATIAL_VAR1: {
            buildP2run3<DIM_TYPE, ACCU_TYPE, CUT_TYPE, LEAF_PERF, LEAF_DISP_SPATIAL_VAR1> ();
            break;
        }
        case LEAF_DISP_SPATIAL_VAR2: {
            buildP2run3<DIM_TYPE, ct::STORE, CUT_TYPE, LEAF_PERF, LEAF_DISP_SPATIAL_VAR2> ();
            break;
        }
        }

    }

    template<class DIM_TYPE, ct::AccuType ACCU_TYPE, CutType CUT_TYPE,
        LeafPerformanceType LEAF_PERF, LeafDispersionType LEAF_DISP>
    void buildP2run3()
    {
        switch (malternatives.selection_type) {
        case LEAF_SEL_PERFORMANCE: {
            buildP2run4<DIM_TYPE, ACCU_TYPE, CUT_TYPE, LEAF_PERF, LEAF_DISP, LEAF_SEL_PERFORMANCE> ();
            break;
        }
        case LEAF_SEL_BETA: {
            buildP2run4<DIM_TYPE, ACCU_TYPE, CUT_TYPE, LEAF_PERF, LEAF_DISP, LEAF_SEL_BETA> ();
            break;
        }
        case LEAF_SEL_CONC_BETA: {
            buildP2run4<DIM_TYPE, ACCU_TYPE, CUT_TYPE, LEAF_PERF, LEAF_DISP, LEAF_SEL_CONC_BETA> ();
            break;
        }
        }

    }

    template<class DIM_TYPE, ct::AccuType ACCU_TYPE, CutType CUT_TYPE,
        LeafPerformanceType LEAF_PERF, LeafDispersionType LEAF_DISP,
        LeafSelectionType LEAF_SEL>
    void buildP2run4()
    {
        P2impl<DIM_TYPE, ACCU_TYPE, CUT_TYPE, LEAF_PERF, LEAF_DISP, LEAF_SEL> p2impl(simulator(),margs, mrand);
        p2impl.optimises(mresults);
    }

    /**
     * @brief P2 alternatives
     */
    P2alternatives malternatives;
    /**
     * @brief P2 parameters
     */
    P2args margs;
    /**
     * @brief Results of optimization process
     */
    vv::Map mresults;
    /**
     * @brief Random number generator
     */
    vu::Rand mrand;
};

}
}
}//namespaces

#endif //CFO_FO_METHODS_P2_HPP_
