/**
 * cfo/fo_methods/p2/utils.hpp
 * @author The RECORD Development Team (INRA) and the VLE Development Team
 * See the AUTHORS or Authors.txt file
 */

/*
 * Copyright (C) 2009-2010 INRA
 *
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef CFO_FO_METHODS_P2_UTILS_HPP_
#define CFO_FO_METHODS_P2_UTILS_HPP_

//tools include for p2
#include <cfo/tools/accumulators/accu_multi.hpp>
#include <cfo/tools/domains/domains.hpp>

namespace cfo {
namespace fo_methods {
namespace p2 {

namespace ct = cfo::tools;

/**
 * @brief A CutType element determines the strategy used
 * for cutting a leaf in two sub leaves. Cf. thesis O. Crespo for more
 * details.
 */
enum CutType
{
    /**
     * @brief Cuts the leaf in the middle of the selected parameter
     */
    CUT_ON_MIDDLE,
    /**
     * @brief Cuts according quantile values of leaf (requires QUANTILE
     * accumulator for StratP2). Default alternative.
     */
    CUT_ON_PIVOT,
    /**
     * @brief Cuts a leaf according the most important mean performance
     * difference between the two sub leaves.
     */
    CUT_ON_LARGEST_MEAN_DIFF,
    /**
     * @brief Cuts a leaf according the most important quantile performance
     * difference between the two sub leaves. (requires QUANTILE
     * accumulator for StratP2)
     */
    CUT_ON_LARGEST_QTL_DIFF
};


/**
 * @brief A LeafPerformanceType element determines the strategy used
 * for evaluating the performance of a leaf.
 */
enum LeafPerformanceType
{
    /**
     *@brief Evaluates the performance of a leaf by the mean value.
     *  Default alternative.
     */
    LEAF_PERF_MEAN,
    /**
     *@brief Evaluates the performance of a leaf by a quantile value.
     * (requires QUANTILE accumulator for StratP2)
     */
    LEAF_PERF_QUANTILE,
    /**
     *@brief Evaluates the performance of a leaf by the minimal value
     * quantile value.
     */
    LEAF_PERF_MIN
};
/**
 * @brief A LeafDispersionType element determines the strategy used
 * for evaluating the dispersion of a leaf.
 */
enum LeafDispersionType
{
    /**
     * @brief Evaluates the dispersion buy the global variance.
     * Default alternative.
     */
    LEAF_DISP_GLOBAL_VAR,
    /**
     * @brief Evaluates the dispersion by the spatial variance (version 1).
     */
    LEAF_DISP_SPATIAL_VAR1,
    /**
     * @brief Evaluates the dispersion by the spatial variance (version 2).
     *  (requires STORE accumulator for StratP2)
     */
    LEAF_DISP_SPATIAL_VAR2
};
/**
 * @brief A LeafSelectionType element determines the strategy used
 * for selecting the best leaf to cut.
 */
enum LeafSelectionType
{
    /**
     * @brief Selects the leaf with the minimal performance
     */
    LEAF_SEL_PERFORMANCE,
    /**
     * @brief Use the beta selection
     */
    LEAF_SEL_BETA,
    /**
     * @brief Use the conc-beta selection
     */
    LEAF_SEL_CONC_BETA
};
/**
 * @brief Different types of distributions for
 * simulating strats in a leaf.
 */
enum LeafSimDistribution
{
    NORMAL, EXPONENTIAL, UNIFORM
};

/**
 * @brief Global structure that refers to a P2 implementation, ie.
 * a composition of alternatives.
 */
struct P2alternatives
{
    /**
     * @brief P2alternatives default constructor (with the best
     * set of alternatives according O. Crespo)
     *
     */
    P2alternatives() :
        cut_type(CUT_ON_PIVOT), eval_perf_type(LEAF_PERF_MEAN),
        eval_disp_type(LEAF_DISP_GLOBAL_VAR),
        selection_type(LEAF_SEL_PERFORMANCE)
    {
    }
    /**
     * @brief P2alternatives copy constructor.
     */
    P2alternatives(const P2alternatives& alt) :
        cut_type(alt.cut_type), eval_perf_type(alt.eval_perf_type),
        eval_disp_type(alt.eval_disp_type), selection_type(alt.selection_type)
    {
    }
    /**
     * @brief Chosen cutting alternative
     */
    CutType cut_type;//4
    /**
     * @brief Chosen leaf performance evaluation
     */
    LeafPerformanceType eval_perf_type;//3
    /**
     * @brief Chosen leaf dispersion evaluation
     */
    LeafDispersionType eval_disp_type;//3
    /**
     * @brief Chosen leaf selection for cutting.
     */
    LeafSelectionType selection_type;//3
};

/**
 * @brief Global structure that stores P2 parameters. Depending on alternative
 * choices, some parameters might not be used.
 */
struct P2args
{

    const static double MAX_DOUBLE = 2000000;
    const static double MIN_DOUBLE = -2000000;

    /**
     * @brief P2args default constructor
     */
    P2args() :
        leaf_params(0), cut_leaf_quantile_order(0.8),
        cut_strat_quantile_order(0.2),
        eval_leaf_quantile_order(0.5),
        eval_strat_quantile_order(0.5),
        eval_target(MIN_DOUBLE),
        sim_output_size(1), sim_output_scales (sim_output_size, 1),
        sim_nb_replicas(1),
        sim_distribution(NORMAL), sim_density(0.1),
        sim_common_seed(false), pool_max_capacity(10),
        max_sim_number(1000)
    {
    }
    /**
     * @brief P2args copy constructor
     */
    P2args(const P2args& args) :
        leaf_params(args.leaf_params),
        cut_leaf_quantile_order(args.cut_leaf_quantile_order),
        cut_strat_quantile_order(args.cut_strat_quantile_order),
        eval_leaf_quantile_order(args.eval_leaf_quantile_order),
        eval_strat_quantile_order(args.eval_strat_quantile_order),
        eval_target(args.eval_target),
        sim_output_size(args.sim_output_size),
        sim_output_scales (args.sim_output_scales),
        sim_nb_replicas(args.sim_nb_replicas),
        sim_distribution(args.sim_distribution),
        sim_density(args.sim_density), sim_common_seed(args.sim_common_seed),
        pool_max_capacity(args.pool_max_capacity),
        max_sim_number(args.max_sim_number)

    {
    }
    /**
     * @brief The set of parameters (and their extents) that defines the
     * search domain.
     */
    std::vector<ct::DCDomain> leaf_params;
    /**
     * @brief Quantile order used for cutting a leaf.
     */
    double cut_leaf_quantile_order;
    /**
     * @brief Quantile order used for cutting a leaf (more speciffically for
     * evaluating a StratP2 while cutting)
     */
    double cut_strat_quantile_order;
    /**
     * @brief Quantile order used for evaluating a leaf.
     */
    double eval_leaf_quantile_order;
    /**
     * @brief Quantile order used for evaluating a strat.
     */
    double eval_strat_quantile_order;
    /**
     * @brief The minimal real value to reach by optimisation.
     *  Optimisation is then stopped.
     */
    double eval_target;
    /**
     * @brief Number of dimensions for output values.
     */
    unsigned int sim_output_size;
    /**
     * @brief Scales for simulation outputs
     * (eg. -1 for maximisation)
     */
    std::vector<double> sim_output_scales;
    /**
     * @brief Number of replicas to simulate for a StratP2.
     */
    unsigned int sim_nb_replicas;
    /**
     * @brief Distribution used for generating strats in a leaf.
     */
    LeafSimDistribution sim_distribution;
    /**
     * @brief Density of a leaf that determines the number of
     * points to generate.
     */
    double sim_density;
    /**
     * @brief True if the same alea is used for two replicas of
     * different strats.
     */
    bool sim_common_seed;
    /**
     * @brief The maximal number of leaves in the pool.
     */
    unsigned int pool_max_capacity;
    /**
     * @brief The maximal number of simulations.
     */
    unsigned int max_sim_number;
};

}
}
}

#endif /* CFO_FO_METHODS_P2_UTILS_HPP_ */
