/**
 * cfo/fo_methods/paradiseo/paradiseo.hpp
 * @author The RECORD Development Team (INRA) and the VLE Development Team
 * See the AUTHORS or Authors.txt file
 */

/*
 * Copyright (C) 2009-2010 INRA
 *
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef CFO_FO_METHODS_PARADISEO_PARADISEO_HPP_
#define CFO_FO_METHODS_PARADISEO_PARADISEO_HPP_

#include <iostream>

#include <vle/value/Map.hpp>
#include <vle/utils/Rand.hpp>

#include <cfo/interfaces/FOmethod.hpp>

#include <eo>
#include <es.h>

namespace cfo {
namespace fo_methods {
namespace paradiseo {

namespace ct = cfo::tools;
namespace ci = cfo::interfaces;
namespace vv = vle::value;
namespace vu = vle::utils;


// REPRESENTATION
//-----------------------------------------------------------------------------
// define your individuals
//typedef eoReal<double> Indi;
typedef eoEsSimple<eoMinimizingFitness> Indi;




// EVAL
//-----------------------------------------------------------------------------
// a simple fitness function that computes the euclidian norm of a real vector
//    @param _indi A real-valued individual

class VlePopEvaluator : public eoPopEvalFunc<Indi>
{
public:
  /** Ctor: set value of embedded eoEvalFunc */
  VlePopEvaluator(unsigned int dimension,
  ct::VleSimulator& simulator, vu::Rand& r) :
              dim(dimension), sim(simulator), rng(r)
  {

  }

  /** Do the job: simple loop over the offspring */
  void operator()(eoPop<Indi>& /*parents*/, eoPop<Indi>& offspring)
  {
      ct::InputVector inputs;
      inputs.resize(offspring.size());
      for (unsigned int s = 0; s < offspring.size(); s++) {
          vv::Set* input = new vv::Set(dim);
          Indi& ind = offspring[s];
          for (unsigned i = 0; i < dim; i++) {
              input->set(i,new vv::Double(ind[i]));
          }
          inputs.setInput(s,input);
      }
      const ct::OutputPlan& outputs = sim.simulatePlan(inputs,1,false,0);
      for (unsigned int s = 0; s < offspring.size(); s++) {
          double out = outputs.getOutput(s,0).toDouble().value();
          Indi& ind = offspring[s];
          ind.fitness(out);
      }
  }

private:
  unsigned int dim;
  ct::VleSimulator&  sim;
  vu::Rand& rng;
};

/**
 * @brief Optimizer for input and output domain into R.
 * An instance of a VERY simple Real-coded Genetic Algorithm
 * (based on Paradiseo Library)
 */
class paradiseo: public ci::FOmethod
{
public:
    /**
     * @brief paradiseo constructor
     */
    paradiseo(const vv::Map& init, ct::VleSimulator& sim, vu::Rand& rand) :
        ci::FOmethod(init,sim,rand),
        mmins(), mmaxs(), mresults(), mdimension(1)
    {
        const vv::Map& model = init.getMap("model");
        const vv::Map& parameters = init.getMap("parameters");
        const vv::Set& inputs = model.getSet("inputs");
        const vv::Map& domains = parameters.getMap("domains");

        mdimension = inputs.size();
        mmins.resize(mdimension);
        mmaxs.resize(mdimension);

        for (unsigned int p = 0; p < mdimension; p++) {
            const vv::Map& dom = domains.getMap(inputs.getString(p));
            mmins[p] = dom.getDouble("min");
            mmaxs[p] = dom.getDouble("max");
        }
        //TODO parameters
        seed = (parameters.exist("seed")) ? parameters.getInt("seed") : 12;
        popSize = (parameters.exist("pop-size")) ? parameters.getInt("pop-size") : 10;
        tSize = (parameters.exist("tournament-size")) ? parameters.getInt("tournament-size") : 2;
        maxGen = (parameters.exist("max-gen")) ? parameters.getInt("max-gen") : 100;
        minGen = (parameters.exist("min-gen")) ? parameters.getInt("min-gen") : 100;
        steadyGen = (parameters.exist("steady-gen")) ? parameters.getInt("steady-gen") : 100;
        pCross = (parameters.exist("p-cross")) ? parameters.getDouble("p-cross") : 0.6;
        pMut = (parameters.exist("p-mut")) ? parameters.getDouble("p-mut") : 0.1;
        hypercubeRate = (parameters.exist("hypercube-rate")) ? parameters.getDouble("hypercube-rate") : 1;
        segmentRate = (parameters.exist("segment-rate")) ? parameters.getDouble("segment-rate") : 1;
        epsilon = (parameters.exist("epsilon")) ? parameters.getDouble("epsilon") : 0.01;
        sigma = (parameters.exist("sigma")) ? parameters.getDouble("sigma") : 0.3;
        uniformMutRate = (parameters.exist("uniform-mut-rate")) ? parameters.getDouble("uniform-mut-rate") : 1;
        detMutRate = (parameters.exist("det-mut-rate")) ? parameters.getDouble("det-mut-rate") : 1;
        normalMutRate = (parameters.exist("normal-mut-rate")) ? parameters.getDouble("normal-mut-rate") : 1;
    }
    /**
     * @brief paradiseo destructor
     */
    ~paradiseo()
    {
    }


    /**
     * @brief Function to implement by an optimization method.
     * Not that the overall optimization process can be implemented into
     * one step (e.g. if the concept of step has no meaning for the method).
     * @param current_step the current step
     * @return true : always stop the process
     */
    bool optimizeStep(unsigned int /*current_step*/)
    {


        // EVAL
        /////////////////////////////
        // Fitness function
        ////////////////////////////
        // Evaluation: from a plain C++ fn to an EvalFunc Object
        // you need to give the full description of the function

        VlePopEvaluator popeval(mdimension,simulator(),rand());


        // INIT
          ////////////////////////////////
          // Initilisation of population
          ////////////////////////////////
          // Either load or initialize
          // create an empty pop
          eoPop<Indi> pop;
          // create a state for reading
          eoState inState;      // a state for loading - WITHOUT the parser
          // register the rng and the pop in the state, so they can be loaded,
          // and the present run will be the exact conitnuation of the saved run
          // eventually with different parameters
          inState.registerObject(rng);
          inState.registerObject(pop);

          rng.reseed(seed);

          // a Indi random initializer
          eoRealVectorBounds bounds = eoRealVectorBounds(mmins, mmaxs);
          eoRealInitBounded<Indi> random = eoRealInitBounded<Indi>(bounds);

          // Init pop from the randomizer: need to use the append function
          pop.append(popSize, random);
          // and evaluate pop (STL syntax)
          popeval.operator()(pop,pop);

        // OUTPUT
          // sort pop before printing it!
          pop.sort();
          // Print (sorted) intial population (raw printout)
          std::cout << "Initial Population" << std::endl;
          std::cout << pop;

        // ENGINE
          /////////////////////////////////////
          // selection and replacement
          ////////////////////////////////////
        // SELECT
          // The robust tournament selection
          eoDetTournamentSelect<Indi> selectOne(tSize);
          // is now encapsulated in a eoSelectPerc (entage)
          eoSelectPerc<Indi> select(selectOne);// by default rate==1

        // REPLACE
          // And we now have the full slection/replacement - though with
          // no replacement (== generational replacement) at the moment :-)
          eoGenerationalReplacement<Indi> replace;

        // OPERATORS
          //////////////////////////////////////
          // The variation operators
          //////////////////////////////////////
        // CROSSOVER
          // uniform chooce on segment made by the parents
          eoSegmentCrossover<Indi> xoverS;
          // uniform choice in hypercube built by the parents
          eoHypercubeCrossover<Indi> xoverA;
          // Combine them with relative weights
          eoPropCombinedQuadOp<Indi> xover(xoverS, segmentRate);
          xover.add(xoverA, hypercubeRate, true);

        // MUTATION
          // offspring(i) uniformly chosen in [parent(i)-epsilon, parent(i)+epsilon]
          eoUniformMutation<Indi>  mutationU(epsilon);
          // k (=1) coordinates of parents are uniformly modified
          eoDetUniformMutation<Indi>  mutationD(epsilon);
          // all coordinates of parents are normally modified (stDev SIGMA)
          eoNormalMutation<Indi>  mutationN(sigma);
          // Combine them with relative weights
          eoPropCombinedMonOp<Indi> mutation(mutationU, uniformMutRate);
          mutation.add(mutationD, detMutRate);
          mutation.add(mutationN, normalMutRate, true);

          // The operators are  encapsulated into an eoTRansform object
          eoSGATransform<Indi> transform(xover, pCross, mutation, pMut);

        // STOP
          //////////////////////////////////////
          // termination condition see FirstBitEA.cpp
          /////////////////////////////////////
          eoGenContinue<Indi> genCont(maxGen);
          eoSteadyFitContinue<Indi> steadyCont(minGen, steadyGen);
          eoFitContinue<Indi> fitCont(0);
          eoCombinedContinue<Indi> continuator(genCont);
          continuator.add(steadyCont);
          continuator.add(fitCont);

        // GENERATION
          /////////////////////////////////////////
          // the algorithm
          ////////////////////////////////////////

          // Easy EA requires
          // stopping criterion, eval, selection, transformation, replacement
          eoEasyEA<Indi> gga(continuator, popeval, select, transform, replace);

          // Apply algo to pop - that's it!
          gga(pop);

        // OUTPUT
          // Print (sorted) intial population
          pop.sort();
          std::cout << "FINAL Population\n" << pop << std::endl;
          std::cout << "FINAL Pop best \n" << pop.best_element().fitness() << std::endl;

          mresults.addDouble("y-min",pop.best_element().fitness());

          vv::Tuple* xmin = new vv::Tuple();
          for(unsigned int i = 0; i < pop.best_element().size(); i++){
              xmin->add(pop.best_element()[i]);
          }
          mresults.add("x-min",xmin);
          return true;
    }
    /**
     * @brief Observation method
     * @param port, the name of the port to observe
     * @return the observed value
     */
    virtual vv::Value* observe(const std::string& port) const
    {
        if(port == "y-min"){
            if(mresults.exist("y-min")){
                return mresults.get("y-min")->clone();
            }
        }
        if(port == "x-min"){
            if(mresults.exist("x-min")){
                return mresults.get("x-min")->clone();
            }
        }
        return 0;
    }

    /**
     * @brief The minimal values for each dimension of the search problem.
     */
    std::vector<double> mmins;
    /**
     * @brief The maximal values for each dimension of the search problem.
     */
    std::vector<double> mmaxs;
    /**
     * @brief Map of results
     */
    vv::Map mresults;
    /**
     * @brief Dimension of the problem
     */
    unsigned int mdimension;
    /**
     * @brief Seed of the paradiseo number generator;
     */
    unsigned int seed;
    /**
     * @brief Population size
     */
    unsigned int popSize;
    /**
     * @brief Tournament size
     */
    unsigned int tSize;
    /**
     * @brief Maximum number of generations
     */
    unsigned int maxGen;
    /**
     * @brief Minimum number of generations
     */
    unsigned int minGen;
    /**
     * @brief Number of generations with no improvement
     */
    unsigned int steadyGen;
    /**
     * @brief Probability of Crossover
     */
    double pCross;
    /**
     * @brief Probability of Mutation
     */
    double pMut;
    /**
     * @brief Relative rate for hypercube crossover
     */
    double hypercubeRate;
    /**
     * @brief Relative rate for segment crossover
     */
    double segmentRate;
    /**
     * @brief Width for uniform mutation
     */
    double epsilon;
    /**
     * @brief Sigma for normal mutation
     */
    double sigma;
    /**
     * @brief Relative rate for uniform mutation
     */
    double uniformMutRate;
    /**
     * @brief Relative rate for det-uniform mutation
     */
    double detMutRate;
    /**
     * @brief Relative rate for normal mutation
     */
    double normalMutRate;
};

}
}
}//namespaces

#endif /* CFO_FO_METHODS_PARADISEO_PARADISEO_HPP_ */
