/**
 * cfo/fo_methods/random/random.hpp
 * @author The RECORD Development Team (INRA) and the VLE Development Team
 * See the AUTHORS or Authors.txt file
 */

/*
 * Copyright (C) 2009-2010 INRA
 *
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef CFO_FO_METHODS_RANDOM_RANDOM_HPP_
#define CFO_FO_METHODS_RANDOM_RANDOM_HPP_

#include <iostream>

#include <vle/value/Map.hpp>
#include <vle/value/Set.hpp>
#include <vle/utils/Rand.hpp>

#include <cfo/interfaces/FOmethod.hpp>

namespace cfo {
namespace fo_methods {
namespace random {

namespace ct = cfo::tools;
namespace ci = cfo::interfaces;
namespace vv = vle::value;
namespace vu = vle::utils;

/**
 * @brief Optimizer for input and output domain into R.
 * It simulates n points and return the minimal value.
 */
class random: public ci::FOmethod
{
public:
    /**
     * @brief random constructor
     */
    random(const vv::Map& init, ct::VleSimulator& sim) :
        ci::FOmethod(init,sim), msimulationNbr(10), mmins(), mmaxs(),
        museExpPlan(true), mresults(), mdimension(1), mrand()
    {
    	const vv::Map& model = init.getMap("model");
    	const vv::Map& parameters = init.getMap("parameters");
    	const vv::Set& inputs = model.getSet("inputs");
        const vv::Map& domains = parameters.getMap("domains");

        mdimension = inputs.size();
        mmins.resize(mdimension);
        mmaxs.resize(mdimension);

        for (unsigned int p = 0; p < mdimension; p++) {
            const vv::Map& dom = domains.getMap(inputs.getString(p));
            mmins[p] = dom.getDouble("min");
            mmaxs[p] = dom.getDouble("max");
        }
        msimulationNbr = parameters.getInt("number-points");
        if (parameters.exist("use-experiment-plan")) {
            museExpPlan = parameters.getBoolean("use-experiment-plan");
        }
        if (parameters.exist("seed")) {
            mrand.seed(parameters.getInt("seed"));
        }
    }
    /**
     * @brief random destructor
     */
    ~random()
    {
    }
    /**
     * @brief Function to implement by an optimization method.
     * Not that the overall optimization process can be implemented into
     * one step (e.g. if the concept of step has no meaning for the method).
     * @param current_step the current step
     * @return true : always stop the process
     */
    virtual bool optimizeStep(unsigned int /*current_step*/)
    {
    	if (museExpPlan) {
    		optimisesWithExpPlan();
    	} else {
    		optimisesWithoutExpPlan();
    	}
    	return true;
    }
    /**
     * @brief Observation method
     * @param port, the name of the port to observe
     * @return the observed value
     */
    virtual vv::Value* observe(const std::string& port) const
    {
    	if(port == "y-min"){
    		if(mresults.exist("y-min")){
    			return mresults.get("y-min")->clone();
    		}
    	}
    	if(port == "x-min"){
    		if(mresults.exist("x-min")){
    			return mresults.get("x-min")->clone();
    		}
    	}
    	return 0;
	}

private:
    /**
     * @brief Use experiment plan (method VleSimulator::simulatePlan)
     * for optimization.
     */
    void optimisesWithExpPlan()
    {
        double ymin = 1000;
        vv::Set xmin(mdimension);
        vv::Set inputs(msimulationNbr);
        for (int s = 0; s < msimulationNbr; s++) {
            vv::Set* input = new vv::Set(mdimension);
            for (unsigned i = 0; i < mdimension; i++) {
                input->set(i,new vv::Double(mrand.getDouble(mmins[i],mmaxs[i])));
            }
            inputs.set(s,input);
        }
        const vv::Set& outputs = simulator().simulatePlan(inputs, true);
        for (int s = 0; s < msimulationNbr; s++) {
            double out = outputs.getDouble(s);
            if (out < ymin) {
                ymin = out;
                const vv::Set& input_min = inputs.getSet(s);
                for (unsigned i = 0; i < mdimension; i++) {
                    xmin.set(i,new vv::Double(input_min.getDouble(i)));
                }
            }
        }
        mresults.add("y-min",new vv::Double(ymin));
        mresults.add("x-min",xmin);
    }
    /**
     * @brief Use simple simulation (method VleSimulator::simulate)
     * for optimization.
     */
    void optimisesWithoutExpPlan()
    {
        double ymin = 1000;
        vv::Set xmin(mdimension);
        double ycurr = 1000;
        vv::Set xcurr(mdimension);
        for (int s = 0; s < msimulationNbr; s++) {
            for (unsigned i = 0; i < mdimension; i++) {
                xcurr.set(i,new vv::Double(mrand.getDouble(mmins[i],mmaxs[i])));
            }
            ycurr = simulator().simulate(xcurr).toDouble().value();
            if (ycurr < ymin) {
                ymin = ycurr;
                for (unsigned i = 0; i < mdimension; i++) {
                    xmin.set(i,new vv::Double(xcurr.getDouble(i)));
                }
            }
        }
        mresults.add("y-min",new vv::Double(ymin));
        mresults.add("x-min",xmin);
    }
    /**
     * @brief The number of simulations to do.
     */
    int msimulationNbr;
    /**
     * @brief The minimal values for each dimension of the search problem.
     */
    std::vector<double> mmins;
    /**
     * @brief The maximal values for each dimension of the search problem.
     */
    std::vector<double> mmaxs;
    /**
     * @brief If true, use simulation of experiment plans.
     */
    bool museExpPlan;
    /**
     * @brief Map of results
     */
    vv::Map mresults;
    /**
     * @brief Dimension of the problem
     */
    unsigned int mdimension;
    /**
     * @brief a random number generator
     */
    vu::Rand mrand;
};

}
}
}//namespaces

#endif /* CFO_FO_METHODS_RANDOM_RANDOM_HPP_ */
