/**
 * cfo/fo_methods/shark/sharCMA.hpp
 * @author The RECORD Development Team (INRA) and the VLE Development Team
 * See the AUTHORS or Authors.txt file
 */

/*
 * Copyright (C) 2009-2010 INRA
 *
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef CFO_FO_METHODS_SHARK_CMA_HPP_
#define CFO_FO_METHODS_SHARK_CMA_HPP_


#include <vle/value/Map.hpp>
#include <vle/value/Set.hpp>
#include <vle/utils/Rand.hpp>

#include <boost/format.hpp>

#include <cfo/fo_methods/shark/utils.hpp>
#include <cfo/interfaces/FOmethod.hpp>

#include <EALib/ChromosomeCMA.h>

namespace vv = vle::value;
namespace vu = vle::utils;


namespace cfo {
namespace fo_methods {
namespace shark {

namespace ct = cfo::tools;
namespace ci = cfo::interfaces;

/**
 * @brief sharkCMA implementation of FOmethod.
 * This implementation is the one proposed by Shark
 * in his tutorial.
 */
class sharkCMA: public ci::FOmethod
{

public:
    /**
     * @brief sharkCMA constructor
     */
    sharkCMA(const vv::Map& init, ct::VleSimulator& sim, vu::Rand& randgen) :
        FOmethod(init,sim,randgen), margs(), simInput(0), mparent(),
        mfitnessParent(0), moffspring(), mresults()
    {
    	//look parameters
        const vv::Map& model = init.getMap("model");
        const vv::Map& parameters = init.getMap("parameters");
        const vv::Set& inputs = model.getSet("inputs");
        const vv::Set& outputs = model.getSet("outputs");
        const vv::Map& domains = parameters.getMap("domains");

        //load input config
        margs.minInit.resize(inputs.size());
        margs.maxInit.resize(inputs.size());
        for (unsigned int p = 0; p < inputs.size(); p++) {
            const vv::Map& dom = domains.getMap(inputs.getString(p));
            margs.minInit[p] = dom.getDouble("min");
            margs.maxInit[p] = dom.getDouble("max");
        }
        //look output config
        if(outputs.size() != 1){
            throw vu::ArgError(_("sharkCMA method can handle only one output "));
        }
        if(parameters.exist("scales")){
        	margs.sim_output_scale = parameters.getDouble("scale");
        } else {
            margs.sim_output_scale = 1;
        }
        if (parameters.exist("global-step-init")) {
            double v = parameters.getDouble("global-step-init");
            margs.global_step_init = v;
        }
        if (parameters.exist("nb-iterations")) {
            unsigned int v = parameters.getInt("nb-iterations");
            margs.nb_iterations = v;
        }
        if(parameters.exist("sim-common-seed")) {
            bool v = parameters.getBoolean("sim-common-seed");
            margs.sim_common_seed = v;
        }
        if(parameters.exist("sim-nb-replicas")) {
            int v = parameters.getInt("sim-nb-replicas");
            if(v <= 0){
                throw vu::ArgError(vle::fmt(_(
                    "bad number of replicas '%1%'")) % v);
            }
            margs.sim_nb_replicas = v;
        }
        //prepare default input for InputVector
        vv::Set defInput;
        for(unsigned int i = 0; i < inputs.size(); i++){
            defInput.addDouble(0);
        }
        simInput = new ct::InputVector(defInput);
        simInput->resize(1);
        //Initialization of the optimization process
        const unsigned dimension = margs.minInit.size();
        std::vector<double> stddev_init(dimension, 1);
        mfixedSeed = rand().getInt();
        mparent.init(dimension, stddev_init, margs.global_step_init,
                margs.minInit, margs.maxInit);
        mfitnessParent = fitness(mparent);
        //Prepare results
        mresults.addDouble("y-min",mfitnessParent);
        vv::Set& best_x = mresults.addSet("x-min");
        for (unsigned int i = 0; i < mparent.size(); i++) {
            best_x.addDouble(mparent[i]);
        }

    }
    /**
     * @brief sharkCMA destructor
     */
    ~sharkCMA()
    {
        if(simInput){
            delete simInput;
        }
    }
    /**
     * @brief Function to implement by an optimization method.
     * Not that the overall optimization process can be implemented into
     * one step (e.g. if the concept of step has no meaning for the method).
     * @param current_step the current step
     * @return true if the process should stop
     */
    virtual bool optimizeStep(unsigned int current_step)
    {
    	double fitnessOffspring;
    	moffspring = mparent;
    	moffspring.mutate();
    	fitnessOffspring = fitness(moffspring);
    	bool success = (fitnessOffspring < mfitnessParent);
    	if (success) {
    		moffspring.updateCovariance();
    		mparent = moffspring;
    		mfitnessParent = fitnessOffspring;
    		//update results
    		mresults.getDouble("y-min") = mfitnessParent;
    		vv::Set& best_x = mresults.getSet("x-min");
    		for (unsigned int i = 0; i < mparent.size(); i++) {
    			best_x.getDouble(i) = mparent[i];
    		}
    	}
    	mparent.updateGlobalStepsize(success);
    	return (current_step >= margs.nb_iterations);
    }
    /**
     * @brief Observation method
     * @param port, the name of the port to observe
     * @return the observed value
     */
    virtual vv::Value* observe(const std::string& port) const
    {
    	if(mresults.exist(port)){
    		return mresults.get(port)->clone();
    	}
    	return 0;
	}

private:
    /**
     * @brief get the seed value for simulation
     * @return seed, the current seed
     */
    int getSeed()
    {
        if(margs.sim_common_seed){
            return mfixedSeed;
        } else {
            return mrand.getInt();
        }

    }
    /**
     * @brief fitness computation relies on simulation
     * @param c, the chromosome which fitness needs to be evaluated
     * @return fitness evaluation of c
     */
    double fitness(ChromosomeT<double> &c)
    {
        vv::Set& input = simInput->getInput(0).toSet();
        for(unsigned i=0; i < c.size(); i++){
            input.set(i,new vv::Double(c[i]));
        }
        const ct::OutputPlan& res = simulator().simulatePlan(*simInput,
            margs.sim_nb_replicas, margs.sim_common_seed, getSeed());
        double sum = 0;
        for(int r = 0; r < margs.sim_nb_replicas ; r++){
            sum += res.getOutput(0,r).toDouble().value();
        }
        return (sum/margs.sim_nb_replicas) * margs.sim_output_scale;
    }
    /**
     * @brief sharkCMA arguments
     */
    SharkCMAargs margs;
    /**
     * @brief Input of simulation
     */
    ct::InputVector* simInput;
    /**
     *@brief the fixed seed in case common-seed is true
     */
    bool mfixedSeed;
    /**
     * @brief Parent chromosome
     */
    ChromosomeCMA mparent;
    /**
     * @brief The best fitness
     */
    double mfitnessParent;
    /**
     * @brief Offspring Chromosome
     */
    ChromosomeCMA moffspring;
    /**
     * @brief Map of results
     */
    vv::Map mresults;
};

}
}
}//namespaces

#endif //CFO_FO_METHODS_SHARK_CMA_HPP_
