/**
 * cfo/interfaces/FOmethod.hpp
 * @author The RECORD Development Team (INRA) and the VLE Development Team
 * See the AUTHORS or Authors.txt file
 */

/*
 * Copyright (C) 2009-2010 INRA
 *
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef CFO_INTERFACES_FOmethod_HPP_
#define CFO_INTERFACES_FOmethod_HPP_

#include <vector>
#include <utility>
#include <cmath>
#include <iostream>

#include <vle/value/Map.hpp>
#include <vle/devs/Dynamics.hpp>

#include <cfo/tools/VleSimulator.hpp>
#include <cfo/interfaces/FOexpe.hpp>

namespace cfo {
namespace interfaces {

namespace vu = vle::utils;
namespace vv = vle::value;
namespace ct = cfo::tools;

/**
 * @brief Class that has to be implemented by all
 * function optimization methods
 */
class FOmethod
{
public:
    /**
     * @brief FOmethod constructor
     * @param init a map that contains initialization parameters for the
     * optimization, note that the same Map is used to initialized the simulator.
     * @param simulator a reference to the simulator
     */
    FOmethod(const vv::Map& init, ct::VleSimulator& simulator):
        minit(init), msimulator(simulator)
    {
    }
    /**
     * @brief FOmethod destructor
     */
    virtual ~FOmethod()
    {
    }
    /**
     * @brief Function to implement by an optimization method.
     * Not that the overall optimization process can be implemented into
     * one step (e.g. if the concept of step has no meaning for the method).
     * @param current_step the current step
     * @return true if the process should stop
     */
    virtual bool optimizeStep(unsigned int current_step) = 0;
    /**
     * @brief Observation method
     * @param port, the name of the port to observe
     * @return the observed value
     */
    virtual vv::Value* observe(const std::string& /*port*/) const
    {
        return 0;
    }

    /**
     * @brief Access method to the simulator
     * return a reference to the simulator
     */
    ct::VleSimulator& simulator()
    {
        return msimulator;
    }

protected:
    /**
     * @brief A reference on the initialization object
     */
    const vv::Map& minit;
    /**
     * @brief A reference on the the simulator
     */
    ct::VleSimulator& msimulator;

};

}}//namespaces

#define DECLARE_FO_METHOD(method)\
    DECLARE_DYNAMICS(cfo::interfaces::FOexpe<method >)

#endif
