/**
 * @file cfo/interfaces/RLagent.hpp
 * @author The RECORD Development Team (INRA)
 */

/*
 * Copyright (C) 2009 INRA
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef CFO_INTERFACES_RLAGENT_HPP_
#define CFO_INTERFACES_RLAGENT_HPP_

#include <vle/devs/Dynamics.hpp>
#include <vle/utils/Exception.hpp>

namespace vd = vle::devs;
namespace vv = vle::value;
namespace vu = vle::utils;

namespace cfo {
namespace interfaces {


/**
 * @brief Dynamic of an agent, the template RL_METHOD is an instantiation of a
 * reinforcement learning method
 */
template <class RL_AGENT_METHOD>
class RLagent: public vd::Dynamics
{
private:
    /**
     * @brief Dynamic state type of the RLagent
     */
    enum State {START, ASLEEP, PROCESS };
    /**
     * @brief The RL method
     */
    RL_AGENT_METHOD mrlmethod;
    /**
     * @brief  Dynamic internal state of the RLagent
     */
    State mstate;
    /**
     * @brief The current action to send to the environment
     */
    vv::Value* maction;
    /**
     * @brief The set of ports to observe
     */
    vv::Set* mobserved;
public:
    /**
     * @brief RLagent constructor
     */
    RLagent(const vd::DynamicsInit& init, const vd::InitEventList& events) :
        vd::Dynamics(init,events), mrlmethod(events),
        mstate(START), maction(0), mobserved(0)
    {
    	if(events.exist("parameters")){
    		const vv::Map& params = events.getMap("parameters");
    		if(params.exist("observed")){
    			mobserved = new vv::Set(params.getSet("observed"));
    		}
    	}

    }
    /**
     * @brief RLagent destructor
     */
    ~RLagent()
    {
        if(maction){
            delete maction;
        }
        if(mobserved){
        	delete mobserved;
        }
    }
    /****/
    vd::Time init(const vd::Time& /* time */)
    {
        return vd::infinity;
    }
    /****/
    void output(const vd::Time& /* time */, vd::ExternalEventList& output) const
    {
        if(mstate == PROCESS){
            if(!maction){
                throw vu::InternalError("[RLagent] action to send is null");
            }
        	vd::ExternalEvent* evt = new vd::ExternalEvent("agent-action");
            evt->putAttribute("value",maction->clone());
            evt->putAttribute("name",new vv::String("agent-action"));
            output.push_back(evt);
        }
    }
    /****/
    vd::Time timeAdvance() const
    {
        switch(mstate){
        case START :
        case ASLEEP :
            return vd::infinity;
            break;
        default :
            return 0;
            break;
        }
        return vd::infinity;
    }
    /****/
    void internalTransition(const vd::Time& /* time */)
    {
        switch(mstate){
        case PROCESS :
            mstate = ASLEEP;
            break;
        default :
            throw vu::InternalError(
                "[RLagent] unexpected state in internalTransition");
            break;
        }
    }
    /****/
    void externalTransition(const vd::ExternalEventList& event, const vd::Time& /* time */)
    {

        vd::ExternalEventList::const_iterator itb = event.begin();
        vd::ExternalEventList::const_iterator ite = event.end();
        for(;itb!=ite;itb++){
            const vd::ExternalEvent& e = *(*itb);
            if(e.onPort("info-env")){
                switch(mstate){
                case START :
                    if(maction){
                        delete maction;
                    }
                    maction = mrlmethod.agent_start(
                        e.getAttributeValue("observation")).clone();
                    mstate = PROCESS;
                    break;
                case ASLEEP :
                    if(maction){
                        delete maction;
                    }
                    if(!e.existAttributeValue("observation")){
                        mrlmethod.agent_episode_ends(
                            e.getDoubleAttributeValue("reward"));
                        maction = 0;
                        mstate = ASLEEP;
                    } else {
                        maction = mrlmethod.agent_step(
                            e.getDoubleAttributeValue("reward"),
                            e.getAttributeValue("observation")).clone();
                        mstate = PROCESS;
                    }

                    break;
                default :
                    throw vu::InternalError(
                        "[RLagent] unexpected state in externalTransition");
                    break;
                }
            }
        }
    }
    /****/
    vv::Value* observation(const vd::ObservationEvent& e) const
    {
        if(e.onPort("episodeResults")){
        	vv::Map* res = new vv::Map();
        	res->addBoolean("stop-process",mrlmethod.is_last_episode());
            if(mobserved){
            	vv::Map& o = res->addMap("observed");
            	for(unsigned int i = 0; i < mobserved->size();i++){
            		o.add(mobserved->getString(i),
            				mrlmethod.observe(mobserved->getString(i)));
            	}
            }
            return res;
        }
        return 0;
    }

};

}}//namespaces

#endif
