/**
 * cfo/interfaces/RLagentMethod.hpp
 * @author The RECORD Development Team (INRA) and the VLE Development Team
 * See the AUTHORS or Authors.txt file
 */

/*
 * Copyright (C) 2009-2010 INRA
 *
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef CFO_INTERFACES_RLagentMethod_HPP_
#define CFO_INTERFACES_RLagentMethod_HPP_

#include <vector>
#include <utility>
#include <cmath>
#include <iostream>

#include <vle/value/Map.hpp>

#include <cfo/interfaces/RLagent.hpp>


namespace cfo {
namespace interfaces {

namespace vv = vle::value;

/**
 * @brief Class that must be implemented by all
 * reinforcement learning methods
 */
class RLagentMethod
{
public:
    /**
     * @brief RLagentMethod constructor
     * @param init a map that contains initialization parameters for the
     * optimization, note that the same Map is used to initialized the simulator.
     * @param currentPolicy the policy learned so far
     * @param currentEpisode the number of the current episode
     */
    RLagentMethod(const vv::Map& init): minit(init)
    {
    }

    /**
     * @brief RLagentMethod destructor
     */
    virtual ~RLagentMethod()
    {
    }
    /**
     * @brief Answers to the first observation
     * @param observation, first observation of environment
     * @return the first action to perform
     */
    virtual const vv::Value& agent_start(const vv::Value& /*observation*/) = 0;

    /**
     * @brief Answers to a step
     * @param reward, the reward of the precedent action
     * @param observation, observation after the precedent action
     * @return the next action to perform
     */
    virtual const vv::Value& agent_step(double /*reward*/,
        const vv::Value& /*observation*/) = 0;

    /**
     * @brief In finite-horizon RL problems, agent can receive a final
     * reward at the end of the episode, no action is required by the
     * environment.
     * @param reward, the reward of the precedent action
     */
    virtual void agent_episode_ends(double /*reward*/)
    {
    }

    /**
     * @brief Tells if the optimization process
     * should stop.
     * @return true if the process has to stop
     */
    virtual bool is_last_episode() const
    {
        return false;
    }

    /**
     * @brief Observation method
     * @param port, the name of the port to observe
     * @return the observed value
     */
    virtual vv::Value* observe(const std::string& /*port*/) const
    {
        return 0;
    }
protected:
    /**
     * @brief A reference on the initialization object
     */
    const vv::Map& minit;

};

}}//namespaces

#define DECLARE_RL_AGENT_METHOD(method) \
        DECLARE_DYNAMICS(cfo::interfaces::RLagent<method >)


#endif
