/**
 * cfo/interfaces/RLexpeMethod.hpp
 * @author The RECORD Development Team (INRA) and the VLE Development Team
 * See the AUTHORS or Authors.txt file
 */

/*
 * Copyright (C) 2009-2010 INRA
 *
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef CFO_INTERFACES_RL_EXPE_METHOD_HPP_
#define CFO_INTERFACES_RL_EXPE_METHOD_HPP_


#include <vle/devs/Dynamics.hpp>
#include <vle/value/Value.hpp>

#include <cfo/tools/VleSimulator.hpp>

#include <cfo/interfaces/AbstractExpe.hpp>
#include <cfo/interfaces/RLexpe.hpp>


#include <iostream>

namespace cfo {
namespace interfaces {


/**
 * @brief Instantiation of an experiment for reinforcement learning problem.
 */
class RLexpeMethod
{

public:
    /**
     * @brief RLexpeMethod constructor
     */
    RLexpeMethod(const vv::Map& /*init*/, ct::VleSimulator& simulator):
        msim(simulator), mlastObserved(0)
    {
    }


    /**
     * @brief Method for initializing the experiment
     */
    virtual void initExpe()
    {
    }

    /**
     * @brief RLexpeMethod destructor
     */
    virtual ~RLexpeMethod()
    {
    	if(mlastObserved){
    		delete mlastObserved;
    	}

    }
    /**
     * @brief Method for a process step
     * @param current_step the current step
     * @return true if the process should stop
     */
    bool processStep(const vv::Integer& current_step)
    {
        const vv::Value& results = msim.simulate(current_step);
        if(results.isMap()){
            const vv::Map& episodeResults = results.toMap();
            if(episodeResults.exist("observed")){
                if(mlastObserved){
                    delete mlastObserved;
                }
                mlastObserved = new vv::Map(
                        episodeResults.getMap("observed"));
            }
            return (episodeResults.get("stop-process") &&
                 episodeResults.getBoolean("stop-process"));
        } else {
            //multiple outputs
            bool stop = true;
            const vv::Set& resultsSet = results.toSet();
            for(unsigned int i=0; i < resultsSet.size(); i++){
                const std::string& oname = msim.getOutputName(i);
                if(oname == "episodeResults"){
                    const vv::Map& episodeResults = resultsSet.getMap(i);
                    if(episodeResults.exist("observed")){
                        if(mlastObserved){
                            delete mlastObserved;
                        }
                        mlastObserved = new vv::Map(
                                episodeResults.getMap("observed"));
                    }
                    stop = (episodeResults.get("stop-process") &&
                            episodeResults.getBoolean("stop-process"));
                }
            }
            for(unsigned int i=0; i < resultsSet.size(); i++){
                const std::string& oname = msim.getOutputName(i);
                if(oname != "episodeResults"){
                    const vv::Value& otherRes = *resultsSet.get(i);
                    if(mlastObserved && !mlastObserved->exist(oname)){
                        mlastObserved->add(oname,otherRes.clone());
                    }
                }
            }
            return stop;
        }

    }

    /**
     * @brief Method that observes one port
     * @param obs, the port observed
     * @return the observed value
     */
    vv::Value* observe(const std::string& port) const
	{
    	if(mlastObserved){
    		if(mlastObserved->exist(port)){
    		    if(mlastObserved->get(port) != 0){
    		        return mlastObserved->get(port)->clone();
    		    }
    		}
    	}
    	return 0;
	}
private:
    ct::VleSimulator& msim;
    /**
     * @brief the maps of the le last observations
     */
    vv::Map* mlastObserved;
};

}}// namespaces

#define DECLARE_RL_EXPE_METHOD(method) \
        DECLARE_DYNAMICS(cfo::interfaces::RLexpe<method >)

#endif
