/**
 * cfo/rl_methods/sarsa/sarsa_agent.hpp
 * @author The RECORD Development Team (INRA) and the VLE Development Team
 * See the AUTHORS or Authors.txt file
 */

/*
 * Copyright (C) 2009-2010 INRA
 *
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef CFO_RL_METHODS_SARSA_SARSA_HPP_
#define CFO_RL_METHODS_SARSA_SARSA_HPP_

#include <vle/value/Map.hpp>
#include <vle/value/Integer.hpp>
#include <vle/value/Set.hpp>
#include <vle/utils/Rand.hpp>


#include <cfo/tools/policies/q_matrix.hpp>

#include <cfo/interfaces/RLagentMethod.hpp>


namespace ci = cfo::interfaces;
namespace ctp = cfo::tools::policies;
namespace vv = vle::value;
namespace vu = vle::utils;


namespace cfo {
namespace rl_methods {
namespace sarsa{

/**
 * sarsa Reinforcement learning agent
 */

class sarsa_agent : public ci::RLagentMethod
{
public:
    /**
     * @brief sarsa RLmethod constructor
     * @param init a map that contains initialization parameters for the
     * optimization, note that the same Map is used to initialized the simulator.
     */
    sarsa_agent(const vv::Map& init):
            ci::RLagentMethod(init),
            mepisode(0), mcurrState(0), mcurrAction(0),
            mepsilon(0.1), mgamma(1.0), malpha(0.1), mnbMoves(0),
            mmaxEpisodes(6)
    {
        std::cout << " DBG init agent " << init << std::endl;

        mepisode = init.getInt("currentEpisode");
        const vv::Map& parameters = init.getMap("parameters");
    	if(mepisode == 1){
    		policy().initFromActions(parameters.getSet("actions"));
    	}
    	if(parameters.exist("epsilon")){
//    		std::cout << " Sarsa epsilon " << mepsilon << std::endl;
    		mepsilon = parameters.getDouble("epsilon");
    	}
    	if(parameters.exist("gamma")){
    		mgamma = parameters.getDouble("gamma");
    	}
    	if(parameters.exist("alpha")){
    		malpha = parameters.getDouble("alpha");
    	}
    	if(parameters.exist("maxEpisodes")){
//    		std::cout << " Sarsa max set " << std::endl;
    		mmaxEpisodes = parameters.getInt("maxEpisodes");
    	}
//    	std::cout << " Sarsa max " << mmaxEpisodes << std::endl;

    }

    ~sarsa_agent()
    {
    }
    /**
     * @brief Answers to the first observation
     * @param observation, first observation of environment
     * @return the first action to perform (a copy)
     */
    const vv::Value& agent_start(const vv::Value& observation)
    {
        ctp::Qmatrix& qmatrix = policy();
        mcurrState = qmatrix.getState(observation);
        std::pair<unsigned int, bool> epsGreedy = epsilon_greedy(mcurrState);
        mcurrAction = epsGreedy.first;
        mcurrExplore = epsGreedy.second;
        return qmatrix.getAction(mcurrAction);
    }

    /**
     * @brief Answers to a step
     * @param reward, the reward of the precedent action
     * @param observation, observation after the precedent action
     * @return the next action to perform (a copy)
     */
    const vv::Value& agent_step(double reward,
        const vv::Value& observation)
    {

        ctp::Qmatrix& qmatrix = policy();
        ctp::Qmatrix::state_id_t newstate = qmatrix.getState(observation);
        std::pair<unsigned int, bool> epsGreedy = epsilon_greedy(newstate);
        ctp::Qmatrix::action_id_t newaction = epsGreedy.first;
        qmatrix.updateSarsa(mcurrState, mcurrAction, newstate, newaction, reward,
            mgamma, malpha );
        mcurrAction = newaction;
        mcurrExplore = epsGreedy.second;
        mcurrState = newstate;
        mnbMoves ++;
        return qmatrix.getAction(mcurrAction);
    }
    /**
     * @brief Implementation
     * @param reward, the reward of the precedent action
     */
    void agent_episode_ends(double reward)
    {
        ctp::Qmatrix& qmatrix = policy();
        qmatrix.updateSarsaEnd(mcurrState, mcurrAction,reward, malpha);
    }
    /**
     * @brief Tells if the optimization process
     * should stop.
     * @return true if the process has to stop
     */
    bool is_last_episode() const
    {
    	return (mepisode >= mmaxEpisodes);
    }

    /**
     * @brief Observation method
     * @param port, the name of the port to observe
     * @return the observed value
     */
    vv::Value* observe(const std::string& port) const
    {
        std::cout << " DBG obs agent " << port << std::endl;
    	if(port == "nb-moves"){
    		return new vv::Integer(mnbMoves);
    	}
    	if(port == "qmatrix"){
    		return policy().buildPolicyRepresentation();
    	}
    	return 0;
	}
    /**
     * @brief Static access to the policy singleton
     * @return the policy singleton
     */
    static ctp::Qmatrix& policy()
    {
        if (sarsa_agent::mpolicy == 0) {
            sarsa_agent::mpolicy = new ctp::Qmatrix();
        }
        return *mpolicy;
    }
private:

    /**
     * @brief Epsilon greedy returns the greedy action
     * (the action with the best Q value) with probability 1-epsilon. Otherwise,
     * it picks an action according to a uniform distribution.
     * @param state id of the current state
     * @return the id of the action to perform, and a boolean that indicates if
     * this is an exploration.
     */
    std::pair<unsigned int,bool> epsilon_greedy(unsigned int state)
    {
        const vv::Set& actions = policy().getActions();
        if(mrand.getDouble() < mepsilon){
            //explore
            return std::make_pair(mrand.getInt(0,actions.size()-1),true);
        }
        return std::make_pair(policy().getBestAction(state),false);
    }
    /**
     * @brief the current episode
     */
    unsigned mepisode;
    /**
     * @brief index of the current state of the agent.
     */
    ctp::Qmatrix::state_id_t mcurrState;
    /**
     * @brief index of the last action performed by the agent.
     */
    ctp::Qmatrix::action_id_t mcurrAction;
    /**
     * @brief tells if the current action is an exploration
     */
    bool mcurrExplore;
    /**
     * @brief epsilon parameter of Sarsa
     */
    double mepsilon;
    /**
     * @brief gamma parameter of Sarsa
     */
    double mgamma;
    /**
     * @brief alpha parameter of Sarsa
     */
    double malpha;
    /**
     * @brief Number of moves during the episode
     */
    double mnbMoves;
    /**
     * @brief The maximal number of episodes for the learning process
     */
    unsigned int mmaxEpisodes;
    /**
     * @brief Policy singleton
     */
    static ctp::Qmatrix* mpolicy;
    /**
     * @brief a random number generator
     */
    vu::Rand mrand;
};



}}} //namespaces



#endif
