/**
 * cfo/rl_methods/xitek/xitek_agent.hpp
 * @author The RECORD Development Team (INRA) and the VLE Development Team
 * See the AUTHORS or Authors.txt file
 */

/*
 * Copyright (C) 2009-2010 INRA
 *
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef CFO_RL_METHODS_XITEK_XITEK_HPP_
#define CFO_RL_METHODS_XITEK_XITEK_HPP_


#include <math.h>

#include <boost/algorithm/string/predicate.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/algorithm/string/split.hpp>
#include <boost/algorithm/string/classification.hpp>

#include <vle/value/Map.hpp>
#include <vle/value/Tuple.hpp>
#include <vle/value/Set.hpp>
#include <vle/utils/Rand.hpp>

#include <cfo/tools/regression/CMAC.hpp>
#include <cfo/tools/policies/policy.hpp>
#include <cfo/interfaces/RLagentMethod.hpp>


namespace ctp = cfo::tools::policies;
namespace ci = cfo::interfaces;
namespace ctr = cfo::tools::regression;
namespace vv = vle::value;
namespace vu = vle::utils;


namespace cfo {
namespace rl_methods {
namespace xitek{


/**
 * @brief a xitekStage configuration
 * the template class REGRESSOR is a IncrRegressor (see tools)
 */
template <class REGRESSOR>
struct xitekStage
{
    /**
     * @brief default XITEK stage constructor
     */
    xitekStage() : mregressors(),mobsNames()
    {
    }
    /**
     * @brief A component of the xitek policy (ie. one stage)
     * @param init, a map that should contain :
     *  - an integer "nb-actions"
     *  - a set of string "observables"
     *  - eventually, the configuration of the regressors
     */

    xitekStage(const vv::Map& init) :
            mregressors(),mobsNames()
    {
        const vv::Set& obsConfig = init.getSet("observables");
        unsigned int nbactions = init.getInt("nb-actions");
        //build the parameters for observalbes
        for(unsigned int o = 0 ; o < obsConfig.size() ; o++){
            mobsNames.push_back(obsConfig.getString(o));
        }
        //build the parameters for regressors
        mregressors.resize(nbactions);
        for(unsigned int a = 0 ; a < nbactions ; a++){
            mregressors[a].init(init);
        }
    }
    /**
     * @brief Get the number of dimensions of the xitek stage problem
     * @return the number of dimensions
     */
    unsigned int dimension()
    {
        return mobsNames.size();
    }
    /**
     * @brief Get the number of possible actions for the xitek stage problem
     * @return the number of actions
     */
    unsigned int nbActions() const
    {
        return mregressors.size();
    }
    /**
     * @brief Constant access to the corresponding regressor for one action
     * @param act the action index
     * @return the corresponding regressor
     */
    const REGRESSOR& getRegressor(unsigned int act) const
    {
        return mregressors[act];
    }
    /**
     * @brief Gets the corresponding regressor for one action
     * @param act the action index
     * @return the corresponding regressor
     */
    REGRESSOR& getRegressor(unsigned int act)
    {
        return mregressors[act];
    }
    /**
     * @brief Builds the tuple by filtering observations
     * @param observations, all the observations
     * @return a tuple that contains only values of observations
     * required for this xitekStage
     */
    vv::Tuple getObsTuple(const vv::Map& observations)
    {
        vv::Tuple tuple(mobsNames.size());
        for(unsigned int obs = 0; obs < mobsNames.size() ; obs++){
            const std::string& obsName = mobsNames[obs];
            tuple.value()[obs] = observations.getDouble(obsName);
        }
        return tuple;
    }

    /**
     * @brief Get the best current action given an observation
     * @param observations all the observations
     * @return the a pair : the best action, and the predicted value
     */
    unsigned int getBestAction(const vv::Tuple& point)
    {
        double bestVal = -10000;
        unsigned int bestAction = 0;
        double prediction;
        //browse the action space to find the best action
        for(unsigned int a = 0; a < mregressors.size() ; a++){
            prediction = mregressors[a].predict(point);
            if(prediction > bestVal){
                bestVal = prediction;
                bestAction = a;
            }
        }
        return bestAction;
    }

    /**
     * @brief Converts the xitekStage into a value
     * @return the value that represents the xitekStage
     */
    vv::Value* toValue() const
    {
        vv::Map* xitekStageValue = new vv::Map();
        vv::Set* regressorsValue = new vv::Set();
        vv::Set* obsValue = new vv::Set();
        for(unsigned int i = 0; i < mregressors.size(); i++){
            regressorsValue->add(mregressors[i].toValue());
        }
        for(unsigned int i = 0; i < mobsNames.size(); i++){
            obsValue->addString(mobsNames[i]);
        }
        xitekStageValue->add("regressors",regressorsValue);
        xitekStageValue->add("obsNames",obsValue);
        return xitekStageValue;
    }
    /**
     * @brief Builds the xitekStage from a value
     * @param value, the value that represents the xitekStage
     */
    void fromValue(const vv::Value& value)
    {
        mregressors.clear();
        mobsNames.clear();
        const vv::Map& xitekStageValue = value.toMap();
        const vv::Set& regressorsValue = xitekStageValue.getSet("regressors");
        const vv::Set& obsValue = xitekStageValue.getSet("obsNames");
        mregressors.resize(regressorsValue.size());
        for(unsigned int i = 0; i < regressorsValue.size(); i++){
            mregressors[i].fromValue(*(regressorsValue[i]));
        }
        for(unsigned int i = 0; i < obsValue.size(); i++){
            mobsNames.push_back(obsValue.getString(i));
        }
    }

    /**
     * @brief store regressors, one regressor by action
     */
    std::vector<REGRESSOR> mregressors;
    /**
     * @brief names of the observables
     */
    std::vector<std::string> mobsNames;

};



/**
 * @brief  the xitekPolicy used for the RL method xitek.
 * the template class REGRESSOR is a IncrRegressor (see tools)
 */
template <class REGRESSOR>
class xitekPolicy : public ctp::policy
{
public:


    typedef std::map<std::string, xitekStage<REGRESSOR> > stage_container;
    typedef typename stage_container::const_iterator stage_container_cit;
    typedef typename stage_container::iterator stage_container_it;

    /**
     * @brief XITEK policy constructor
     */
    xitekPolicy() : mstages(), mrho(0), mQhlearning(false)
    {
    }
    /**
     * @brief XITEK policy destructor
     */
    ~xitekPolicy()
    {
    }

    /**
     * @brief Static function for initializing the singleton
     * @param init, a map which each entry correponds to a stage. An entry
     * must be a map that contains :
     *   - an integer "nb-actions"
     *   - a set of strings "observables"
     * @param Qhlearning, tells if Qh-learning is performed
     */
    void initFromStageSpecifications(const vv::Map& stages, bool Qhlearning)
    {
        vv::Map::const_iterator itb = stages.begin();
        vv::Map::const_iterator ite = stages.end();
        for(;itb != ite; itb++){
            mstages.insert(
                std::pair<std::string,xitekStage<REGRESSOR> >(itb->first,
                    xitekStage<REGRESSOR>((itb->second)->toMap())));
        }
	mQhlearning = Qhlearning;
    }

    /**
     * @brief Get the observable values, for a given stage,
     * from all the observations
     * @param stage, the name of the stage
     * @param observations, all the observations
     * @return a tuple that contains observables values
     */
    vv::Tuple getObsTuple(const std::string& stage, const vv::Map& observations)
    {
        return(getXitekStage(stage).getObsTuple(observations));
    }
    /**
     * @brief Get the best action, for a given stage,
     * from observables values
     * @param stage, the name of the stage
     * @param obs, the observables values
     * @return the best action for inputs=obs in stage stage
     */
    unsigned int getBestAction(const std::string& stage, const vv::Tuple& obs)
    {
        return getXitekStage(stage).getBestAction(obs);
    }
    /**
     * @brief Get the number actions for a given stage
     * @param stage the name of the stage
     * @return the number of actions for stage stage
     */
    unsigned int nbActions(const std::string& stage)
    {
        return getXitekStage(stage).nbActions();
    }
    /**
     * @brief Get the number of stages
     * @return the number of stages
     */
    unsigned int nbStages()
    {
        return mstages.size();
    }
    /**
     * @brief Gets the current value of Rho
     * @return the current value of Rho
     */
    double getRho()
    {
        return mrho;
    }
    /**
     * @brief Tells if Qh learning is performed
     * @return if Qh -learning is performed
     */
    bool isQhlearning()
    {
        return mQhlearning;
    }
    /**
     * @brief Constant access to the set of xitekStages
     * @return the set of stages
     */
    const stage_container& stages() const
    {
        return mstages;
    }
    /**
     * @brief Gets the set of xitekStages
     * @return the set of stages
     */
    stage_container& stages()
    {
        return mstages;
    }
    /**
     * @brief Sets the current value of Rho
     * @param the new value of Rho
     */
    void setRho(double rho)
    {
        mrho = rho;
    }
    /**
     * @brief Get the corresponding regressor for stage and act
     * @param stage the name of stage
     * @param act the action index
     * @return the corresponding regressor
     */
    REGRESSOR& getRegressor(const std::string& stage, unsigned int act)
    {
        return mstages[stage].mregressors[act];
    }
    /**
     * @brief Converts the policy into a value
     * @return the value that represents the policy
     */
    virtual vv::Value* toValue() const
    {
        vv::Map* xitekPolicyValue = new vv::Map();
        xitekPolicyValue->addDouble("rho",mrho);
        xitekPolicyValue->addBoolean("Qh-learning",mQhlearning);
        vv::Map& stagesValues = xitekPolicyValue->addMap("stages");
        stage_container_cit itb = mstages.begin();
        stage_container_cit ite = mstages.end();
        for(;itb!=ite;itb++){
            stagesValues.add(itb->first,itb->second.toValue());
        }
        return xitekPolicyValue;
    }
    /**
     * @brief Builds the xitekPolicy from a value
     * @param value, the value that represents the xitekPolicy
     */
    void fromValue(const vv::Value& value)
    {
        mstages.clear();
        const vv::Map& xitekPolicyValue = value.toMap();
        mrho = xitekPolicyValue.getDouble("rho");
        mQhlearning = xitekPolicyValue.getBoolean("Qh-learning");
        const vv::Map& stagesValue = xitekPolicyValue.getMap("stages");
        vv::Map::const_iterator itb = stagesValue.begin();
        vv::Map::const_iterator ite = stagesValue.end();
        for(;itb!=ite;itb++){
            std::pair<stage_container_it,bool> res =
                    mstages.insert(
                        std::pair<std::string,xitekStage<REGRESSOR> >(
                                itb->first,
                                xitekStage<REGRESSOR>()));
            stage_container_it itCont = res.first;
            xitekStage<REGRESSOR>& reg = itCont->second;
            reg.fromValue(*(itb->second));
        }
    }
private :

    /**
     * @brief Gets a xitek stage
     * @param stage, the name of the stage
     * @return the stage
     */
    xitekStage<REGRESSOR>& getXitekStage(const std::string& stage)
    {
        if(mstages.find(stage) == mstages.end()){
            throw vu::ArgError(vle::fmt(_(
                "[xitek] stage '%1%' not found ")) %  stage);
        }
        return mstages.find(stage)->second;
    }
    /**
     * @brief Stores the stages of the xitek policy
     */
    stage_container mstages;
    /**
     * @brief the Rho value for R-learning
     */
    double mrho;
    /**
     * @brief tells if Qh learning is performed
     */
    bool mQhlearning;
};

/**
 * @brief A transition stores information in order to update the policy
 * with the R learning rules
 */
template <class REGRESSOR>
class xitekTransition
{
public:
    /**
     * @brief xitekTransition constructor
     * @param stage1, name of the departure stage
     * @param s1, departure state
     */
    xitekTransition(const std::string& stage1, const vv::Tuple& s1):
        stage1(stage1), s1(s1), d1(0), greedy(false), R1_d1_s1(0), r1(0),
        R1_d1_s1_error(0)
    {
    }
    /**
     * @brief xitekTransition constructor
     * @param value, a value that represeents the trnasition
     */
    xitekTransition(const vv::Value& value)
    {
        fromValue(value);
    }
    /**
     * @brief React from an observation : selects a epsilon greedy decision
     * and computes prediction
     * @param rand, the random number generator
     * @param epsilon, the epsilon parameter
     * @return the chosen decision
     */
    unsigned int select_decision(vu::Rand& rand, const double& epsilon
            , xitekPolicy<REGRESSOR>& pol)
    {
        //chooses the action to perform
        d1 = 0;
        greedy = false;
        if(rand.getDouble() < epsilon){
            greedy = false;
            d1 = rand.getInt(0,pol.nbActions(stage1)-1);
        } else {
            greedy = true;
            d1 = pol.getBestAction(stage1, s1);
        }
        //predicts the R value
        R1_d1_s1 = pol.getRegressor(stage1,d1).predict(s1);
        return d1;
    }
    /**
     * @brief Implementation of the R-learning update rules
     * @param R2_max_s2, the expected maximal value of next R value.
     * @param alpha0, the alpha0 parameter of R learning
     * @param beta, the beta parameter of R learning
     */
    void finalize(double R2_max_s2,
            double alpha0, double beta, xitekPolicy<REGRESSOR>& pol)
    {
        //if pb in prediction (e.g. not in CMAC bounds)
        //no policy update
        if ((R1_d1_s1 == std::numeric_limits<double>::max()) ||
                (R2_max_s2 == std::numeric_limits<double>::max())){
            R1_d1_s1_error = std::numeric_limits<double>::max();
        } else {
            REGRESSOR& R1_d1 = pol.getRegressor(stage1, d1);
            double rho = pol.getRho();
            //update policy using R-learning update rules
            double upValue = (r1 - rho + R2_max_s2);
            R1_d1_s1_error = upValue - R1_d1_s1;
            R1_d1.updateWeight(s1, upValue, alpha0);//TODO
            if(greedy){
                if(pol.isQhlearning()){
                    pol.setRho(0.0);
		} else { 
                    pol.setRho(rho + beta * (R1_d1_s1_error));
                }
            }
        }
    }

    /**
     * @brief Converts the StageExecution into a value
     * @return the value that represents the StageExecution
     */
    vv::Value* toValue() const
    {
        vv::Map* stageExeValue = new vv::Map();
        stageExeValue->addString("stage1",stage1);
        stageExeValue->addInt("d1",d1);
        stageExeValue->addBoolean("greedy",greedy);
        stageExeValue->add("s1",s1.clone());
        stageExeValue->addDouble("r1",r1);
        if(R1_d1_s1 == std::numeric_limits<double>::max()){
            stageExeValue->addString("R1_d1_s1","regError");
            stageExeValue->addString("R1_d1_s1_error","regError");
        } else {
            stageExeValue->addDouble("R1_d1_s1",R1_d1_s1);
            stageExeValue->addDouble("R1_d1_s1_error",R1_d1_s1_error);
        }
        return stageExeValue;
    }
    /**
     * @brief Builds the StageExecution from a value
     * @param value, the value that represents the StageExecution
     */
    void fromValue(const vv::Value& value)
    {
        const vv::Map& stageExeValue = value.toMap();
        const vv::Tuple& observationValue =
                stageExeValue.getTuple("s1");
        vv::TupleValue& s1Vector = s1.value();
        s1Vector.resize(observationValue.size());
        for(unsigned int i=0; i<observationValue.size() ;i++){
            s1Vector[i] = observationValue.at(i);
        }
        stage1 = stageExeValue.getString("stage1");
        d1 = stageExeValue.getInt("d1");
        greedy = stageExeValue.getBoolean("greedy");
        r1 = stageExeValue.getDouble("r1");
        if(stageExeValue.get("R1_d1_s1")->isDouble()){
            R1_d1_s1 = stageExeValue.getDouble("R1_d1_s1");
            R1_d1_s1_error = stageExeValue.getDouble("R1_d1_s1_error");
        } else {
            R1_d1_s1 = std::numeric_limits<double>::max();
            R1_d1_s1_error = std::numeric_limits<double>::max();
        }
    }

    /**
     * The name of the departure stage. Kown when built.
     */
    std::string stage1;
    /**
     * @brief The departure state. Kown when built.
     */
    vv::Tuple s1;
    /**
     * @brief The decision chosen. Computed on select_decision
     */
    unsigned int d1;
    /**
     * @brief Is the decision greedy. Computed on select_decision
     */
    bool greedy;
    /**
     * @brief The prediction R(s1,d1). Computed on select_decision
     */
    double R1_d1_s1;
    /**
     * @brief The received reward. Computed on finalize
     */
    double r1;
    /**
     * @brief The error on prediction. Computed on finalize
     */
    double R1_d1_s1_error;
};



/**
 * xitek Reinforcement learning agent
 * the template class REGRESSOR is a IncrRegressor (see tools)
 */
template <class REGRESSOR>
class xitek_agent : public ci::RLagentMethod
{
public:

    /**
     * @brief xitek_agent RLagent constructor
     * @param init a map that contains initialization parameters for the
     * optimization, note that the same Map is used to initialized the simulator.
     * @param rand, the random number generator
     */
    xitek_agent(const vv::Map& init):
            ci::RLagentMethod(init), mtransitions(), mepisode(0),
            mbeta(1.0), malpha0(1.0), mepsilon(0.1), mmaxEpisodes(6), 
            mcurrAction(), mrand()
    {
        mepisode = init.getInt("currentEpisode");
        if(init.exist("parameters")){
            const vv::Map& params = init.getMap("parameters");
            if(params.exist("beta0")){
                mbeta = params.getDouble("beta0");
            }
            if(params.exist("alpha0")){
                malpha0 = params.getDouble("alpha0");
            }
            if(params.exist("epsilon")){
                mepsilon = params.getDouble("epsilon");
            }
            if(params.exist("maxEpisodes")){
                mmaxEpisodes = params.getInt("maxEpisodes");
            }
            if(mepisode == 1){
                bool qhlearning = false;
        	if(params.exist("Qh-learning")){
                    qhlearning = params.getBoolean("Qh-learning");
                }
                policy().initFromStageSpecifications(
                    params.getMap("stage-actions"),qhlearning);
            }
        }
        //set current beta according the episode
        mbeta = mbeta/(1.0 + log((double)mepisode));
        //build the action action
        mcurrAction.addInt("action",0);
        mcurrAction.addString("stage","");

    }

    /**
     * @brief xitek_agent RLagent destructor
     */
    ~xitek_agent()
    {
        for(unsigned int s = 0; s < mtransitions.size() ; s++){
            delete mtransitions[s];
        }
    }
    /**
     * @brief Answers to the first observation
     * @param observation, first observation of environment
     * @return the first action to perform
     */
    const vv::Value& agent_start(const vv::Value& observation)
    {
        const std::string& stage1 = observation.toMap().getString("stage");
        const vv::Map& obs = observation.toMap().getMap("observables");
        vv::Tuple s1 = policy().getObsTuple(stage1, obs);
        //build a new transition
        mtransitions.push_back(new xitekTransition<REGRESSOR>(stage1,s1));
        xitekTransition<REGRESSOR>& tr1 = getLastTransition();
        unsigned int d1 = tr1.select_decision(mrand,mepsilon, policy());
        mcurrAction.set("action",vv::Integer(d1));
        mcurrAction.set("stage",vv::String(stage1));
        return mcurrAction;
    }

    /**
     * @brief Answers to a step
     * @param r1, the reward of the precedent action
     * @param observation, observation after the precedent action
     * @return the next action to perform
     */
    const vv::Value& agent_step(double r1,
        const vv::Value& observation)
    {
        const std::string& stage2 = observation.toMap().getString("stage");
        const vv::Map& obs = observation.toMap().getMap("observables");
        vv::Tuple s2 = policy().getObsTuple(stage2, obs);
        //update policy
        xitekTransition<REGRESSOR>& tr1 = getLastTransition();
        tr1.r1 = r1;
        //Predict the best expected R2 from the new state s2
        double R2_max_s2 = policy().getRegressor(
                stage2, policy().getBestAction(stage2, s2)).predict(s2);
        tr1.finalize(R2_max_s2, malpha0, mbeta, policy());
        //build a new transition
        mtransitions.push_back(new xitekTransition<REGRESSOR>(stage2,s2));
        xitekTransition<REGRESSOR>& tr2 = getLastTransition();
        unsigned int d2 = tr2.select_decision(mrand,mepsilon, policy());
        mcurrAction.set("action",vv::Integer(d2));
        mcurrAction.set("stage",vv::String(stage2));
        return mcurrAction;
    }

    /**
     * @brief In finite-horizon RL problems, agent can receive a final
     * reward at the end of the episode, no action is required by the
     * environment
     * @param r1, the reward of the precedent action
     */
    void agent_episode_ends(double r1)
    {
        //update policy
        xitekTransition<REGRESSOR>& tr1 = getLastTransition();
        tr1.r1 = r1;
        tr1.finalize(0, malpha0, mbeta, policy());
    }

    /**
     * @brief Tells if the optimization process
     * should stop.
     * @return true if the process has to stop
     */
    virtual bool is_last_episode() const
    {
    	return (mepisode >= mmaxEpisodes);
    }

    /**
     * @brief Observation method
     * @param port, the name of the port to observe
     * @return the observed value
     */
    virtual vv::Value* observe(const std::string& port) const
    {
    	if(port == "episode-traces"){
    		vv::Set* episodeTraces = new vv::Set();
    		for(unsigned int s=0;s<mtransitions.size();s++){
    			const xitekTransition<REGRESSOR>& tr = *(mtransitions[s]);
    			episodeTraces->add(tr.toValue());
    		}
    		return episodeTraces;
    	}
    	if(port == "total-reward"){
    		double totalReward = 0;
    		for(unsigned int s=0;s<mtransitions.size();s++){
    			const xitekTransition<REGRESSOR>& tr = *(mtransitions[s]);
    			totalReward += tr.r1;
    		}
    		return new vv::Double(totalReward);
    	}
    	if(port == "greedy"){
    	    vv::Set* res = new vv::Set();
    	    for(unsigned int s=0;s<mtransitions.size();s++){
    	        const xitekTransition<REGRESSOR>& tr = *(mtransitions[s]);
    	        res->addBoolean(tr.greedy);
    	    }
    	    return res;
    	}
    	if(port == "action"){
    	    vv::Set* res = new vv::Set();
    	    for(unsigned int s=0;s<mtransitions.size();s++){
    	        const xitekTransition<REGRESSOR>& tr = *(mtransitions[s]);
    	        res->addInt(tr.d1);
    	    }
    	    return res;
    	}
        if(port == "rho"){
             return new vv::Double(policy().getRho());
        }
        if(port == "prediction"){
            vv::Set* res = new vv::Set();
            for(unsigned int s=0;s<mtransitions.size();s++){
                const xitekTransition<REGRESSOR>& tr = *(mtransitions[s]);
                if (tr.R1_d1_s1 == std::numeric_limits<double>::max()){
                    res->addString("regError");
                } else {
                    res->addDouble(tr.R1_d1_s1);
                }
            }
            return res;
        }
        if(port == "observation"){
            vv::Set* res = new vv::Set();
            for(unsigned int s=0;s<mtransitions.size();s++){
                const xitekTransition<REGRESSOR>& tr = *(mtransitions[s]);
                res->add(tr.s1);
            }
            return res;
        }
        if(port == "beta"){
            return new vv::Double(mbeta);
        }
        if(port == "policy"){
            return policy().toValue();
        }
        if(boost::starts_with(port,"obs:")){
            std::vector<std::string> splitVect;
            boost::split(splitVect,port,boost::is_any_of(":"));
            if(splitVect.size() != 3){
                return 0;
            }
            unsigned int stage = boost::lexical_cast<unsigned int>(splitVect[1]);
            unsigned int obsIndex = boost::lexical_cast<unsigned int>(splitVect[2]);
            if(mtransitions.size() <= stage){
                return 0;
            }
            const xitekTransition<REGRESSOR>& tr = *(mtransitions[stage]);
            if(tr.s1.size() <= obsIndex){
                return 0;
            }
            return new vv::Double(tr.s1[obsIndex]);
        }
        if(boost::starts_with(port,"predError:")){
            std::vector<std::string> splitVect;
            boost::split(splitVect,port,boost::is_any_of(":"));
            if(splitVect.size() != 2){
                return 0;
            }
            unsigned int stage = boost::lexical_cast<unsigned int>(splitVect[1]);
            if(mtransitions.size() <= stage){
                return 0;
            }
            const xitekTransition<REGRESSOR>& tr = *(mtransitions[stage]);
            if (tr.R1_d1_s1_error == std::numeric_limits<double>::max()){
                return new vv::String("regError");
            } else {
                return new vv::Double(tr.R1_d1_s1_error);
            }
        }
        if(boost::starts_with(port,"pred:")){
            std::vector<std::string> splitVect;
            boost::split(splitVect,port,boost::is_any_of(":"));
            if(splitVect.size() != 2){
                return 0;
            }
            unsigned int stage = boost::lexical_cast<unsigned int>(splitVect[1]);
            if(mtransitions.size() <= stage){
                return 0;
            }
            const xitekTransition<REGRESSOR>& tr = *(mtransitions[stage]);
            if (tr.R1_d1_s1 == std::numeric_limits<double>::max()){
                return new vv::String("regError");
            } else {
                return new vv::Double(tr.R1_d1_s1);
            }
        }
        if(boost::starts_with(port,"act:")){
            std::vector<std::string> splitVect;
            boost::split(splitVect,port,boost::is_any_of(":"));
            if(splitVect.size() != 2){
                return 0;
            }
            unsigned int stage = boost::lexical_cast<unsigned int>(splitVect[1]);
            if(mtransitions.size() <= stage){
                return 0;
            }
            const xitekTransition<REGRESSOR>& tr = *(mtransitions[stage]);
            return new vv::Integer(tr.d1);
        }
        if(boost::starts_with(port,"rew:")){
            std::vector<std::string> splitVect;
            boost::split(splitVect,port,boost::is_any_of(":"));
            if(splitVect.size() != 2){
                return 0;
            }
            unsigned int stage = boost::lexical_cast<unsigned int>(splitVect[1]);
            if(mtransitions.size() <= stage){
                return 0;
            }
            const xitekTransition<REGRESSOR>& tr = *(mtransitions[stage]);
            return new vv::Integer(tr.r1);
        }
        return 0;
	}
    /**
     * @brief Static access to the policy singleton
     * @return the policy singleton
     */
    static xitekPolicy<REGRESSOR>& policy()
    {
    	if (xitek_agent<REGRESSOR>::mpolicy == 0) {
    		xitek_agent<REGRESSOR>::mpolicy = new xitekPolicy<REGRESSOR>();
    	}
    	return *mpolicy;
    }

private:

    /**
     * @brief Get the last transition declared
     * @return the last transition
     */
    xitekTransition<REGRESSOR>& getLastTransition()
    {
        return *(mtransitions[mtransitions.size()-1]);
    }
    /**
     * @brief The execution path
     */
    std::vector<xitekTransition<REGRESSOR>*> mtransitions;
    /**
     * @brief the current episode
     */
    unsigned mepisode;
    /**
     * @brief gamma parameter of Xitek
     */
    double mbeta;
    /**
     * @brief alpha parameter of Xitek
     */
    double malpha0;
    /**
     * @brief Rate of exploration
     */
    double mepsilon;
    /**
     * @brief The maximal number of episodes for the learning process
     */
    unsigned int mmaxEpisodes;
    /**
     * @brief The current action that is performed, a pair stage - action
     */
    vv::Map mcurrAction;
    /**
     * @brief Policy singleton
     */
    static xitekPolicy<REGRESSOR>* mpolicy;
    /**
     * @brief a random number generator
     */
    vu::Rand mrand;

};
}}} //namespaces



#endif
