/**
 * cfo/stat_methods/standard/standard.hpp
 * @author The RECORD Development Team (INRA) and the VLE Development Team
 * See the AUTHORS or Authors.txt file
 */

/*
 * Copyright (C) 2009-2010 INRA
 *
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef CFO_STAT_METHODS_STANDARD_STANDARD_HPP_
#define CFO_STAT_METHODS_STANDARD_STANDARD_HPP_

#include <iostream>

#include <vle/value/Set.hpp>
#include <vle/value/Map.hpp>
#include <vle/value/Double.hpp>
#include <vle/utils/Rand.hpp>

#include <cfo/interfaces/STATmethod.hpp>
#include <cfo/tools/domains/domains.hpp>
#include <cfo/tools/accumulators/accu_mono.hpp>

namespace cfo {
namespace stat_methods {
namespace standard {

namespace ct = cfo::tools;
namespace ci = cfo::interfaces;
namespace vv = vle::value;
namespace vu = vle::utils;

/**
 * @brief Simple statistics
 */
class standard: public ci::STATmethod
{
public:
    /**
     * @brief standard constructor
     */
    standard(const vv::Map& init, ct::VleSimulator& sim) :
        ci::STATmethod(init,sim)
    {
        const vv::Map& parameters = init.getMap("parameters");
        msimulationNbr = parameters.getInt("number-points");
        const vv::Map& domains = parameters.getMap("domains");
        vv::Map::const_iterator itb = domains.begin();
        vv::Map::const_iterator ite = domains.end();
        for(;itb!=ite;itb++){
            const vv::Value& dom = *(itb->second);
            mdomains.addDomain(dom);
        }
    }
    /**
     * @brief standard destructor
     */
    virtual ~standard()
    {
    }
    /**
     * @brief Function to implement by a statistic method.
     */
    virtual void make_stats()
    {
        //set random inputs
        vv::Set inputs(msimulationNbr);
        vv::Set input(mdomains.size());
        for(unsigned int i = 0; i < inputs.size(); i++){
            for(unsigned int d =0; d < mdomains.size(); d++){
                const ct::Domain& domain = mdomains.get(d);
                if (domain.isDiscretDomain()){
                    const ct::DiscretDomain& discretDom =
                            ct::Domain::toDiscretDomain(domain);
                    unsigned int r =  mrand.getInt(0,discretDom.size()-1);
                    input.set(d, discretDom.val(r));
                } else if (domain.isContinuousDomain()){
                    const ct::ContinuousDomain& contDom =
                            ct::Domain::toContinuousDomain(domain);
                    double r = mrand.getDouble(contDom.min(),contDom.max());
                    input.set(d,vv::Double(r));
                }
            }
            inputs.set(i,input);
        }
        //simulate the plan
        const vv::Set& outputs =
                simulator().simulatePlan(inputs, true);
        //conpute statistics on outputs
        for (int s = 0; s < msimulationNbr; s++) {
            double out = outputs.getDouble(s);
            maccumulator.insert(out);
        }
    }
    /**
     * @brief Observation method
     * @param port, the name of the port to observe
     * @return the observed value
     */
    virtual vv::Value* observe(const std::string& port) const
    {
        if(port == "mean"){
            if(maccumulator.count() >= 1){
                return new vv::Double(maccumulator.mean());
            }
        }
        if(port == "stdDeviation"){
            if(maccumulator.count() >= 1){
                return new vv::Double(maccumulator.stdDeviation());
            }
        }
        return 0;
    }
private:
    /**
     * @brief The number of simulations to do.
     */
    int msimulationNbr;
    /**
     * @brief The statistic accumulator
     */
    ct::Accu<ct::MonoDim, ct::MEAN_VAR> maccumulator;
    /**
     * @brief Container of domains
     */
    ct::DomainCont mdomains;
    /**
     *@brief a random number generator
     */
    vu::Rand mrand;

};

}
}
}//namespaces

#endif /* CFO_STAT_METHODS_STANDARD_SIMPLE_HPP_ */
