/**
 * cfo/tools/VleSimulator.hpp
 * @author The RECORD Development Team (INRA) and the VLE Development Team
 * See the AUTHORS or Authors.txt file
 */

/*
 * Copyright (C) 2009-2010 INRA
 *
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef CFO_TOOLS_VLESIMULATOR_HPP_
#define CFO_TOOLS_VLESIMULATOR_HPP_

#include <vle/vpz/Vpz.hpp>
#include <vle/utils/Exception.hpp>
#include <vle/utils/Package.hpp>
#include <vle/utils/Rand.hpp>
#include <vle/manager/Manager.hpp>
#include <vle/manager/Simulation.hpp>
#include <vle/value/Double.hpp>

#include <cfo/tools/VleAPIfacilities.hpp>

namespace cfo {
namespace tools {


namespace vu = vle::utils;
namespace vm = vle::manager;
namespace vz = vle::vpz;
namespace vv = vle::value;
namespace vo = vle::oov;

/**
 * @brief Class that implements an abstract simulator of
 * the simOptim library based on a VLE simulator.
 */
class VleSimulator
{

public:
    /**
     * @brief VleSimulator constructor
     */
    VleSimulator(const vv::Map& initObj) :
        moutputValues(0), mnbthread(1), minputs(), moutputs(), mvpzDyn(),
        mmodules(), mSimpleSimulator(0), mPlanSimulator(0), merror(),
        mrand()
    {
        const vv::Map& model = initObj.getMap("model");
        const std::string& modelpkg = model.getString("package");
        const std::string& modelvpz = model.getString("vpz");
        const vv::Set& inputs = model.getSet("inputs");
        const vv::Set& outputs = model.getSet("outputs");
        for(unsigned int i = 0; i< inputs.size(); i++){
            minputs.push_back(inputs.getString(i));
        }
        for(unsigned int i = 0; i< outputs.size(); i++){
            moutputs.push_back(outputs.getString(i));
        }
        if(initObj.exist("parameters")){
            const vv::Map& params = initObj.getMap("parameters");
            if(params.exist("nb-threads")){
                mnbthread = params.getInt("nb-threads");
            }
        }
        vu::Package pkg(modelpkg);
        std::string vpzFile = pkg.getExpFile(modelvpz, vu::PKG_BINARY);
        mvpzDyn.parseFile(vpzFile);
        //handle views
        VleAPIfacilities::changeAllPlugin(mvpzDyn,"dummy");
        VleAPIfacilities::changePlugin(mvpzDyn,"CallForOptim", "storage");
        mSimpleSimulator = new  vm::Simulation(vm::LOG_NONE,
                vm::SIMULATION_NONE, NULL);
        mPlanSimulator = new vm::Manager(vm::LOG_NONE, vm::SIMULATION_NONE,
                NULL);
    }
    /**
     * @brief VleSimulator destructor
     */
    virtual ~VleSimulator()
    {
        delete moutputValues;
        delete mPlanSimulator;
        delete mSimpleSimulator;
    }

    /**
     * @brief simulate only one input
     * @param input, the input to simulate
     * @return the output of simulation of input
     */
    const vv::Value& simulate(const vv::Value& input)
    {
        delete moutputValues;
        clearAllCondPort();
        postInput(input, false, 1, false);
        vv::Map* res = mSimpleSimulator->run(new vz::Vpz(mvpzDyn),
                mmodules, &merror);
        if (merror.code != 0) {
            throw vle::utils::InternalError(
                    vle::fmt("Error in VleSimulator::simulate '%1%'")
                    % merror.message);
        }
        recoverOutput(*res, 1, 1);
        delete res;
        return *moutputValues;
    }

    /**
     * @brief simulate only one input, but with more than one replica
     * @param input, the input to simulate
     * @param nbrepl, number of replicates
     * @return the output of simulation of input
     */
    const vv::Set& simulate(const vv::Value& input, unsigned int nbrepl)
    {
        delete moutputValues;
        clearAllCondPort();
        postInput(input, false, nbrepl, false);
        vv::Matrix* res = mPlanSimulator->run(new vz::Vpz(mvpzDyn), mmodules,
                mnbthread, 0, 1, &merror);
        if (merror.code != 0) {
            throw vle::utils::InternalError(
                    vle::fmt("Error in VleSimulator::simulate '%1%'")
                    % merror.message);
        }
        recoverOutput(*res, 1, nbrepl);
        delete res;
        return moutputValues->toSet();
    }


    /**
     * @brief Simulates an experiment plan
     * @param inputs, the experiment plan
     * @param commonSeed, should all the combinations be simulated with
     *   the same seed ?
     * @return the set of results
     */
    const vv::Set& simulatePlan(const vv::Set& inputs, bool commonSeed)
    {
        delete moutputValues;
        clearAllCondPort();
        postInput(inputs, true, 1, commonSeed);
        mvpzDyn.project().experiment().setCombination("linear");
        vv::Matrix* res = mPlanSimulator->run(new vz::Vpz(mvpzDyn), mmodules,
                mnbthread, 0, 1, &merror);
        if (merror.code != 0) {
            throw vle::utils::InternalError(
                    vle::fmt("Error in VleSimulator::simulatePlan '%1%'")
                    % merror.message);
        }
        //fill outputs
        recoverOutput(*res, inputs.size(), 1);
        delete res;
        return moutputValues->toSet();
    }

    /**
     * @brief Simulates an experiment plan :
     *   many combinations * mans replicates
     * @param inputs, the experiment plan
     * @param commonSeed, should all the combinations be simulated with
     *   the same seed ?
     * @return the Matrix of results
     */
    const vv::Matrix& simulatePlan(const vv::Set& inputs, unsigned int nbrepl,
            bool commonSeed)
    {
        delete moutputValues;
        clearAllCondPort();
        postInput(inputs, true, nbrepl, commonSeed);
        mvpzDyn.project().experiment().setCombination("linear");
        vv::Matrix* res = mPlanSimulator->run(new vz::Vpz(mvpzDyn), mmodules,
                mnbthread, 0, 1, &merror);
        if (merror.code != 0) {
            throw vle::utils::InternalError(
                    vle::fmt("Error in VleSimulator::simulatePlan '%1%'")
                    % merror.message);
        }
        //fill outputs
        recoverOutput(*res, inputs.size(), nbrepl, true);
        delete res;
        return moutputValues->toMatrix();
    }

    /**
     * @brief Sets a value on a port of a condition of the vpz
     * @param condName, the name of the condition
     * @param portName, the name of the condition port
     * @param value, the value to set on port portName of
     * the condition condName
     */
    void setCondValue(const std::string& condName, const std::string& portName,
    		const vv::Value& value)
    {
    	vz::Conditions& conds = mvpzDyn.project().experiment().conditions();
    	vz::Condition& cond = conds.get(condName);
    	cond.del(portName);
    	cond.add(portName);
    	cond.addValueToPort(portName,value);
    }
    /**
     * @brief Set the number of threads for simulation
     */
    void setNbThreads(unsigned int nbThreads)
    {
        mnbthread = nbThreads;
    }
    /**
     * @brief Get the name of output
     */
    const std::string& getOutputName(unsigned int i)
    {
        return moutputs[i];
    }
    /**
     * @brief Get the name of input
     */
    const std::string& getInputName(unsigned int i)
    {
        return minputs[i];
    }
private:

    void clearAllCondPort()
    {
        vz::Conditions& conds = mvpzDyn.project().experiment().conditions();
        for(unsigned int i = 0; i < minputs.size();i++){
            const std::string& condPort = minputs[i];
            conds.get("CallForOptim").clearValueOfPort(condPort);
        }
    }

    const vz::Vpz& getVpz() const
    {
        return mvpzDyn;
    }

    /**
     * @brief Fills the vpz cond port from an input
     * @param input a vv::Value containing the values of an input
     * @param multi, if true, input is a set representing multiple input values
     * @param nbrepl, number of replicates
     * @param commonSeed, should the combinations be simulated with the same
     * seed ? (only if multi = =true)
     */
    void postInput(const vv::Value& input,  bool multi,
            unsigned nbrepl, bool commonSeed)
    {
        unsigned int nbcomb = 1;
        if (multi) {
            nbcomb = input.toSet().size();
        }
        if ((nbcomb == 1) && (nbrepl == 1)){
            //simple simulation
            if (multi) {
                postOneInput(*input.toSet().get(0));
            } else {
                postOneInput(input);
            }
        } else if ((nbcomb > 1) && (nbrepl == 1)){
            // plans simulation whithout replicates
            const vv::Set& inputs = input.toSet();
            for (unsigned int c=0; c < inputs.size() ; c++) {
                postOneInput(*inputs.get(c));
            }
        } else if ((nbcomb == 1) && (nbrepl > 1)){
            //simulation of one combination with many replicates
            for (unsigned int r=0; r < nbrepl ; r++) {
                postOneInput(input);
                postOneSeed(commonSeed);
            }
        } else if ((nbcomb == 1) && (nbrepl > 1)){
            //simulation of many combinations with many replicates
            const vv::Set& inputs = input.toSet();
            for (unsigned int r=0; r < nbrepl ; r++) {
                bool first = true;
                for (unsigned int c=0; c < inputs.size() ; c++) {
                    const vv::Value& val = *inputs.get(c);
                    postOneInput(val);
                    postOneSeed(commonSeed && !first);
                    first = false;
                }
            }
        }
    }

    /**
     * @brief Fills the vpz cond port from one input dimension
     * @param input a vv::Value containing the values of one input dim
     */
    void postOneInput(const vv::Value& input)
    {
        vz::Conditions& conds = mvpzDyn.project().experiment().conditions();
        vz::Condition& condCFO = conds.get("CallForOptim");
        if (minputs.size() == 1) {
            condCFO.addValueToPort(minputs[0],input.clone());
        } else {
            if(input.isSet()){
                const vv::Set& inset = input.toSet();
                for(unsigned int i =0 ; i < minputs.size(); i ++){
                    condCFO.addValueToPort(minputs[i], inset.get(i)->clone());
                }
            } else if(input.isTuple()) {
                const vv::Tuple& intuple = input.toTuple();
                for(unsigned int i =0 ; i < minputs.size(); i ++){
                    condCFO.addValueToPort(minputs[i],
                            vv::Double(intuple.at(i)));
                }
            } else {
                throw vu::ArgError(vle::fmt(_(
                        "VleSimulator : a set or tuple was expected (got %1%)"))
                % input);
            }
        }
    }

    /**
     * @brief Post one new seed
     * @param commonSeed, should the seed be all the same
     */
    void postOneSeed(bool copyFromPrevious)
    {
        vz::Conditions& conds = mvpzDyn.project().experiment().conditions();
        vz::Condition& condCFO = conds.get("CallForOptim");
        vz::Condition::iterator itb = condCFO.begin();
        vz::Condition::iterator ite = condCFO.end();
        for (; itb != ite; itb++) {
            if (itb->first == "seed") {
                vv::Set* condValues = itb->second;
                if (copyFromPrevious) {
                    condValues->add(*condValues->get(condValues->size()-1));
                } else {
                    condValues->add(vv::Integer(mrand.getInt()));
                }
            }
        }
        if (itb == ite) {
            throw vu::ArgError(vle::fmt(_(
                    "VleSimulator : missing 'seed' port on 'CallForOptim'"
                    "condition ")));
        }
    }

    /**
     * @brief Get outputs from a list of view
     * @param viewList, list of views
     */
    vv::Value* getOutputValues(const vv::Map& viewList)
    {
        vv::Value* tmpRes = 0;
        if (moutputs.size() == 1) {
            tmpRes = VleAPIfacilities::findLastOutputValue(
                    viewList, "CallForOptim", moutputs[0])->clone();
        } else {
            tmpRes = new vv::Set(moutputs.size());
            for(unsigned int i =0 ; i < moutputs.size(); i++){
                const vv::Value* v = VleAPIfacilities::findLastOutputValue(
                        viewList, "CallForOptim", moutputs[i]);
                tmpRes->toSet().set(i, *v);
            }
        }
        return tmpRes;
    }


    /**
     * @brief Fills the output (either the simple vv:Value or the outputPlan
     * depending on the value of nbcomb or nb repl) from observable ports of vle.
     * @param res, a vle simulation result.
     * @param nbcomb, the number of combination of the plan
     * @param nbrepl, the number of replicate
     * @param forceMatrix, forces the output to be built in a matrix,
     * whatever the values of nbcomb and nbrepl
     */
    void recoverOutput(const vv::Value& res,
        unsigned int nbcomb, unsigned int nbrepl, bool forceMatrix = false)
    {
        if ((nbcomb == 1) && (nbrepl == 1) && !forceMatrix){
            //simple simulation
            const vv::Map& viewList = res.toMap();
            moutputValues = getOutputValues(viewList);

        } else if ((nbcomb > 1) && (nbrepl == 1) && !forceMatrix){
            // plans simulation whithout replicates
            moutputValues = new vv::Set(nbcomb);
            const vv::Matrix& simsRes = res.toMatrix();
            for (unsigned int c=0; c < nbcomb ; c++) {
                const vv::Map& viewList = simsRes.get(c,0)->toMap();
                moutputValues->toSet().set(c, getOutputValues(viewList));
            }
        } else if ((nbcomb == 1) && (nbrepl > 1) && !forceMatrix){
            //simulation of one combination with many replicates
            moutputValues = new vv::Set(nbrepl);
            const vv::Matrix& simsRes = res.toMatrix();
            for (unsigned int r=0; r < nbrepl ; r++) {
                const vv::Map& viewList = simsRes.get(r,0)->toMap();
                moutputValues->toSet().set(r, getOutputValues(viewList));
            }
        } else if (((nbcomb > 1) && (nbrepl > 1)) || forceMatrix){
            //simulation of many combinations with many replicates
            moutputValues = new vv::Matrix(nbcomb, nbrepl, 1, 1);
            const vv::Matrix& simsRes = res.toMatrix();
            for (unsigned int r=0; r < nbrepl ; r++) {
                for (unsigned int c=0; c < nbcomb ; c++) {
                    const vv::Map& viewList =
                            simsRes.get(r * nbcomb + c,0)->toMap();
                    moutputValues->toMatrix().set(c, r,
                            getOutputValues(viewList));
                }
            }
        }
    }
    /**
     * @brief Output plan
     */
    vv::Value* moutputValues;
    /**
     *@brief The number of threads for experiment plan simulation
     */
    int mnbthread;
    /**
     *@brief Names of the inputs
     */
    std::vector<std::string> minputs;
    /**
     *@brief Names of the outputs
     */
    std::vector<std::string> moutputs;
    /**
     *@brief vpz object representation, which defines
     *an experiment in vle (see vle)
     */
    vz::Vpz mvpzDyn;
    /**
     * @brief the module manager
     */
    vu::ModuleManager mmodules;
    /**
     * @brief the Simulation engine for simple simulation
     */
    vm::Simulation* mSimpleSimulator;
    /**
     * @brief the Simulation engine for plans
     */
    vm::Manager* mPlanSimulator;
    /**
     * @brief Error structure for simulation
     */
    vm::Error merror;
    /**
     * @brief a random number generator
     */
    vu::Rand mrand;
};

}}//namespaces

#endif
