/**
 * cfo/tools/accumulators/accu_mono.hpp
 * @author The RECORD Development Team (INRA) and the VLE Development Team
 * See the AUTHORS or Authors.txt file
 */

/*
 * Copyright (C) 2009-2010 INRA
 *
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef CFO_TOOLS_ACCUM_MONO_HPP_
#define CFO_TOOLS_ACCUM_MONO_HPP_

#include <algorithm>
#include <vector>
#include <math.h>
#include <functional>
#include <numeric>
#include <limits>

#include<vle/utils/Exception.hpp>

namespace cfo {
namespace tools {

namespace vu = vle::utils;
/**
 * @brief Different types of accumulators.
 */
enum AccuType
{
    MEAN_VAR, QUANTILE, STORE
};

/**
 * @brief MonoDim is the data type handled by
 * one dimensional accumulators
 */
typedef double MonoDim;
/**
 * @brief MultiDim is the data type handled by
 * multi dimensional accumulators
 */
typedef std::vector<double> MultiDim;

/**
 * @brief Common structure to initialize accumulators.
 * Can be used for initializing accumulators with unknown type
 */
struct AccuArgs
{
    /**
     * @brief AccuArgs constructor
     * @param qtl a quantile order
     * @param nbCrit number of dimensions of the multi dimensional
     * accumulator
     */
    AccuArgs(double qtl = 0.5, unsigned int nbCrit = 1) :
        quantileOrder(qtl), nbCriteria(nbCrit)
    {
    }
    /**
     * @brief A quantile order
     */
    double quantileOrder;
    /**
     * @brief The number of dimensions
     */
    unsigned int nbCriteria;
};

/**
 * @brief Accumulator type template declaration
 */
template<class DIM_TYPE, AccuType ACCU_TYPE>
class Accu
{
};

/**
 * @brief Instantiation of one dimensional accumulator MEAN_VAR
 */
template<>
class Accu<MonoDim, MEAN_VAR>
{
private :
    const static double MAX_DOUBLE = 2000000;
    const static double MIN_DOUBLE = -2000000;

public:
    /**
     * @brief Accu constructor
     * @param  args, common accumulator initialization structure
     */
    Accu(const AccuArgs& /*args*/) :
        msum(0), mcount(0), msquareSum(0), mmin(MAX_DOUBLE)
    {
    }
    /**
     * @brief Default Accu constructor
     */
    Accu() :
        msum(0), mcount(0), msquareSum(0), mmin(MAX_DOUBLE)
    {
    }
    /**
     * @brief Inserts a real into the accumulator
     * @param v the real value
     */
    void insert(double v)
    {
        msum += v;
        mcount++;
        msquareSum += pow(v,2);
        if (v < mmin) {
            mmin = v;
        }
    }
    /**
     * @brief Inserts all from an accumulator of the same type
     * @param acc the accumulator to insert
     */
    void insertAcc(const Accu<MonoDim, MEAN_VAR>& acc)
    {
        msum += acc.msum;
        mcount += acc.mcount;
        msquareSum += acc.msquareSum;
        if (acc.mmin < mmin) {
            mmin = acc.mmin;
        }
    }
    /**
     * @brief Mean statistic extractor
     * @return the mean value
     */
    inline double mean() const
    {
        return msum / mcount;
    }
    /**
     * @brief Moment of order 2 extractor :
     * (sum of squares) divided by count
     * @return the moment of order 2 value
     */

    inline double moment2() const
    {
        return msquareSum / mcount;
    }
    /**
     * @brief Deviation Square sum extractor (requires rewriting) :
     * let N = count ; X = mean ; x_i the ieme value inserted.
     * sum_i [(x_i - X) ^ 2] = N * X^2 + sum_i (x_i ^ 2) - 2X sum_i(x_i)
     * @return the deviation square sum
     */
    inline double deviationSquareSum() const
    {
        double mean = msum / mcount;
        return mcount * pow(mean,2) + msquareSum - 2 * mean * msum;
    }
    /**
     * @brief Standad deviation,
     * note: this is not the sample standard deviation
     * @return the stdDeviation
     */
    inline double stdDeviation() const
    {
        return sqrt(deviationSquareSum()/mcount);
    }
    /**
     * @brief Square sum extractor
     * @return suqre sum value
     */
    inline double squareSum() const
    {
        return msquareSum;
    }
    /**
     * @brief Count extractor
     * @return the count value
     */
    inline double count() const
    {
        return mcount;
    }
    /**
     * @brief Sum extractor
     * @return the sum value
     */
    inline double sum() const
    {
        return msum;
    }
    /**
     * @brief Minimal value extractor
     * @return the minimal value
     */
    inline double min() const
    {
        return mmin;
    }
protected:
    /**
     * @brief The sum value
     */
    double msum;
    /**
     * @brief The count value
     */
    double mcount;
    /**
     * @brief The sum of square value
     */
    double msquareSum;
    /**
     * @brief The minimal value
     */
    double mmin;
};

/**
 * @brief Instantiation of one dimensional accumulator QUANTILE
 */
template<>
class Accu<MonoDim, QUANTILE>
{
public:
    /**
     * @brief Accu constructor
     * @param  args, common accumulator initialization structure
     */
    Accu(const AccuArgs& arg) :
        morderedValues(0), msorted(true), mquantileOrder(arg.quantileOrder)
    {
    }
    /**
     * @brief Accu specific constructor
     * @param  quantileOrder the default quantile order
     */
    Accu(double quantileOrder = 0.5) :
        morderedValues(0), msorted(true), mquantileOrder(quantileOrder)
    {
    }
    /**
     * @brief Inserts a real into the accumulator
     * @param v the real value
     */
    void insert(double v)
    {
        morderedValues.push_back(v);
        msorted = false;
    }
    /**
     * @brief Inserts all from an accumulator of the same type
     * @param acc the accumulator to insert
     */
    void insertAcc(const Accu<MonoDim, QUANTILE>& acc)
    {
        std::vector<double>::const_iterator itb = acc.morderedValues.begin();
        std::vector<double>::const_iterator ite = acc.morderedValues.end();
        for (; itb != ite; itb++) {
            morderedValues.push_back(*itb);
        }
        msorted = false;
    }
    /**
     * @brief Quantile value extractor (weigthed average method)
     * @param quantileOrder the quantile order
     * @returns the quantile value of order quantileOrder
     */
    inline double quantile(double quantileOrder)
    {
        if (!msorted) {
            std::sort(morderedValues.begin(),morderedValues.end());
            msorted = true;
        }
        double quantile = 0;
        if (quantileOrder == 1.0) {
            quantile = morderedValues[morderedValues.size() - 1];
        } else {
            double rang_quantile = (morderedValues.size() - 1) * quantileOrder;
            int ent_rang = floor(rang_quantile);
            double frac_rang = rang_quantile - ent_rang;
            quantile = morderedValues[ent_rang] + frac_rang *
                    (morderedValues[ent_rang + 1] - morderedValues[ent_rang]);
        }
        return quantile;
    }
    /**
     * @brief Mean statistic extractor
     * @return the mean value
     */
    inline double mean() const
    {
        double sum = std::accumulate(morderedValues.begin(),morderedValues.end(),0.0);
        return sum / morderedValues.size();
    }
    /**
     * @brief Sum extractor
     * @return the sum value
     */
    inline double sum() const
    {
        return std::accumulate(morderedValues.begin(),morderedValues.end(),0.0);
    }
    /**
     * @brief Count extractor
     * @return the count value
     */
    inline double count() const
    {
        return morderedValues.size();
    }
    /**
     * @brief Square sum extractor
     * @return suqre sum value
     */
    inline double squareSum() const
    {
        double squareSum = 0;
        std::vector<double>::const_iterator itb = morderedValues.begin();
        std::vector<double>::const_iterator ite = morderedValues.end();
        for (; itb != ite; itb++) {
            squareSum += pow(*itb,2);
        }
        return squareSum;
    }
    /**
     * @brief Quantile value extractor
     * @returns the quantile value of order default specified quantile order
     */
    inline double quantile()
    {
        return quantile(mquantileOrder);
    }
private:
    /**
     * @brief The sorted set of accumulated real
     */
    std::vector<double> morderedValues;
    /**
     * @brief Boolean that specifies if the set off values is sorted
     */
    bool msorted;
    /**
     * @bried The default quantile order
     */
    double mquantileOrder;
};

/**
 * @brief Instantiation of one dimensional accumulator STORE
 */
template<>
class Accu<MonoDim, STORE>
{
public:
    /**
     * @brief Accu constructor
     * @param  args, common accumulator initialization structure
     */
    Accu(const AccuArgs& arg) :
        mvalues(0), mquantileOrder(arg.quantileOrder)
    {
    }
    /**
     * @brief Accu specific constructor
     * @param  quantileOrder the default quantile order
     */
    Accu(double quantileOrder = 0.5) :
        mvalues(0), mquantileOrder(quantileOrder)
    {
    }
    /**
     * @brief Inserts a real into the accumulator
     * @param v the real value
     */
    void insert(double v)
    {
        mvalues.push_back(v);
    }
    /**
     * @brief Inserts all from an accumulator of the same type
     * @param acc the accumulator to insert
     */
    void insertAcc(const Accu<MonoDim, STORE>& acc)
    {
        std::vector<double>::const_iterator itb = acc.mvalues.begin();
        std::vector<double>::const_iterator ite = acc.mvalues.end();
        for (; itb != ite; itb++) {
            mvalues.push_back(*itb);
        }
    }
    /**
     * @brief Mean statistic extractor
     * @return the mean value
     */
    inline virtual double mean() const
    {
        double sum = std::accumulate(mvalues.begin(),mvalues.end(),0.0);
        return sum / mvalues.size();
    }
    /**
     * @brief Count extractor
     * @return the count value
     */
    inline unsigned int count() const
    {
        return mvalues.size();
    }
    /**
     * @brief An accumulated real extractor
     * @param i the index in the storage set
     * @return the ith real value
     */
    inline double access(unsigned int i)
    {
        return mvalues[i];
    }
    /**
     * @brief Quantile value extractor (weigthed average method)
     * @param quantileOrder the quantile order
     * @returns the quantile value of order quantileOrder
     */
    inline double quantile(double quantileOrder)
    {
        std::vector<double> ordered(mvalues);
        std::sort(mvalues.begin(),mvalues.end());
        double quantile = 0;
        if (quantileOrder == 1.0) {
            quantile = ordered[mvalues.size() - 1];
        } else {
            double rang_quantile, frac_rang;
            int ent_rang;
            rang_quantile = (mvalues.size() - 1) * quantileOrder;
            ent_rang = floor(rang_quantile);
            frac_rang = rang_quantile - ent_rang;
            quantile = ordered[ent_rang] + frac_rang *
                    (ordered[ent_rang + 1] - ordered[ent_rang]);
        }
        return quantile;
    }
    /**
     * @brief Quantile value extractor
     * @returns the quantile value of order default specified quantile order
     */
    inline double quantile()
    {
        return quantile(mquantileOrder);
    }
private:
    /**
     * @brief The set of accumulated reals
     */
    std::vector<double> mvalues;
    /**
     * The default quantile order
     */
    double mquantileOrder;
};

}
}

#endif /* SOT_ACCUM_MONO_HPP_ */

