/**
 * cfo/tools/accumulators/accu_multi.hpp
 * @author The RECORD Development Team (INRA) and the VLE Development Team
 * See the AUTHORS or Authors.txt file
 */

/*
 * Copyright (C) 2009-2010 INRA
 *
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef CFO_TOOLS_ACCUM_MULTI_HPP_
#define CFO_TOOLS_ACCUM_MULTI_HPP_

#include <algorithm>
#include <vector>
#include <math.h>
#include <functional>
#include <numeric>
#include <limits>

#include <vle/utils/Exception.hpp>
#include <cfo/tools/accumulators/accu_mono.hpp>

namespace vu = vle::utils;

namespace cfo {
namespace tools {

/**
 * @brief Instantiation of multi dimensional accumulator STORE
 */
template<>
class Accu<MultiDim, STORE>
{
public:
    /**
     * @brief Accu constructor
     * @param  args, common accumulator initialization structure
     */
    Accu(const AccuArgs& arg) :
        mstats(arg.nbCriteria,Accu<MonoDim, STORE> (arg))
    {
    }
    /**
     * @brief Accu specific constructor
     * @param  nbCriteria the number of dimensions
     */
    Accu(unsigned int nbCriteria = 5) :
        mstats(nbCriteria)
    {
    }
    /**
     * @brief Inserts multi dimensional data  into the accumulator
     * @param vec the real values
     */
    void insert(const MultiDim& vec)
    {
        unsigned int nbCriteria = vec.size();
        if (mstats.size() != nbCriteria) {
            mstats.resize(nbCriteria);
        }
        for (unsigned int acc = 0; acc < nbCriteria; acc++) {
            mstats[acc].insert(vec[acc]);
        }
    }
    /**
     * @brief Inserts all from an accumulator of the same type
     * @param acc the accumulator to insert
     */
    void insertAcc(const Accu<MultiDim, STORE>& accs)
    {
        unsigned int nbCriteria = mstats.size();
        if (accs.mstats.size() != nbCriteria) {
            throw vle::utils::InternalError(" ");
        }
        for (unsigned int acc = 0; acc < nbCriteria; acc++) {
            mstats[acc].insertAcc(accs.mstats[acc]);
        }
    }
    /**
     * @brief Mean vector extractor
     * @return the mean values
     */
    std::vector<double> mean() const
    {
        std::vector<double> res(mstats.size());
        for (unsigned int i = 0; i < res.size(); i++) {
            res[i] = mstats[i].mean();
        }
        return res;
    }
    /**
     * @brief Quantile vector extractor
     * @param quantileOrder the quantile order
     * @return return the vector of quantile values of order
     * quantileOrder
     */
    std::vector<double> quantile(double quantileOrder)
    {
        std::vector<double> res(mstats.size());
        for (unsigned int i = 0; i < res.size(); i++) {
            res[i] = mstats[i].quantile(quantileOrder);
        }
        return res;
    }
    /**
     * @brief Quantile vector extractor
     * @return return the vector of quantile values of default quantile order
     */
    std::vector<double> quantile()
    {
        std::vector<double> res(mstats.size());
        for (unsigned int i = 0; i < res.size(); i++) {
            res[i] = mstats[i].quantile();
        }
        return res;
    }
    /**
     * @brief Quantile vector extractor
     * @return return the vector of quantile values of default quantile order
     */
    std::vector<double> access(unsigned int index)
    {
        std::vector<double> res(mstats.size());
        for (unsigned int i = 0; i < res.size(); i++) {
            res[i] = mstats[i].access(index);
        }
        return res;
    }
    /**
     * @brief Count extractor
     * @return the number of MultiDim objects accumulated
     */
    inline unsigned int count()
    {
        return mstats[0].count();
    }

private:
    /**
     * @brief the multi dimensional vector of accumulators
     */
    std::vector<Accu<MonoDim, STORE> > mstats;
};

}
}//namespaces

#endif /* SOT_ACCUM_MULTI_HPP_ */

