/**
 * cfo/tools/data/Data.hpp
 * @author The RECORD Development Team (INRA) and the VLE Development Team
 * See the AUTHORS or Authors.txt file
 */

/*
 * Copyright (C) 2009-2010 INRA
 *
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef CFO_TOOLS_DATA_DATA_HPP_
#define CFO_TOOLS_DATA_DATA_HPP_

#include <iostream>
#include <vector>

#include <boost/multi_array.hpp>

#include <vle/utils/Exception.hpp>



namespace cfo {
namespace tools {
namespace data {


/**
 * @brief Class that represents a header, i.e.
 * a property for each variable of
 * data (e.g. name, type, ...)
 */
class Header
{
public:
    /**
     * @brief Header constructor
     */
    Header()
    {
    }
    /**
     * @brief Add a property
     * @param idx, the index of the variable
     * @param prop, the value of the property
     */
    void addProperty(unsigned int idx, const std::string& prop)
    {
        mheader.insert(std::make_pair(idx,prop));
    }
    /**
     * @brief Gives the property of a variable
     * @param idx, the index of the variable
     * @return the property of idx
     */
    const std::string& property(unsigned int idx) const
    {
        std::map<int,std::string>::const_iterator it;
        it = mheader.find(idx);
        return it->second;
    }

private:
    /**
     * @brief An header is a mapping from the variable index
     * to a property
     */
    std::map<int,std::string> mheader;


};

/**
 * @brief Class that represents a 2D matrix of doubles.
 * Columns represent attributes. Lines represent Data.
 */
class Data
{
public:
    /**
     * @brief The 2D container type
     */
    typedef boost::multi_array<double, 2> array_type;
    /**
     * @brief The index type of the container
     */
    typedef array_type::index index;
    /**
     * @brief The range type of the data
     */
    typedef boost::multi_array_types::index_range range;
    /**
     * @brief Default Data constructor.
     */
    Data() :
        mnbLines(0), mnbCols(0), mextents(),  mdata(mextents[1][1])
    {
        mdata[0][0] = 0;
    }
    /**
     *@brief Data destructor
     */
    virtual ~Data()
    {
    }
    /**
     * @brief Set the (line, col) value.
     * @param line the index of the line
     * @param col the index of the column
     * @param the value at (line,col)
     */
    void set(unsigned int line, unsigned int col, double v)
    {
        if ((line >= mnbLines) || (col >= mnbCols)) {
            throw vle::utils::InternalError(
                vle::fmt(" out of range access (%1%,%2%) ; max = (%3%,%4%)")
                % line % col % mnbLines % mnbCols);
        }
        mdata[line][col] = v;
    }
    /**
     * @brief Constant access to the (line, col) data
     * @param line the index of the line
     * @param col the index of the column
     * @return the value at (line,col)
     */
    const double& get(unsigned int line, unsigned int col) const
    {
        if ((line >= mnbLines) || (col >= mnbCols)) {
            throw vle::utils::InternalError(
                " out of range access ");
        }
        return mdata[line][col];
    }
    /**
     * @brief Gives the number of lines
     * @return the number of lines
     */
    inline unsigned int getLines() const
    {
        return mnbLines;
    }
    /**
     * @brief Gives the number of columns.
     * @return the number of columns
     */
    inline unsigned int getCols() const
    {
        return mnbCols;
    }
    /**
     * @brief Set the number of columns
     * @param cols, the number of columns
     */
    void setCols(unsigned int cols)
    {
        mnbCols = cols;
        resize(mnbLines,mnbCols);
    }
    /**
     * @brief Adds an header to the data
     * @param headerName, the name of the header
     * @return the header inserted
     */
    Header& addHeader(const std::string& headerName)
    {
        std::pair<std::map<std::string,Header>::iterator,bool> ret;
        ret = mheaders.insert(std::pair<std::string,Header>(
                headerName,Header()));
        return ret.first->second;
    }
    /**
     * @brief Gets an header
     * @param name, the name of the header
     * @return a reference to the header
     */
    const Header& getHeader(const std::string& name) const
    {
        std::map<std::string,Header>::const_iterator it = mheaders.find(name);
        return it->second;
    }
    /**
     * @brief Adds a line to the data
     */
    void addLine(std::vector<double> values)
    {
        if(values.size() != mnbCols){
            throw vle::utils::InternalError(vle::fmt(
                " unexpected size : %1%, expected : %2% ")
                % values.size() % mnbCols);
        }
        mnbLines ++;
        resize(mnbLines, mnbCols);
        for(unsigned int j = 0; j< values.size() ; j++){
            set(mnbLines-1,j,values[j]);
        }
    }

private:
    /**
     * @brief Proceeds to the resize of the container
     * @param nbLines is the new number of combination.
     * @param nbCols is the new number of replicas.
     */
    void resize(unsigned int nbLines, unsigned int nbCols)
    {
        unsigned int s0 = mdata.size();
        unsigned int s1 = mdata[0].size();
        if ((nbLines > s0) || (nbCols > s1)) {
            mdata.resize(
                mextents[std::max(nbLines,s0)][std::max(nbCols,s1)]);
        }
        mnbLines = nbLines;
        mnbCols = nbCols;
    }
    /**
     * @brief the number of combinations (or inputs of the plan)
     */
    unsigned int mnbLines;
    /**
     * @brief the number of replicas
     */
    unsigned int mnbCols;
    /**
     *@brief Dimensions of the container
     */
    array_type::extent_gen mextents;
    /**
     * @brief The 2D container
     */
    array_type mdata;
    /**
     * @brief Set of headers attached to this data
     */
    std::map<std::string, Header> mheaders;
};
}}}//namepsaces
#endif
