/**
 * cfo/tools/FOmethod.hpp
 * @author The RECORD Development Team (INRA) and the VLE Development Team
 * See the AUTHORS or Authors.txt file
 */

/*
 * Copyright (C) 2009-2010 INRA
 *
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef CFO_TOOLS_doe_doe_HPP_
#define CFO_TOOLS_doe_doe_HPP_

#include <vector>
#include <utility>
#include <cmath>

#include <vle/utils/Exception.hpp>
#include <vle/utils/Rand.hpp>
#include <vle/value/Set.hpp>
#include <vle/value/Tuple.hpp>

namespace cfo {
namespace tools {
namespace doe {


namespace vu = vle::utils;
namespace vv = vle::value;

/**
 * @brief Class that represents a full factorial design of experiment
 */
class FullFactorial
{
public:
    /**
     * @brief FullFactorial constructor
     * @param levels, numbers of levels for each dimension
     */
    FullFactorial(const std::vector<unsigned int>& levels)
    {
        //check
        for(unsigned int l=0; l < levels.size(); l++){
            if(levels[l] == 0){
                throw vu::ArgError("FullFactorial : unexpected levels");
            }
        }
        //fill the doe
        vv::Tuple curr(levels.size(),0);
        unsigned int si = 0;
        fillWith(curr,levels,si);
    }
    /**
     * @brief FullFactorial constructor
     * @param levels, numbers of levels for each dimension
     * @param bounds, bounds for each dimension
     */
    FullFactorial(const std::vector<unsigned int>& levels,
        const vv::Set& bounds)
    {
        //check
        for(unsigned int l=0; l < levels.size(); l++){
            if(levels[l] == 0){
                throw vu::ArgError("FullFactorial : unexpected levels");
            }
        }
        if(bounds.size() != levels.size()){
            throw vu::ArgError("FullFactorial : unexpected bounds");
        }
        //fill the doe
        vv::Tuple curr(levels.size(),0);
        unsigned int si = 0;
        fillWith(curr,levels,si);
        //adapt with the bounds
        for(unsigned int d = 0; d < levels.size(); d++){
            const vv::Tuple& bound = bounds.getTuple(d);
            double min = bound.at(0);
            double max = bound.at(1);
            double nbLevels = levels[d]+1;
            double step = (max - min)/nbLevels;
            for(unsigned int p = 0; p < mpoints.size(); p++){
                vv::Tuple& pt = point(p);
                pt[d] = min + (pt[d] + 1)*step;
            }
        }
    }

    /**
     * @brief FullFactorial destructor
     */
    virtual ~FullFactorial()
    {
        for(unsigned int i =0; i< mpoints.size(); i++){
            delete mpoints[i];
        }
    }

    /**
     * @brief Get the number of points generated
     * @return the number of points generated
     */
    unsigned int size() const
    {
        return mpoints.size();
    }
    /**
     * @brief Constant access to a point
     * @brief index, index of the point in the doe
     */
    const vv::Tuple& point(unsigned int index) const
    {
        return *(mpoints[index]);
    }
    /**
     * @brief Access to a point
     * @brief index, index of the point in the doe
     */
    vv::Tuple& point(unsigned int index)
    {
        return *(mpoints[index]);
    }
    /**
     * @brief Ostream
     */
    friend std::ostream& operator<<(std::ostream& out, const FullFactorial& obj)
    {
        out << " [FullFactorial : " ;
        for(unsigned int i=0;i< obj.mpoints.size();i++){
            out << *(obj.mpoints[i]) << " ; ";
        }
        out << "]";
        return out;
    }

private:
    /**
     * @brief Recursive implementation of the full factorial design
     */
    void fillWith(vv::Tuple& curr, const std::vector<unsigned int>& levels,
        unsigned int si)
    {
        for(unsigned int i=0; i< levels[si]; i++){
            curr[si] = i;
            if(si == levels.size()-1){
                mpoints.push_back(new vv::Tuple(curr));
            } else {
                fillWith(curr,levels,si+1);
            }
        }
    }
    /**
     * @brief The random generator for the points generation
     */
    vu::Rand mrand;
    /**
     * @brief Points of the doe
     */
    std::vector<vv::Tuple*> mpoints;

};

}}}//namespaces

#endif
