/**
 * cfo/tools/domains/domains.hpp
 * @author The RECORD Development Team (INRA) and the VLE Development Team
 * See the AUTHORS or Authors.txt file
 */

/*
 * Copyright (C) 2009-2010 INRA
 *
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef CFO_TOOLS_DOMAINS_DOMAINS_HPP_
#define CFO_TOOLS_DOMAINS_DOMAINS_HPP_

#include <vector>
#include <iostream>
#include <limits>
#include <math.h>

#include <vle/utils/Exception.hpp>

namespace cfo {
namespace tools {

namespace vv = vle::value;

class DiscretDomain;
class ContinuousDomain;


/**
 * @brief Abstract domain
 */
class Domain
{
public:
    /**
     * @brief Domain type
     */
    enum TYPE {DISCRET, CONTINUOUS, DISCRETIZED_CONTINUOUS};
    /**
     * @brief Domain default constructor
     */
    Domain()
    {
    }
    /**
     * @brief Domain default destructor
     */
    virtual ~Domain()
    {
    }
    /**
     * @brief Gets the type of the domain
     * @return the type of the domain
     */
    virtual TYPE getType() const = 0;

    /**
     * @brief Tells if the domain is CONTINUOUS
     * @return true if the domain is CONTINUOUS
     */
    bool isContinuousDomain() const
    {
        return getType() == CONTINUOUS;
    }
    /**
     * @brief Tells if the domain is DISCRET
     * @return true if the domain is DISCRET
     */
    bool isDiscretDomain() const
    {
        return getType() == DISCRET;
    }
    /**
     * @brief Tells if the domain is DISCRETIZED_CONTINUOUS
     * @return true if the domain is DISCRETIZED_CONTINUOUS
     */
    bool isDCDomain() const
    {
        return getType() == DISCRETIZED_CONTINUOUS;
    }
    /**
     * @brief Cast the domain to a DiscretDomain
     * @param the domain to cast
     * @return the resulting DiscretDomain
     */
    static const DiscretDomain& toDiscretDomain(const Domain&);
    /**
     * @brief Cast the domain to a DiscretDomain
     * @param the domain to cast
     * @return the resulting DiscretDomain
     */
    static const ContinuousDomain& toContinuousDomain(const Domain&);
};

/**
 * @brief ContinuousDomain, representation of a continuous domain
 */
class ContinuousDomain : public Domain
{
public:
    /**
     * @brief ContinuousDomain constructor
     */
    ContinuousDomain():
        Domain(), mmin(0), mmax(1)
    {
    };
    /**
     * @brief ContinuousDomain default destructor
     */
    virtual ~ContinuousDomain()
    {
    }
    /**
     * @brief ContinuousDomain constructor
     * @param bounds, a map that contains the bounds of the domain
     */
    ContinuousDomain(const vv::Map& bounds):
        Domain(), mmin(0), mmax(1)
    {
        mmin = bounds.getDouble("min");
        mmax = bounds.getDouble("max");
    };
    /**
     * @brief ContinuousDomain constructor
     * @brief mini, the minimal value of the domain
     * @brief mini, the maximal value of the domain
     */
    ContinuousDomain(double mini, double maxi):
        Domain(), mmin(mini), mmax(maxi)
    {
    };
    /**
     * @brief Gets the minimal value
     */
    double min() const
    {
        return mmin;
    };
    /**
     * @brief Gets the maximal value
     */
    double max() const
    {
        return mmax;
    };
    /**
     *@brief Set the minimal value of the Domain
     *@param mini, the minimal value of the Domain
     */
    void setMin(double mini)
    {
        mmin = mini;
    }
    /**
     *@brief Set the maximal value of the Domain
     *@param maxi, the maximal value of the Domain
     */
    void setMax(double maxi)
    {
        mmax = maxi;
    }
    /**
     * @brief Gets the type of the domain
     * @return the type of the domain
     */
    virtual TYPE getType() const
    {
        return Domain::CONTINUOUS;
    }

protected:
    /**
     * @brief The minimal value of the domain
     */
    double mmin;
    /**
     * @brief The minimal value of the domain
     */
    double mmax;
};

/**
 * @brief DiscretDomain, representation of a discret domain
 */
class DiscretDomain : public Domain
{
public:
    /**
     * @brief DiscretDomain default constructor
     */
    DiscretDomain():
        Domain(), mvals()
    {
    }
    /**
     * @brief DiscretDomain default destructor
     */
    virtual ~DiscretDomain()
    {
    }
    /**
     * @brief DiscretDomain constructor
     * @pbrief the of discrete values
     */
    DiscretDomain(const vv::Set& values):
        Domain(), mvals(values)
    {
    }
    /**
     * @brief Gets one value of the domain
     * @param i the index of the value
     * @return the ith value of the domain
     */
    const vv::Value& val(unsigned int i) const
    {
        return *mvals.get(i);
    }
    /**
     * @brief Gets the size of the domain
     * @return return the size of the domain
     */
    unsigned int size() const
    {
        return mvals.size();
    }
    /**
     * @brief Gets the type of the domain
     * @return the type of the domain
     */
    virtual TYPE getType() const
    {
        return Domain::DISCRET;
    }

    friend std::ostream& operator<<(std::ostream& out, const DiscretDomain& obj)
    {
        out << "(DiscretDomain : " << obj.mvals << " )";
        return out;
    }
private:
    /**
     * @brief The set of discrete values
     */
    vv::Set mvals;
};

/**
 *@brief Object that represents a discretized continuous domain
 */
class DCDomain : public ContinuousDomain
{
public:
    /**
     *@brief DCDomain default constructor
     */
    DCDomain() :
        ContinuousDomain(), min_interval(0)
    {
    }
    /**
     * @brief DCDomain default destructor
     */
    virtual ~DCDomain()
    {
    }
    /**
     * @brief DCDomain copy constructor
     * @param p the parameter to copy
     */
    DCDomain(const DCDomain& p) :
        ContinuousDomain(p.min(), p.max()), min_interval(p.min_interval)
    {
    }
    /**
     * @brief DCDomain copy constructor
     * @param bounds, a map containing bounds
     */
    DCDomain(const vv::Map& bounds) :
        ContinuousDomain(), min_interval(1.0)
    {
        setMin(bounds.getDouble("min"));
        setMax(bounds.getDouble("max"));
        setMinInterval(bounds.getDouble("min-interval"));
    }
    /**
     *@brief DCDomain constructor
     *@param mini the minimum value for this parameter
     *@param maxi the maximum value for this parameter
     *@param step the minimal step for this parameter
     */
    DCDomain(double mini, double maxi, double sstep) :
        ContinuousDomain(mini,maxi), min_interval(sstep)
    {
    }
    /**
     *@brief Gives the extent of this domain, which is the number of
     * steps that the length  of the domain contains
     *@return the extent of the parameter
     */
    double getExtent() const
    {
        return (max() - min()) / this->min_interval;
    }
    /**
     *@brief Gives the minimal interval of the DCDomain
     *@return the minimal interval of the DCDomain
     */
    inline double minInterval() const
    {
        return min_interval;
    }

    /**
     *@brief Set the minimal interval of the DCDomain
     *@param the minimal interval of the DCDomain
     */
    void setMinInterval(double mininterval)
    {
        min_interval = mininterval;
    }

    /**
     * @brief Initialises the parameter
     *@param mini the minimum value for this parameter
     *@param maxi the maximum value for this parameter
     *@param step the minimal step for this parameter
     */
    inline void init(double mini, double maxi, double step)
    {
        mmin = mini;
        mmax = maxi;
        min_interval = step;
    }
    /**
     * @brief Gets the type of the domain
     * @return the type of the domain
     */
    virtual TYPE getType() const
    {
        return Domain::DISCRETIZED_CONTINUOUS;
    }

    friend std::ostream& operator<<(std::ostream& out, const DCDomain& obj)
    {
        out << "(DCDomain : min " << obj.min() << " max " << obj.max()
                << " min-interval " << obj.min_interval << " )";
        return out;
    }

    /**
     * @brief Converts the DCDomain into a value
     * @return the value that represents the DCDomain
     */
    vv::Value* toValue() const
    {
        vv::Map* tileEvalValue = new vv::Map();
        tileEvalValue->addDouble("min",mmin);
        tileEvalValue->addDouble("max",mmax);
        tileEvalValue->addDouble("min-interval",min_interval);
        return tileEvalValue;
    }
    /**
     * @brief Builds the DCDomain from a value
     * @param value, the value that represents the DCDomain
     */
    void fromValue(const vv::Value& value)
    {
        const vv::Map& vmap = value.toMap();
        mmin = vmap.getDouble("min");
        mmax = vmap.getDouble("max");
        min_interval = vmap.getDouble("min-interval");
    }
private:
    /**
     *@brief the minimal interval of the param
     */
    double min_interval;
};

/**
 * @brief DomainCont, a container of polymorphics domains
 */
class DomainCont
{
public:
    /**
     * @brief DomainCont default constructor
     */
    DomainCont():
        mdomains()
    {
    }
    /**
     * @brief DomainCont destructor
     */
    ~DomainCont()
    {
        for(unsigned int i=0; i< mdomains.size(); i++){
            delete mdomains[i];
        }
    }
    /**
     * @brief Gets the number of domains
     * @return the number of domains
     */
    unsigned int size() const
    {
        return mdomains.size();
    }
    /**
     * @brief Gets one Domain
     * @param i, the ith Domain to get
     * @return the ith Domain
     */
    inline const Domain& get(unsigned int i) const
    {
        return *mdomains[i];
    }
    /**
     * @brief Add a domain built from a map
     * @param dom, the map for initalizing the domaimn
     */
    void addDomain(const vv::Value& dom)
    {
        if(dom.isSet()){
            mdomains.push_back(new DiscretDomain(dom.toSet()));
        } else if (dom.isMap()){
            const vv::Map& map = dom.toMap();
            if(map.exist("min-interval")){
                mdomains.push_back(new DCDomain(map));
            } else {
                mdomains.push_back(new ContinuousDomain(map));
            }
        }
    }
private:
    /**
     * @brief the container of the domains
     */
    std::vector<Domain*> mdomains;


};

/**
 *@brief Object representing  a leaf (i.e. a discretized hyperectangle)
 */
class Leaf
{
public:
    /**
     *@brief Leaf constructor
     *@param domains the set of variables parameters
     *@param depth the depth of the leaf (given by the method)
     */
    Leaf(const std::vector<DCDomain>& domains, unsigned int depth = 0) :
        mpotentialSubLeaves(0), mdepth(depth), mdomains(domains)
    {
    }

    /**
     *@brief Leaf destructor
     */
    virtual ~Leaf()
    {
    }
    /**
     *@brief Gives the potential number of sub leaves in this leaf.
     * The sub leaves have extent of 1 for each variable
     *@return the potential number of sub leaves
     */
    inline int getNbPotentialSubLeaves()
    {
        mpotentialSubLeaves = 1.0;
        double extent = 0;
        unsigned int extent_int = 0;
        for (unsigned int i = 0; i < mdomains.size(); i++) {
            extent = mdomains[i].getExtent();
            extent_int = static_cast<unsigned int> (ceil(extent));
            mpotentialSubLeaves *= extent_int;
        }
        return mpotentialSubLeaves - 1;
    }
    /**
     * @brief Gives the variable with the maximal extent
     * @return the index of the variable with the maximal extent
     */
    inline unsigned int getGreatestExtent()
    {
        unsigned int p = 0;
        double largest_extent = 0;
        double extent = 0;
        for (unsigned int i = 0; i < mdomains.size(); i++) {
            extent = mdomains[i].getExtent();
            if (extent > largest_extent) {
                largest_extent = extent;
                p = i;
            }
        }
        return p;
    }
    /**
     * @brief Gives the depth of the leaf
     * @return the depth of the leave
     */
    inline unsigned int getDepth()
    {
        return mdepth;
    }
    /**
     * @brief Constant access to a variable (ie. DCDomain)
     * @param in, index of the parameter
     * @return a reference on a parameter
     */
    inline const DCDomain& getDomain(unsigned int in) const
    {
        return mdomains[in];
    }
    /**
     * @brief Access to a variable (ie. DCDomain)
     * @param in, index of the parameter
     * @return a reference on a parameter
     */
    inline DCDomain& getDomain(unsigned int in)
    {
        return mdomains[in];
    }
    /**
     * @brief Gives the number of parameters
     * @return the number of parameters
     */
    inline unsigned int getInputSize() const
    {
        return mdomains.size();
    }


    friend std::ostream& operator<<(std::ostream& out, const Leaf& obj)
    {
        out << " [Leaf : " ;
        for(unsigned int i=0;i< obj.mdomains.size();i++){
            out << obj.mdomains[i];
        }
        out << "]";
        return out;
    }
private:
    /**
     *@brief The potential number of sub leaves
     */
    unsigned int mpotentialSubLeaves;
    /**
     * @brief The depth of the leaf
     */
    unsigned int mdepth;
    /**
     *@brief domains container
     */
    std::vector<DCDomain> mdomains;
};

const DiscretDomain& Domain::toDiscretDomain(const Domain& domain)
{
    return static_cast<const DiscretDomain&>(domain);
}
const ContinuousDomain& Domain::toContinuousDomain(const Domain& domain)
{
    return static_cast<const ContinuousDomain&>(domain);
}

}
}//namespaces

#endif //CFO_TOOLS_DOMAINS_DOMAINS_HPP_
