/**
 * cfo/tools/policies/q_matrix.hpp
 * @author The RECORD Development Team (INRA) and the VLE Development Team
 * See the AUTHORS or Authors.txt file
 */

/*
 * Copyright (C) 2009-2010 INRA
 *
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef CFO_TOOLS_POLICIES_Q_MATRIX_HPP_
#define CFO_TOOLS_POLICIES_Q_MATRIX_HPP_

#include <vector>
#include <map>
#include <utility>

#include <cfo/tools/VleAPIfacilities.hpp>
#include <cfo/tools/policies/policy.hpp>
//#include <cfo/interfaces/RLexpe.hpp>

namespace cfo {
namespace tools {
namespace policies {

namespace ct = cfo::tools;
//namespace ci = cfo::interfaces;

/**
 * @brief a Q matrix for RL algorithms.
 */
class Qmatrix : public policy
{

public:
    typedef vv::Value action_t;
    typedef vv::Value state_t;
    typedef unsigned int action_id_t;
    typedef unsigned int state_id_t;

    typedef vv::Set action_set_t;
    typedef vv::Set state_set_t;

    typedef vv::Tuple action_values_t;
    typedef vv::Set qmatrix_t;
    typedef vv::Map policy_cont_t;

    typedef action_set_t::const_iterator const_actions_it;
    typedef state_set_t::const_iterator const_states_it;
    typedef qmatrix_t::const_iterator const_qmatrix_it;
    typedef qmatrix_t::iterator qmatrix_it;

    /**
     * @brief Qmatrix constructor
     */
    Qmatrix(): mpolicyCont()
    {
    }
    /**
     * @brief  Initializes the Qmatrix from a set of actions
     * @param  actions the set of actions
     */
    void initFromActions(const action_set_t& actions)
    {
        action_set_t& act = mpolicyCont.addSet("actions");
        mpolicyCont.addSet("states");
        mpolicyCont.addSet("qmatrix");
        const_actions_it itb = actions.begin();
        const_actions_it ite = actions.end();
        for (; itb != ite; itb++) {
            act.add(*(*itb));
        }
        if(act.empty()){
            throw vu::InternalError();
        }
    }


    /**
     * @brief Qmatrix destructor
     */
    ~Qmatrix()
    {
    }

    /**
     * @brief Get the set of actions
     * @return the set of actions
     */
    const action_set_t& getActions() const
    {
        return mpolicyCont.getSet("actions");
    }

    /**
     * @brief Gets the state id of a state value. If not present, the state
     * value and the corresponding state-action values are added to the Qmatrix.
     * @param state a state value
     * @return the state id of the state value
     */
    state_id_t getState(const state_t& state)
    {
        state_set_t& states = mpolicyCont.getSet("states");
        action_set_t& actions = mpolicyCont.getSet("actions");
        qmatrix_t& qmatrix = mpolicyCont.getSet("qmatrix");

        const_states_it isb = states.begin();
        const_states_it ise = states.end();
        state_id_t res = 0;
        for (; isb != ise; isb++) {
            if (ct::VleAPIfacilities::equal(state, *(*isb))) {
                return res;
            }
            res++;
        }
        states.add(state);
        //fill qmatrix
        qmatrix.addTuple(actions.size(),0);
        return res;
    }

    /**
     * @brief The action id of an action value
     * @param act the action value
     * @return the id of action act.
     */
    action_id_t getAction(const action_t& act) const
    {
        const action_set_t& actions = mpolicyCont.getSet("actions");
        const_actions_it iab = actions.begin();
        const_actions_it iae = actions.end();
        action_id_t res = 0;
        for (; iab != iae; iab++) {
            if (ct::VleAPIfacilities::equal(act, *(*iab))) {
                return res;
            }
            res++;
        }
        throw vu::ArgError();
    }
    /**
     * @brief Gets the action value corresponding to an action id.
     * @param act the action id
     * @return the action value with id act.
     */
    const action_t& getAction(const action_id_t& act) const
    {
        return *mpolicyCont.getSet("actions").get(act);
    }
    /**
     * @brief Gets the state value corresponding to a state id.
     * @param sta the state id
     * @return the state value with id sta.
     */
    const state_t& getState(const state_id_t& sta)
    {
        return *mpolicyCont.getSet("states").get(sta);
    }

    /**
     * @brief Updates the Qmatrix according the Sarsa rule.
     * @param currState the current state id
     * @param currAction the action performed in currState
     * @param nextState the next state id
     * @param nextAction id of action to perform in nextState
     * @param reward the reward obtained by performing currAction
     * @param gama the gamma parameter of Sarsa
     * @param alpha the alpha parameter of Sarsa
     */
    void updateSarsa(const state_id_t& currState, const action_id_t& currAction,
        const state_id_t& nextState, const action_id_t& nextAction,
        double reward, double gama, double alpha)
    {
        double& qsaCurr = getQref(currState, currAction);
        double& qsaNext = getQref(nextState, nextAction);
        double newqsa = qsaCurr + alpha * (reward + gama * qsaNext - qsaCurr);
        qsaCurr = newqsa;
    }
    /**
     * @brief Updates the Qmatrix according an ad-hoc rule for the end of the
     * episode. Next state-action does not exist.
     * @param currState the current state id
     * @param currAction the action performed in currState
     * @param reward the reward obtained by performing currAction
     * @param gama the gamma parameter of Sarsa
     * @param alpha the alpha parameter of Sarsa
     */
    void updateSarsaEnd(const state_id_t& currState, const action_id_t& currAction,
        double reward, double alpha)
    {
        double& qsaCurr = getQref(currState, currAction);
        double newqsa = qsaCurr + alpha * (reward - qsaCurr);
        qsaCurr = newqsa;
    }


    /**
     * @brief Get a state-action value (or Qvalue)
     * @param state a state value
     * @param action an action value
     * @return the Q value of (state, action) pair
     */
    double getQ(const state_t& state, const action_t& action)
    {
        const state_id_t& s = getState(state);
        const action_id_t& a = getAction(action);
        return getQ(s,a);
    }

    /**
     * @brief Get a constant reference on a Q value
     * @param state state id
     * @param action action id
     * @return a the Q value of (state,action) pair
     */
    double getQ(const state_id_t& state, const action_id_t& action) const
    {
        const qmatrix_t& qmat = mpolicyCont.getSet("qmatrix");
        return qmat.getTuple(state).at(action);
    }

    /**
     * @brief Get the best action to perfom in a state
     * @param state the state id
     * @return the id of the best action to perform
     */
    action_id_t getBestAction(const state_id_t& state) const
    {
        const action_set_t& actions = mpolicyCont.getSet("actions");
        const_actions_it iab = actions.begin();
        const_actions_it iae = actions.end();
        action_id_t best = -1;
        action_id_t curr = 0;
        double resBest = -10000000;
        for (; iab != iae; iab++) {
            double v = getQ(state,curr);
            if(v > resBest){
                best = curr;
                resBest = v;
            }
            curr++;
        }
        return best;
    }

    /**
     * @brief Builds a representation of the policy where the best actions
     * are given (an not Q values).
     * @return a map containing the (real) policy computed from the policy
     */
    vv::Set* buildPolicyRepresentation() const
    {
        vv::Set* res = new vv::Set();
        const state_set_t& states = mpolicyCont.getSet("states");
        for (unsigned int s = 0; s < states.size(); s++) {
            const state_t& sta = *states.get(s);
            vv::Map* staAct = new vv::Map();
            staAct->add("state",sta);
            staAct->add("best-action",getAction(getBestAction(s)));
            res->add(staAct);
        }
        return res;
    }

private:
    /**
     * @brief Get a reference on a Q value
     * @param state state id
     * @param action action id
     * @return a reference on the Q value of (state,action) pair
     */
    double& getQref(const state_id_t& state, const action_id_t& action)
    {
        qmatrix_t& qmat = mpolicyCont.getSet("qmatrix");
        return qmat.getTuple(state).at(action);
    }


    /**
     * @brief The policy container (a map containing the set of states, the set
     * of actions and the qmatrix)
     */
    policy_cont_t mpolicyCont;


};

}}} // namespace cfo::tools::policies

#endif /* CFO_TOOLS_POLICIES_Q_MATRIX_HPP_ */
