/**
 * cfo/tools/regression/CMAC.hpp
 * @author The RECORD Development Team (INRA) and the VLE Development Team
 * See the AUTHORS or Authors.txt file
 */

/*
 * Copyright (C) 2009-2010 INRA
 *
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef CFO_TOOLS_REGRESSION_CMAC_HPP_
#define CFO_TOOLS_REGRESSION_CMAC_HPP_

#include <vector>
#include <numeric>
#include <math.h>

#include <boost/multi_array.hpp>

#include <vle/value/Set.hpp>
#include <vle/value/Map.hpp>
#include <vle/value/Tuple.hpp>
#include <vle/utils/Exception.hpp>

#include <cfo/tools/domains/domains.hpp>
#include <cfo/tools/regression/IncrRegressor.hpp>



namespace cfo {
namespace tools {
namespace regression {

namespace vu = vle::utils;
namespace vv = vle::value;
namespace ct = cfo::tools;

/**
 * @brief Grid objects stored in a CMAC
 */
class Grid
{
private:
    /**
     * @brief Tile identifier. A tile is identified by
     * the indexes (over all dimensions) in the grid.
     */
    struct TileId
    {
        /**
         * @brief TileId constructor
         * @param size, the number of dimensions
         */
        TileId(unsigned int size):
            index(size,0)
        {
        }
        /**
         * @brief TileId constructor
         * @param setVal, the set of indexes
         */
        TileId(const vv::Set& setVal)
        {
            fromValue(setVal);
        }
        /**
         * @brief Converts the TileId into a value
         * @return the value that represents the TileId
         */
        virtual vv::Value* toValue() const
        {
            vv::Set* tileIdValue = new vv::Set();
            for(unsigned int i = 0; i < index.size() ; i++){
                tileIdValue->addInt(index[i]);
            }
            return tileIdValue;
        }
        /**
         * @brief Builds the TileId from a value
         * @param value, the value that represents the TileId
         */
        void fromValue(const vv::Value& value)
        {
            const vv::Set& vset = value.toSet();
            index.resize(vset.size());
            for(unsigned int i=0; i< vset.size(); i++){
                index[i] = vset.getInt(i);
            }
        }
        /**
         * @brief Comparison operator
         * @param b, another TileId
         * @return true if "this" is inferior to b
         */
        bool operator <(const TileId& b) const
        {
            if(index.size() != b.index.size()){
                throw vu::ArgError(vle::fmt(_(
                    "[CMAC] unexpeced TileId size (got '%1%', expected '%2%')"))
                % b.index.size() % index.size());
            }
            for(unsigned int dim = 0; dim < index.size(); dim++){
                if(index[dim] != b.index[dim]){
                    return (index[dim] < b.index[dim]);
                }
            }
            return false;
        }
        /**
         * @brief Indexes for each dimension
         */
        std::vector<unsigned int> index;
    };
    /**
     * @brief Tile evaluation.
     * Stores the evaluation of one Tile
     */
    struct TileEval
    {
        /**
         * @brief TileEval default constructor
         */
        TileEval():
            numCall(0),weight(0)
        {
        }
        /**
         * @brief TileEval constructor
         * @param m, the map that contains TileEval members values
         */
        TileEval(const vv::Map& m)
        {
            fromValue(m);
        }
        /**
         * @brief Converts the TileEval into a value
         * @return the value that represents the TileEval
         */
        virtual vv::Value* toValue() const
        {
            vv::Map* tileEvalValue = new vv::Map();
            tileEvalValue->addInt("numCall",numCall);
            tileEvalValue->addDouble("weight",weight);
            return tileEvalValue;
        }
        /**
         * @brief Builds the TileEval from a value
         * @param value, the value that represents the TileEval
         */
        void fromValue(const vv::Value& value)
        {
            const vv::Map& vmap = value.toMap();
            numCall = vmap.getInt("numCall");
            weight = vmap.getDouble("weight");
        }
        /**
         * @brief the number of calls into this tile
         */
        unsigned int numCall;
        /**
         * @brief the weight parameter (for regression) of the tile
         */
        double weight;
    };
    /**
     * @brief Get the corresponding tile iterator for a point
     * @brief point, the point to search
     * @return the Tile iterator into which the point falls,
     * if not found, the tile is added into the map of tiles.
     */
    std::map<TileId,TileEval>::iterator tile(const vv::Tuple& point)
    {
        if(point.size() != mintervals.size()){
            throw vu::ArgError(vle::fmt(_(
                "[CMAC] unexpected point size (got '%1%', expected '%2%')"))
            % point.size() % mintervals.size());
        }
        TileId tile(mintervals.size());
        for(unsigned int dim = 0 ; dim < mintervals.size() ; dim ++){
            const ct::DCDomain& param = mintervals[dim];
            double min = param.min();
            double max = param.max();
            double tileSize = param.minInterval();
            double val = point.at(dim);
            if((val < min) || (val >= max)){
//                std::cout << (vle::fmt(_(
//                    "[CMAC]  DBG out of bounds ('%1%' should be into {%2%,%3%})"))
//                % val % min % max)
//                << std::endl;
                return mtiles.end();
            }
            unsigned int id = 0;
            bool found = false;
            while(not found){
                if((min + ((double) id)* tileSize) <= val ){
                    if(val < (min + ((double) id +1) * tileSize)){
                        found = true;
                    }
                }
                id++;
            }
            tile.index[dim] = id;
        }
        TileIterator it = mtiles.find(tile);
        if(it == mtiles.end()){
            std::pair<TileIterator,bool> r = mtiles.insert(
                    std::pair<TileId,TileEval>(tile,TileEval()));
            if(!r.second){
                throw "unexpected error";
            }
            it = r.first;
        }
        return it;
    }

    /**
     * @brief The tiles container (id and evaluation)
     */
    std::map<TileId, TileEval> mtiles;
    /**
     * @brief Domains of each dimension
     */
    std::vector<ct::DCDomain> mintervals;
    typedef std::map<TileId,TileEval>::iterator TileIterator;
    typedef std::map<TileId,TileEval>::const_iterator TileIteratorConst;
public:
    /**
     * @brief Grid constructor
     * @param intervals, intervals for each dimension
     */
    Grid(const std::vector<ct::DCDomain>& intervals):
        mtiles(), mintervals(intervals)
    {
    }
    /**
     * @brief Grid constructor
     */
    Grid(const vv::Map& m)
    {
        fromValue(m);
    }
    /**
     * @brief Gets the weight for a given point
     * @param pt, the point
     * @return the weight of point point or infinity if not in bounds
     */
    double getWeight(const vv::Tuple& pt)
    {
        TileIterator it = tile(pt);
        if (it == mtiles.end()) {
            return std::numeric_limits<double>::max();
        } else {
            return it->second.weight;
        }
    }
    /**
     * @brief Gets the number of calls
     * in the tile that contains a point
     * @param point the point under study
     * @return the number of calls in the tile that contains point
     * or infinity if not in bounds
     */
    unsigned int getNbCall(const vv::Tuple& point)
    {
        TileIterator it = tile(point);
        if (it == mtiles.end()) {
            return std::numeric_limits<unsigned int>::max();
        } else {
            TileEval& tileEval = it->second;
            return tileEval.numCall;
        }
    }
    /**
     * @brief Updates the weight of the tile that matches the point based on
     * the difference between observed value and predicted value,
     * according to an update rate parameter
     * @param point the point considered
     * @param diff the difference between the observed and predicted value at
     * point point
     * @param alpha0 the initial update rate parameter
     *  @return true if the update is a success
     */
    bool updateWeight(const vv::Tuple& point, double diff, double alpha0)
    {
        TileIterator it = tile(point);
        if(it == mtiles.end()){
            return false;
        }
        TileEval& tileEval = it->second;
        tileEval.numCall += 1;
        tileEval.weight +=
                (alpha0/(1.0 + log ((double)tileEval.numCall))) * diff;
        return true;
    }
    /**
     * @brief Converts the Grid and its parameters into a value
     * @return the value that represents the grid
     */
    vv::Value* toValue() const
    {
        vv::Map* gridValue = new vv::Map();
        vv::Set& doms = gridValue->addSet("domains");
        for(unsigned int d=0; d<mintervals.size(); d++){
            doms.add(mintervals[d].toValue());
        }
        vv::Set& tiles = gridValue->addSet("tiles");
        TileIteratorConst itb = mtiles.begin();
        TileIteratorConst ite = mtiles.end();
        for(;itb!=ite;itb++){
            vv::Map* tileValue = new vv::Map();
            tileValue->add("TileId",itb->first.toValue());
            tileValue->add("TileEval",itb->second.toValue());
            tiles.add(tileValue);
        }
        return gridValue;
    }
    /**
     * @brief Builds the Grid from a value
     * @param value, the value that represents the Grid
     */
    void fromValue(const vv::Value& value)
    {
        mintervals.clear();
        mtiles.clear();
        const vv::Map& vmap = value.toMap();
        const vv::Set& doms = vmap.getSet("domains");
        for(unsigned int d=0; d<doms.size(); d++){
            mintervals.push_back(ct::DCDomain(doms.getMap(d)));
        }
        const vv::Set& tiles = vmap.getSet("tiles");
        for(unsigned int t=0; t<tiles.size(); t++){
            const vv::Map& tileI = tiles.getMap(t);
            mtiles.insert(std::pair<TileId,TileEval>(
                    TileId(tileI.getSet("TileId")),
                    TileEval(tileI.getMap("TileEval"))));
        }
    }
};


/**
 * @brief a CMAC regression tool
 */
class CMAC : public IncrRegressor
{
public:
    /**
     * @brief CMAC constructor
     */
    CMAC()
    {
    }
    /**
     * @brief assignement operator
     */
    CMAC& operator=(const CMAC& o)
    {
        vv::Value* CMACvalue = o.toValue();
        fromValue(*CMACvalue);
        delete CMACvalue;
        return *this;
    }
    /**
     * @brief Initialization of the CMAC regressor
     * @param parameters the parameters of the method.
     * it is a map that must contain :
     *  - an integer value "nb-grids" that gives the number of grids
     *  the tile size for shifting grids
     *  - an integer value "nb-tiles" that gives the number of tiles
     *  for all dimension
     *  - a set of tuple "bounds" that gives the set of
     *  minimal and maximal values of each input
     */
    void init(const vv::Value& parameters)
    {
        const vv::Map& params = parameters.toMap();
        const vv::Set& boundset = params.getSet("bounds");
        unsigned int nbGrids = params.getInt("nb-grids");
        mnbTiles = params.getInt("nb-tiles");
        //initialisation of bounds
        for(unsigned int d=0; d< boundset.size(); d++){
            const vv::Tuple& bounds = boundset.getTuple(d);
            mbounds.add(bounds.clone());
        }
        //initalisation of grids
        std::vector<ct::DCDomain> gridInit(mbounds.size());
        for (unsigned int g = 0; g < nbGrids; g++){
            for (unsigned int d = 0; d < mbounds.size(); d++){
                const vv::Tuple& bounds = mbounds.getTuple(d);
                ct::DCDomain& param = gridInit[d];
                double tileSize =  (bounds.at(1) - bounds.at(0))/
                        ((double) mnbTiles);
                double shift = tileSize * (g/ ((double) nbGrids));
                param.setMin(bounds.at(0) - shift);
                param.setMax(bounds.at(1) - shift + tileSize);
                param.setMinInterval(tileSize);
            }
            mgrids.push_back(Grid(gridInit));
        }
    }

    /**
     * @brief Gets the number of dimensions of the problem
     * @return the number of dimensions
     */
    unsigned int getDimension() const
    {
        return mbounds.size();
    }
    /**
     * @brief Gets the bounds of dimensions for this regressor
     * @return the bounds of the CMAC regressor
     */
    const vv::Set& bounds() const
    {
        return mbounds;
    }
    /**
     * @brief Gets the number of tiles by dimension
     * @return the number of tiles
     */
    unsigned int nbTiles() const
    {
        return mnbTiles;
    }
    /**
     * @brief Predicts the output value at a point
     * @param point the input value
     * @return the predicted value or infinity if not in bounds
     */
    double predict(const vv::Tuple& point)
    {
        double res = 0;
        std::vector<Grid>::iterator itb = mgrids.begin();
        std::vector<Grid>::iterator ite = mgrids.end();
        for(;itb!=ite;itb++){
            double  w = itb->getWeight(point);
            if (w == std::numeric_limits<double>:: max()){
                return w;
            } else {
                res += w;
            }
        }
        return res/(double)mgrids.size();
    }
    /**
     * @brief Gets the mean number of calls
     * in the neighborough of a point
     * @param point the point under study
     * @return the mean number of calls in the neighborough
     * of point
     */
    double getMeanNbCall(const vv::Tuple& point)
    {
        double res = 0;
        std::vector<Grid>::iterator itb = mgrids.begin();
        std::vector<Grid>::iterator ite = mgrids.end();
        for(;itb!=ite;itb++){
            unsigned int n = itb->getNbCall(point);
            if(n == std::numeric_limits<unsigned int>::max()){
                return std::numeric_limits<double>::max();
            } else {
                res += n;
            }
        }
        return ((double)res)/((double) mnbTiles);
    }
    /**
     * @brief Updates the weights of all tills containing a given point
     * accordning an update rate parameter
     * @param point the point considered
     * @param upValue the observed value at point point
     * @param alpha0 the initial update rate parameter
     * @return true if the update is a success
     */
    bool updateWeight( const vv::Tuple& point, double upValue, double alpha0)
    {
        double diff = upValue - predict(point);
        std::vector<Grid>::iterator itb = mgrids.begin();
        std::vector<Grid>::iterator ite = mgrids.end();
        bool up = true;
        for(;itb!=ite;itb++){
            up = itb->updateWeight(point, diff, alpha0);
            if(!up){
                return false;
            }
        }
        return true;
    }

    /**
     * @brief Converts the CMAC regressor and its parameters into a value
     * @return the value that represents the regressor
     */
    vv::Value* toValue() const
    {
        vv::Map* cmac = new vv::Map();
        vv::Set& regValue = cmac->addSet("grids");
        for(unsigned int i = 0; i < mgrids.size(); i++){
            regValue.add(mgrids[i].toValue());
        }
        cmac->add("bounds",mbounds.clone());
        cmac->addInt("nb-tiles",mnbTiles);
        return cmac;
    }
    /**
     * @brief Builds the CMAC regressor and its parameters from a value
     * @param v, the value that represents the regressor
     */
    void fromValue(const vv::Value& v)
    {
        mgrids.clear();
        mbounds.clear();
        const vv::Map& cmac = v.toMap();
        const vv::Set& regValue = cmac.getSet("grids");
        for(unsigned int i = 0; i < regValue.size(); i++){
            mgrids.push_back(Grid(regValue.getMap(i)));
        }
        const vv::Set& boundsValue = cmac.getSet("bounds");
        for(unsigned int d=0; d< boundsValue.size(); d++){
            const vv::Tuple& bounds = boundsValue.getTuple(d);
            mbounds.add(bounds.clone());
        }
        mnbTiles = cmac.getInt("nb-tiles");

    }

private:
    /**
     * @brief Set of grids of the CMAC regressor
     */
    std::vector<Grid> mgrids;
    /**
     * @brief Bounds for each dimension
     */
    vv::Set mbounds;
    /**
     * @brief Number of Tiles by dimension
     */
    unsigned int mnbTiles;

};

}}}//namespaces

#endif /* CFO_TOOLS_REGRESSION_CMAC_HPP_ */
