/**
 * @file cfo/examples/ExMinesEnvironment.cpp
 * @author The RECORD Development Team (INRA)
 */

/*
 * Copyright (C) 2009 INRA
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */



#include <vle/devs/DynamicsDbg.hpp>
#include <vle/utils/DateTime.hpp>
#include <vle/utils/Rand.hpp>
#include <vle/value/Matrix.hpp>
#include <vle/extension/fsa/Statechart.hpp>

namespace ve = vle::extension;
namespace vd = vle::devs;
namespace vv = vle::value;
namespace vu = vle::utils;


namespace cfo { namespace examples {

/**
 * @brief ExMinesEnvironment is an example for reinforcement learning.
 * This example is proposed by the developpers of the library RL-Glue.
 * It simulates the move of an agent on a 2D grid. A cell can be either a mine
 * (thus a very low reward -100 is given to the agent), the goal (a high reward
 *  +1000 is given), the border (a reward of -2) or nothing especially
 *  (a reward of -1).
 *
 *
 *  @article{rl-glue,
 *   Author = {Brian Tanner and Adam White},
 *   Date-Added = {2009-05-06 22:44:59 -0600},
 *   Date-Modified = {2009-10-06 13:44:04 -0600},
 *   Journal = {Journal of Machine Learning Research},
 *   Month = {September},
 *   Pages = {2133--2136},
 *   Title = {{RL}-{G}lue : Language-Independent Software for Reinforcement-Learning Experiments},
 *   Volume = {10},
 *   Year = {2009}}
 *
 *
 */
class ExMinesEnvironment: public ve::fsa::Statechart {
public:
    /**
     * @brief ExMinesEnvironment state-chart constructor
     */
    ExMinesEnvironment(const vd::DynamicsInit& init, const vd::InitEventList& events) :
        ve::fsa::Statechart(init, events), mworld(events.getMatrix("world")),
        magentRequireAnswer(false)
    {
        states(this) << INIT << ASLEEP <<  PROCESS << FINISH;
        initialState(INIT);

        //initializes the spatial state of the agent
        bool initDone = false;
        unsigned int x = 0;
        unsigned int y = 0;
        while(not initDone){
            x = mrand.getInt(0,(int) mworld.rows() - 1);
            y = mrand.getInt(0,(int) mworld.columns() - 1);
            initDone = (mworld.getInt(y,x) == NOTHING);
        }
        magentState.addInt("x",x);
        magentState.addInt("y",y);
        magentState.addInt("type",NOTHING);

        state(this, ASLEEP)
                << eventInState("agent-action", &ExMinesEnvironment::agentAction);

        transition(this, INIT, ASLEEP)
                << after(0.)
                << send(&ExMinesEnvironment::agentInit);
        transition(this, ASLEEP, PROCESS)
                << guard(&ExMinesEnvironment::agentRequireAnswer);
        transition(this, ASLEEP, FINISH)
                << guard(&ExMinesEnvironment::finish);
        transition(this, PROCESS, ASLEEP)
                << after(1.0)
                << send(&ExMinesEnvironment::agentAnswer)
                << action(&ExMinesEnvironment::agentAnswered);
    }

    /**
     * @brief An action (between N,E,S,W) has been sent by the agent.
     * This modifies the location of the agent.
     */
    void agentAction(const vle::devs::Time& /*time*/,
        const vle::devs::ExternalEvent* event)
    {
        if(event->onPort("agent-action")){
            vv::Integer& x = magentState.get("x")->toInteger();
            vv::Integer& y = magentState.get("y")->toInteger();
            switch(event->getIntegerAttributeValue("value")){
            case W:{
                if(y.value() > 0){
                    y.set(y.value()-1);
                }
                break;
            }case E:{
                if(y.value() < (int) mworld.columns() - 1){
                    y.set(y.value()+1);
                }
                break;
            }case N:{
                if(x.value() > 0){
                    x.set(x.value()-1);
                }
                break;
            }case S:{
                if(x.value() < (int) mworld.rows() - 1){
                    x.set(x.value() + 1);
                }
                break;
            }default:{
                break;
            }
            }
            magentState.get("type")->toInteger().set(
                mworld.getInt(y.value(),x.value()));
            magentRequireAnswer = true;
        }
    }

    /**
     * @brief Should the agent be answered (observation + reward)
     */
    void agentAnswered(const vd::Time& /* time */)
    {
        magentRequireAnswer = false;
    }

    /**
     * @brief Send an initialization for the agent
     * (the initial observation, ie. location)
     */
    void agentInit(const vle::devs::Time& /*time*/,
        vle::devs::ExternalEventList& output) const
    {
        vd::ExternalEvent* evt = new vd::ExternalEvent("info-env");
        evt->putAttribute("observation",magentState.clone());
        output.push_back(evt);
    };

    /**
     * @brief Send an answer (observation + reward) for the agent
     */
    void agentAnswer(const vle::devs::Time& /*time*/,
        vle::devs::ExternalEventList& output) const
    {
        vd::ExternalEvent* evt = new vd::ExternalEvent("info-env");
        evt->putAttribute("reward",new vv::Double(rewardState()));
        if(not finishState()){
            evt->putAttribute("observation",magentState.clone());
        }
        output.push_back(evt);
    };

    /**
     * @brief Does the simulation should stop.
     * Yes if the current location is a mine or the goal.
     */
    bool finish(const vd::Time& /* time */)
    {
        return finishState() && (not magentRequireAnswer);
    }
    /**
     * @brief Tells that a couple observation + reward has been computed and
     * that it should be sent for the agent.
     */
    bool agentRequireAnswer(const vd::Time& /* time */)
    {
        return magentRequireAnswer;
    }
private:
    enum FsaState { INIT, ASLEEP, PROCESS, FINISH };
    enum SpatialStateType {NOTHING, BORDER, MINE, GOAL};
    enum AgentAction {N,E,S,W};
    bool startState() const
    {
        return (magentState.getInt("type") == NOTHING);
    }
    bool finishState() const
    {
        return((magentState.getInt("type") == MINE) or
                (magentState.getInt("type") == GOAL));
    }
    double rewardState() const
    {
        switch(magentState.getInt("type")){
        case NOTHING:{
            return -1;
            break;
        }case BORDER:{
            return -10;
            break;
        }case MINE:{
            return -20;
            break;
        }case GOAL:{
            return 20;
            break;
        }default :{
            throw vu::ArgError();
        }
        }
    }
    vv::Matrix mworld;
    vv::Map magentState;
    bool magentRequireAnswer;
    vu::Rand mrand;
};
}}//namespace cfo::examples

DECLARE_DYNAMICS_DBG(cfo::examples::ExMinesEnvironment)


