/**
 * @file test/cfo/fo_methods/p2/tests_p2.cpp
 * @author The RECORD Development Team (INRA) and the VLE Development Team
 * See the AUTHORS or Authors.txt file
 */

/*
 * Copyright (C) 2009-2010 INRA
 *
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#define BOOST_TEST_MAIN
#define BOOST_AUTO_TEST_MAIN
#define BOOST_TEST_DYN_LINK
#define BOOST_TEST_MODULE tests_cfo
#include <boost/test/unit_test.hpp>
#include <boost/test/auto_unit_test.hpp>
#include <boost/test/floating_point_comparison.hpp>
#include <boost/lexical_cast.hpp>
#include <stdexcept>
#include <limits>
#include <fstream>
#include <iostream>
#include <vector>

#include <vle/utils/Package.hpp>

#include <cfo/fo_methods/p2/p2.hpp>
#include <cfo/tools.hpp>

namespace mp = cfo::fo_methods::p2;
namespace ct = cfo::tools;
namespace vz = vle::vpz;
namespace vm = vle::manager;


template<ct::AccuType STATSTRAT,
mp::LeafPerformanceType LEAF_PERF = mp::LEAF_PERF_MEAN,
mp::LeafDispersionType LEAF_DISP = mp::LEAF_DISP_SPATIAL_VAR1>
struct p2t1
{
    typedef mp::StratP2<ct::MonoDim, STATSTRAT> strat_type;
    typedef mp::LeafP2<strat_type, LEAF_PERF, LEAF_DISP> leaf_type;
    typedef mp::AccuFactory<leaf_type> factory_type;
    typedef boost::shared_ptr<factory_type> factory_type_ptr;
    typedef boost::shared_ptr<strat_type> strat_type_ptr;
    typedef boost::shared_ptr<leaf_type> leaf_type_ptr;

    mp::P2args args;

    unsigned int nbOutputs;

    p2t1() :
        args(), nbOutputs(1)
    {
        args.leaf_params.push_back(ct::DCDomain(-10,10,1));
        args.leaf_params.push_back(ct::DCDomain(-11,11,1));
        args.leaf_params.push_back(ct::DCDomain(-12,12,1));
    }

    leaf_type_ptr getTestLeaf()
    {

        leaf_type_ptr leaf(new leaf_type(args,1));

        {//strategy 1
            strat_type_ptr s(new strat_type(args.leaf_params.size()));
            s->setInput(0,-9.5);
            s->setInput(1,7.5);
            s->setInput(2,9.2);
            //insert double
            s->insert(10);
            s->insert(9);
            s->insert(8);
            s->insert(6);
            s->insert(4);
            //s mean = 7.4 = (9+10+4+8+6)/5
            //quantile 0.5 = 8
            leaf->insert(s);
        }
        {//strategy 2
            strat_type_ptr s(new strat_type(args.leaf_params.size()));
            s->setInput(0,-8.8);
            s->setInput(1,5.3);
            s->setInput(2,9.8);
            //insert double
            s->insert(9);
            s->insert(8);
            s->insert(7);
            s->insert(4);
            s->insert(1);
            //s mean = 5.8 = (1+8+9+7+4)/5
            //quantile 0.5 = 7
            leaf->insert(s);
        }
        {//strategy 3
            strat_type_ptr s(new strat_type(args.leaf_params.size()));
            s->setInput(0,4.8);
            s->setInput(1,-6.3);
            s->setInput(2,0.8);
            //insert double
            s->insert(5);
            s->insert(3);
            s->insert(1);
            s->insert(0);
            s->insert(-1);
            //s mean = 1.6 = (5+0-1+3+1)/5
            //quantile 0.5 = 1
            leaf->insert(s);
        }
        {//strategy 4
            strat_type_ptr s(new strat_type(args.leaf_params.size()));
            s->setInput(0,8.1);
            s->setInput(1,2.4);
            s->setInput(2,-5.1);
            //insert double
            s->insert(9);
            s->insert(8);
            s->insert(6);
            s->insert(5);
            s->insert(3);
            //s mean = 6.2 = (5+0-1+3+1)/5
            //quantile 0.5 = 6
            leaf->insert(s);
        }
        {//strategy 5
            strat_type_ptr s(new strat_type(args.leaf_params.size()));
            s->setInput(0,-0.5);
            s->setInput(1,-2.6);
            s->setInput(2,-9.8);
            //insert double
            s->insert(9);
            s->insert(8);
            s->insert(5);
            s->insert(1);
            s->insert(-3);
            //s mean = 4
            //quantile 0.5 = 5
            leaf->insert(s);
        }
        //leaf stats[0] min : 1.6 (strategy 3) : min(7.4, 5.8, 1.6, 6.2, 4)
        //leaf stats[0] stdDev : 1.6  : sqrt(c(7.4, 5.8, 1.6, 6.2, 4)-5)^2/5)
        return leaf;
    }
    const mp::P2args& getTestArgsP2()
    {
        return args;
    }
};

//Cutting
BOOST_AUTO_TEST_CASE(test_leaf_cut)
{

    {
        const ct::AccuType stat_type = ct::MEAN_VAR;
        typedef p2t1<stat_type>::leaf_type_ptr leaf_type_ptr;
        typedef p2t1<stat_type>::leaf_type leaf_type;

        p2t1<stat_type> t;
        leaf_type_ptr leaf = t.getTestLeaf();
        mp::Cutting<leaf_type, mp::CUT_ON_MIDDLE> cut(t.getTestArgsP2());
        std::pair<leaf_type_ptr, leaf_type_ptr> childs = cut.cuts(leaf);
        leaf_type_ptr c1 = childs.first;
        leaf_type_ptr c2 = childs.second;

        BOOST_REQUIRE_CLOSE(c1->getDomain(0).min(),-10.0,10e-4);
        BOOST_REQUIRE_CLOSE(c1->getDomain(0).max(),10.0,10e-4);
        BOOST_REQUIRE_CLOSE(c1->getDomain(0).minInterval(),1.0,10e-4);
        BOOST_REQUIRE_CLOSE(c1->getDomain(1).min(),-11.0,10e-4);
        BOOST_REQUIRE_CLOSE(c1->getDomain(1).max(),11.0,10e-4);
        BOOST_REQUIRE_CLOSE(c1->getDomain(1).minInterval(),1.0,10e-4);
        BOOST_REQUIRE_CLOSE(c1->getDomain(2).min(),-12.0,10e-4);
        BOOST_REQUIRE_CLOSE(c1->getDomain(2).max(),0.0,10e-4);
        BOOST_REQUIRE_CLOSE(c1->getDomain(2).minInterval(),1.0,10e-4);
        BOOST_REQUIRE_CLOSE(c2->getDomain(2).min(),0.0,10e-4);
        BOOST_REQUIRE_CLOSE(c2->getDomain(2).max(),12.0,10e-4);
    }
    {
        const ct::AccuType stat_type = ct::QUANTILE;
        typedef p2t1<stat_type>::leaf_type_ptr leaf_type_ptr;
        typedef p2t1<stat_type>::leaf_type leaf_type;
        p2t1<stat_type> t;

        leaf_type_ptr leaf = t.getTestLeaf();
        BOOST_REQUIRE_CLOSE(leaf->access(0)->quantile(0.5),8.0,10e-4);
        mp::Cutting<leaf_type, mp::CUT_ON_PIVOT> cut(t.getTestArgsP2());
        std::pair<leaf_type_ptr, leaf_type_ptr> childs = cut.cuts(leaf);
        leaf_type_ptr c1 = childs.first;
        leaf_type_ptr c2 = childs.second;
        BOOST_REQUIRE_CLOSE(c1->getDomain(2).min(),-12.0,10e-4);
        BOOST_REQUIRE_CLOSE(c1->getDomain(2).max(),-5.1,10e-4);
        BOOST_REQUIRE_CLOSE(c2->getDomain(2).min(),-5.1,10e-4);
        BOOST_REQUIRE_CLOSE(c2->getDomain(2).max(),12.0,10e-4);
    }
    {
        const ct::AccuType stat_type = ct::MEAN_VAR;
        typedef p2t1<stat_type>::leaf_type_ptr leaf_type_ptr;
        typedef p2t1<stat_type>::leaf_type leaf_type;
        p2t1<stat_type> t;
        leaf_type_ptr leaf = t.getTestLeaf();
        mp::Cutting<leaf_type, mp::CUT_ON_LARGEST_MEAN_DIFF>cut(
            t.getTestArgsP2());
        std::pair<leaf_type_ptr, leaf_type_ptr> childs = cut.cuts(leaf);
        leaf_type_ptr c1 = childs.first;
        leaf_type_ptr c2 = childs.second;
        BOOST_REQUIRE_CLOSE(c1->getDomain(2).min(),-12.0,10e-4);
        BOOST_REQUIRE_CLOSE(c1->getDomain(2).max(),5.0,10e-4);
        BOOST_REQUIRE_CLOSE(c2->getDomain(2).min(),5.0,10e-4);
        BOOST_REQUIRE_CLOSE(c2->getDomain(2).max(),12.0,10e-4);
    }
    {
        const ct::AccuType stat_type = ct::QUANTILE;
        typedef p2t1<stat_type>::leaf_type_ptr leaf_type_ptr;
        typedef p2t1<stat_type>::leaf_type leaf_type;
        p2t1<stat_type> t;
        leaf_type_ptr leaf = t.getTestLeaf();
        mp::P2args p2args(t.getTestArgsP2());
        p2args.cut_leaf_quantile_order = 0.5;
        p2args.cut_strat_quantile_order = 0.5;
        mp::Cutting<leaf_type, mp::CUT_ON_LARGEST_QTL_DIFF> cut(p2args);
        std::pair<leaf_type_ptr, leaf_type_ptr> childs = cut.cuts(leaf);
        leaf_type_ptr c1 = childs.first;
        leaf_type_ptr c2 = childs.second;
        BOOST_REQUIRE_CLOSE(c1->getDomain(2).min(),-12.0,10e-4);
        BOOST_REQUIRE_CLOSE(c1->getDomain(2).max(),5.0,10e-4);
        BOOST_REQUIRE_CLOSE(c2->getDomain(2).min(),5.0,10e-4);
        BOOST_REQUIRE_CLOSE(c2->getDomain(2).max(),12.0,10e-4);
    }
}

//Evaluation
BOOST_AUTO_TEST_CASE(test_leaf_eval)
{
    {
        const ct::AccuType stat_type = ct::MEAN_VAR;
        typedef p2t1<stat_type, mp::LEAF_PERF_MEAN, mp::LEAF_DISP_GLOBAL_VAR>
        p2t1impl;
        typedef p2t1impl::leaf_type_ptr leaf_type_ptr;
        typedef p2t1impl::leaf_type leaf_type;
        p2t1impl t;
        leaf_type_ptr leaf = t.getTestLeaf();
        BOOST_REQUIRE_CLOSE(leaf->performance(),5.0,10e-4);
        BOOST_REQUIRE_CLOSE(leaf->dispersion(),12.56,10e-4);
    }
    {
        const ct::AccuType stat_type = ct::MEAN_VAR;
        typedef p2t1<stat_type, mp::LEAF_PERF_MIN, mp::LEAF_DISP_SPATIAL_VAR1>
        p2t1impl;
        typedef p2t1impl::leaf_type_ptr leaf_type_ptr;
        typedef p2t1impl::leaf_type leaf_type;
        p2t1impl t;
        leaf_type_ptr leaf = t.getTestLeaf();
        BOOST_REQUIRE_CLOSE(leaf->performance(),1.6,10e-4);
        BOOST_REQUIRE_CLOSE(leaf->dispersion(),4.08,10e-4);
    }
    {
        const ct::AccuType stat_type = ct::STORE;
        typedef p2t1<stat_type, mp::LEAF_PERF_QUANTILE, mp::LEAF_DISP_SPATIAL_VAR2>
        p2t1impl;
        typedef p2t1impl::leaf_type_ptr leaf_type_ptr;
        typedef p2t1impl::leaf_type leaf_type;
        p2t1impl t;
        leaf_type_ptr leaf = t.getTestLeaf();
        BOOST_REQUIRE_CLOSE(leaf->performance(),6.0,10e-4);
        BOOST_REQUIRE_CLOSE(leaf->dispersion(),4.79,10e-4);
    }
}

//Selection
BOOST_AUTO_TEST_CASE(test_leaf_selection)
{
    const ct::AccuType stat_type = ct::MEAN_VAR;
    typedef p2t1<stat_type, mp::LEAF_PERF_MEAN, mp::LEAF_DISP_GLOBAL_VAR>
    p2t1impl;
    typedef p2t1impl::leaf_type_ptr leaf_type_ptr;
    typedef p2t1impl::leaf_type leaf_type;
    p2t1impl t;
    leaf_type_ptr leaf = t.getTestLeaf();
    std::vector<leaf_type_ptr> pendingLeaves;
    pendingLeaves.push_back(leaf);
    {
        mp::LeafComparator<leaf_type, ct::MonoDim, mp::LEAF_SEL_BETA> comp;
        BOOST_REQUIRE_CLOSE(comp.score(leaf),13.32360,10e-4);
    }
    {
        mp::LeafComparator<leaf_type, ct::MonoDim, mp::LEAF_SEL_CONC_BETA> comp;
        BOOST_REQUIRE_CLOSE(comp.score(leaf),0.7548559,10e-4);
    }
}

//Optimisation
BOOST_AUTO_TEST_CASE(test_optim_p2)
{

    //build the vpz optimizer
    vu::Package pkg("CallForOptim");
    vz::Vpz vpz(pkg.getExpFile("UnitTestMethods.vpz"));
    //set conditions "mod" and "random" to the optimizer
    std::vector<std::string> conds;
    conds.push_back("FOsim");
    conds.push_back("p2");
    conds.push_back("mode");
    ct::VleAPIfacilities::attachConditions(vpz,"Experiments",conds);
    //set the dynamic FOmethod_random to the experiment
    ct::VleAPIfacilities::setDynamics(vpz,"Experiments","dFOp2");
    //set the view of the experiment to storage plugin
    ct::VleAPIfacilities::changePlugin(vpz,"view","storage");
    //launch optimization
    vu::ModuleManager modules;
    vm::Simulation simpleSimulator(vm::LOG_NONE, vm::SIMULATION_NONE, NULL);
    vm::Error error;
    vv::Map* res = simpleSimulator.run(new vz::Vpz(vpz), modules, &error);
    bool simOk = (error.code ==0);
    if (!simOk) {
        std::cout << " error.message " << error.message << std::endl;
    }
    BOOST_REQUIRE(simOk);
    //recover optimiser results
    const vv::Value* val = ct::VleAPIfacilities::findLastOutputValue(
        *res, "view", "UnitTestMethods:Experiments", "min-performance");
    double valDouble = val->toDouble().value();
    //check optimized value
    //very bad value, require more sims thant 200, or new parameters
    //(should be close to 0)
    BOOST_REQUIRE_CLOSE(valDouble,5690.7078806265963,10e-2);
}

