/**
 * @file test/cfo/tools/tests_tools.cpp
 * @author The RECORD Development Team (INRA) and the VLE Development Team
 * See the AUTHORS or Authors.txt file
 */

/*
 * Copyright (C) 2009-2010 INRA
 *
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#define BOOST_TEST_MAIN
#define BOOST_AUTO_TEST_MAIN
#define BOOST_TEST_DYN_LINK
#define BOOST_TEST_MODULE tests_cfo
#include <boost/test/unit_test.hpp>
#include <boost/test/auto_unit_test.hpp>
#include <boost/test/floating_point_comparison.hpp>
#include <boost/lexical_cast.hpp>
#include <stdexcept>
#include <limits>
#include <fstream>
#include <iostream>

#include <cfo/tools.hpp>

#include <vle/value/Double.hpp>
#include <vle/value/Set.hpp>
#include <vle/value/Map.hpp>
#include <vle/value/String.hpp>
#include <vle/vpz/Vpz.hpp>


namespace ct = cfo::tools;
namespace ctp = cfo::tools::policies;
namespace ctr = cfo::tools::regression;
namespace ctd = cfo::tools::doe;
namespace vv = vle::value;
namespace vz = vle::vpz;

//Domain
BOOST_AUTO_TEST_CASE(test_leaf)
{
    {
        std::vector<ct::DCDomain> params;
        params.push_back(ct::DCDomain(-10, 10, 1));
        params.push_back(ct::DCDomain(-11, 10.1, 1));
        params.push_back(ct::DCDomain(-12, 12.6, 1));
        ct::Leaf leaf1(params);
        BOOST_REQUIRE_EQUAL(leaf1.getNbPotentialSubLeaves(),10999);
    }
}

//VleSimulator
BOOST_AUTO_TEST_CASE(test_simulator)
{
    vv::Map init;
    vv::Map& mod = init.addMap("model");
    mod.addString("package","CallForOptim");
    mod.addString("vpz","ExBohachevsky.vpz");
    vv::Set& inputs = mod.addSet("inputs");
    inputs.addString("x1");
    inputs.addString("x2");
    vv::Set& outputs = mod.addSet("outputs");
    outputs.addString("y");
    ct::VleSimulator sim(init);
    vv::Set point;
    point.addDouble(3);
    point.addDouble(-10);
    double out = sim.simulate(point).toDouble().value();
    BOOST_REQUIRE_CLOSE(out,209.6,10e-4);
}

//VleSimulator integration
BOOST_AUTO_TEST_CASE(test_simulator2)
{
    //build an InitEventList from conditions of the optimizer
    vu::Package pkg("CallForOptim");
    vz::Vpz vpz(pkg.getExpFile("UnitTestMethods.vpz"));
    std::vector<std::string> conds;
    conds.push_back("FOsim");
    vv::Map* init = ct::VleAPIfacilities::buildInitEventList(vpz,conds);
    {
        ct::VleSimulator sim(*init);
        vv::Set point;
        point.addDouble(0);
        point.addDouble(0);
        double out = sim.simulate(point).toDouble().value();
        BOOST_CHECK(((out < 10e-5)&&(out > -10e-5)));
        point.set(0,vv::Double(3));
        point.set(1,vv::Double(-10));
        out = sim.simulate(point).toDouble().value();
        BOOST_REQUIRE_CLOSE(out,209.6,10e-4);
    }
    delete init;
}





//Policies
BOOST_AUTO_TEST_CASE(test_policies)
{
    vu::Rand r;
    ctp::Qmatrix qm;
    vv::Set actions;
    actions.addInt(0);
    actions.addInt(1);
    actions.addInt(2);
    qm.initFromActions(actions);
    qm.getState(vv::String("state1"));
    qm.getState(vv::String("state2"));
    qm.updateSarsa(qm.getState(vv::String("state1")),qm.getAction(vv::Integer(0)),
        qm.getState(vv::String("state2")),qm.getAction(vv::Integer(1)),
            -10, 0.5 /*gama*/, 0.2 /* alpha*/);
    double qs1a0 = qm.getQ(vv::String("state1"),vv::Integer(0));
    BOOST_REQUIRE_CLOSE(qs1a0, -2.0,10e-4);
    qm.updateSarsa(qm.getState(vv::String("state2")),qm.getAction(vv::Integer(1)),
        qm.getState(vv::String("state1")),qm.getAction(vv::Integer(0)),
            -3, 0.5 /*gama*/, 0.2 /* alpha*/);
    double qs2a1 = qm.getQ(vv::String("state2"),vv::Integer(1));
    BOOST_REQUIRE_CLOSE(qs2a1, -0.8,10e-4);
}



//FullFactorial doe
BOOST_AUTO_TEST_CASE(test_FullFactorial)
{
    std::vector<unsigned int > levels(3);
    levels[0] = 3;
    levels[1] = 2;
    levels[2] = 4;
    {
        //Test without bounds
        ctd::FullFactorial ff(levels);
        std::cout << " ff " << ff << std::endl;
        const vv::Tuple& point = ff.point(9);
        std::cout << " ff.point(9) " << point << std::endl;
        BOOST_REQUIRE_CLOSE(point.at(0),1.0,10e-4);
        BOOST_REQUIRE_CLOSE(point.at(1),0.0,10e-4);
        BOOST_REQUIRE_CLOSE(point.at(2),1.0,10e-4);
    }
    {
        //Test with bounds
        vv::Set bounds;
        bounds.addTuple();
        vv::Tuple& t0 = bounds.getTuple(0);
        t0.add(10); t0.add(15);
        bounds.addTuple();
        vv::Tuple& t1 = bounds.getTuple(1);
        t1.add(10); t1.add(20);
        bounds.addTuple();
        vv::Tuple& t2 = bounds.getTuple(2);
        t2.add(10); t2.add(25);
        ctd::FullFactorial ff(levels, bounds);
        std::cout << " ff " << ff << std::endl;
        const vv::Tuple& point = ff.point(15);
        std::cout << " ff.point(15) " << point << std::endl;
        BOOST_REQUIRE_CLOSE(point.at(0),12.5,10e-4);
        BOOST_REQUIRE_CLOSE(point.at(1),16.6667,10e-4);
        BOOST_REQUIRE_CLOSE(point.at(2),22,10e-4);
    }
}

